@@ -21,6 +21,11 @@ else
     MISTRAL_BIN_DIR=$(get_python_exec_prefix)
 fi
 
+# Toggle for deploying Mistral API under HTTPD + mod_wsgi
+MISTRAL_USE_MOD_WSGI=${MISTRAL_USE_MOD_WSGI:-True}
+
+MISTRAL_FILES_DIR=$MISTRAL_DIR/devstack/files
+
 # create_mistral_accounts - Set up common required mistral accounts
 #
 # Tenant      User       Roles
@@ -33,14 +38,12 @@ function create_mistral_accounts {
 
     create_service_user "mistral" "admin"
 
-    if [[ "$KEYSTONE_CATALOG_BACKEND" = 'sql' ]]; then
-        get_or_create_service "mistral" "workflowv2" "Workflow Service v2"
-        get_or_create_endpoint "workflowv2" \
-            "$REGION_NAME" \
-            "$MISTRAL_SERVICE_PROTOCOL://$MISTRAL_SERVICE_HOST:$MISTRAL_SERVICE_PORT/v2" \
-            "$MISTRAL_SERVICE_PROTOCOL://$MISTRAL_SERVICE_HOST:$MISTRAL_SERVICE_PORT/v2" \
-            "$MISTRAL_SERVICE_PROTOCOL://$MISTRAL_SERVICE_HOST:$MISTRAL_SERVICE_PORT/v2"
-    fi
+    get_or_create_service "mistral" "workflowv2" "Workflow Service v2"
+    get_or_create_endpoint "workflowv2" \
+        "$REGION_NAME" \
+        "$MISTRAL_SERVICE_PROTOCOL://$MISTRAL_SERVICE_HOST:$MISTRAL_SERVICE_PORT/v2" \
+        "$MISTRAL_SERVICE_PROTOCOL://$MISTRAL_SERVICE_HOST:$MISTRAL_SERVICE_PORT/v2" \
+        "$MISTRAL_SERVICE_PROTOCOL://$MISTRAL_SERVICE_HOST:$MISTRAL_SERVICE_PORT/v2"
 }
 
 
@@ -56,15 +59,17 @@ function mkdir_chown_stack {
 
 # configure_mistral - Set config files, create data dirs, etc
 function configure_mistral {
+
+    # create and clean up auth cache dir
+    mkdir_chown_stack "$MISTRAL_AUTH_CACHE_DIR"
+    rm -f "$MISTRAL_AUTH_CACHE_DIR"/*
+
     mkdir_chown_stack "$MISTRAL_CONF_DIR"
 
     # Generate Mistral configuration file and configure common parameters.
     oslo-config-generator --config-file $MISTRAL_DIR/tools/config/config-generator.mistral.conf --output-file $MISTRAL_CONF_FILE
     iniset $MISTRAL_CONF_FILE DEFAULT debug $MISTRAL_DEBUG
 
-    MISTRAL_POLICY_FILE=$MISTRAL_CONF_DIR/policy.json
-    cp $MISTRAL_DIR/etc/policy.json $MISTRAL_POLICY_FILE
-
     # Run all Mistral processes as a single process
     iniset $MISTRAL_CONF_FILE DEFAULT server all
 
@@ -72,17 +77,11 @@ function configure_mistral {
     #-------------------------
 
     # Setup keystone_authtoken section
-    iniset $MISTRAL_CONF_FILE keystone_authtoken auth_host $KEYSTONE_AUTH_HOST
-    iniset $MISTRAL_CONF_FILE keystone_authtoken auth_port $KEYSTONE_AUTH_PORT
-    iniset $MISTRAL_CONF_FILE keystone_authtoken auth_protocol $KEYSTONE_AUTH_PROTOCOL
-    iniset $MISTRAL_CONF_FILE keystone_authtoken admin_tenant_name $SERVICE_TENANT_NAME
-    iniset $MISTRAL_CONF_FILE keystone_authtoken admin_user $MISTRAL_ADMIN_USER
-    iniset $MISTRAL_CONF_FILE keystone_authtoken admin_password $SERVICE_PASSWORD
-    iniset $MISTRAL_CONF_FILE keystone_authtoken auth_uri "http://${KEYSTONE_AUTH_HOST}:5000/v3"
+    configure_auth_token_middleware $MISTRAL_CONF_FILE mistral $MISTRAL_AUTH_CACHE_DIR
+    iniset $MISTRAL_CONF_FILE keystone_authtoken auth_uri $KEYSTONE_AUTH_URI_V3
 
     # Setup RabbitMQ credentials
-    iniset $MISTRAL_CONF_FILE oslo_messaging_rabbit rabbit_userid $RABBIT_USERID
-    iniset $MISTRAL_CONF_FILE oslo_messaging_rabbit rabbit_password $RABBIT_PASSWORD
+    iniset_rpc_backend mistral $MISTRAL_CONF_FILE
 
     # Configure the database.
     iniset $MISTRAL_CONF_FILE database connection `database_connection_url mistral`
@@ -92,9 +91,6 @@ function configure_mistral {
     # Configure action execution deletion policy
     iniset $MISTRAL_CONF_FILE api allow_action_execution_deletion True
 
-    # Path of policy.json file.
-    iniset $MISTRAL_CONF oslo_policy policy_file $MISTRAL_POLICY_FILE
-
     if [ "$LOG_COLOR" == "True" ] && [ "$SYSLOG" == "False" ]; then
         setup_colorized_logging $MISTRAL_CONF_FILE DEFAULT tenant user
     fi
@@ -102,6 +98,10 @@ function configure_mistral {
     if [ "$MISTRAL_RPC_IMPLEMENTATION" ]; then
         iniset $MISTRAL_CONF_FILE DEFAULT rpc_implementation $MISTRAL_RPC_IMPLEMENTATION
     fi
+
+    if [ "$MISTRAL_USE_MOD_WSGI" == "True" ]; then
+        _config_mistral_apache_wsgi
+    fi
 }
 
 
@@ -109,7 +109,7 @@ function configure_mistral {
 function init_mistral {
     # (re)create Mistral database
     recreate_database mistral utf8
-    python $MISTRAL_DIR/tools/sync_db.py --config-file $MISTRAL_CONF_FILE
+    $PYTHON $MISTRAL_DIR/tools/sync_db.py --config-file $MISTRAL_CONF_FILE
 }
 
 
@@ -123,13 +123,17 @@ function install_mistral {
     if is_service_enabled horizon; then
         _install_mistraldashboard
     fi
+
+    if [ "$MISTRAL_USE_MOD_WSGI" == "True" ]; then
+        install_apache_wsgi
+    fi
 }
 
 
 function _install_mistraldashboard {
     git_clone $MISTRAL_DASHBOARD_REPO $MISTRAL_DASHBOARD_DIR $MISTRAL_DASHBOARD_BRANCH
     setup_develop $MISTRAL_DASHBOARD_DIR
-    ln -fs $MISTRAL_DASHBOARD_DIR/_50_mistral.py.example $HORIZON_DIR/openstack_dashboard/local/enabled/_50_mistral.py
+    ln -fs $MISTRAL_DASHBOARD_DIR/mistraldashboard/enabled/_50_mistral.py $HORIZON_DIR/openstack_dashboard/local/enabled/_50_mistral.py
 }
 
 
@@ -144,12 +148,27 @@ function install_mistral_pythonclient {
     fi
 }
 
+function install_mistral_lib {
+    if use_library_from_git "mistral-lib"; then
+        git_clone $MISTRAL_LIB_REPO $MISTRAL_LIB_DIR $MISTRAL_LIB_BRANCH
+        setup_develop $MISTRAL_LIB_DIR
+    fi
+}
 
-# start_mistral - Start running processes, including screen
+# start_mistral - Start running processes
 function start_mistral {
+    # If the site is not enabled then we are in a grenade scenario
+    local enabled_site_file
+    enabled_site_file=$(apache_site_config_for mistral-api)
+
     if is_service_enabled mistral-api && is_service_enabled mistral-engine && is_service_enabled mistral-executor && is_service_enabled mistral-event-engine ; then
         echo_summary "Installing all mistral services in separate processes"
-        run_process mistral-api "$MISTRAL_BIN_DIR/mistral-server --server api --config-file $MISTRAL_CONF_DIR/mistral.conf"
+        if [ -f ${enabled_site_file} ] && [ "$MISTRAL_USE_MOD_WSGI" == "True" ]; then
+            enable_apache_site mistral-api
+            restart_apache_server
+        else
+            run_process mistral-api "$MISTRAL_BIN_DIR/mistral-server --server api --config-file $MISTRAL_CONF_DIR/mistral.conf"
+        fi
         run_process mistral-engine "$MISTRAL_BIN_DIR/mistral-server --server engine --config-file $MISTRAL_CONF_DIR/mistral.conf"
         run_process mistral-executor "$MISTRAL_BIN_DIR/mistral-server --server executor --config-file $MISTRAL_CONF_DIR/mistral.conf"
         run_process mistral-event-engine "$MISTRAL_BIN_DIR/mistral-server --server event-engine --config-file $MISTRAL_CONF_DIR/mistral.conf"
@@ -162,10 +181,17 @@ function start_mistral {
 
 # stop_mistral - Stop running processes
 function stop_mistral {
-    # Kill the Mistral screen windows
-    for serv in mistral mistral-api mistral-engine mistral-executor mistral-event-engine; do
+    local serv
+    for serv in mistral mistral-engine mistral-executor mistral-event-engine; do
         stop_process $serv
     done
+
+    if [ "$MISTRAL_USE_MOD_WSGI" == "True" ]; then
+        disable_apache_site mistral-api
+        restart_apache_server
+    else
+        stop_process mistral-api
+    fi
 }
 
 
@@ -173,6 +199,11 @@ function cleanup_mistral {
     if is_service_enabled horizon; then
         _mistral_cleanup_mistraldashboard
     fi
+
+    if [ "$MISTRAL_USE_MOD_WSGI" == "True" ]; then
+        _mistral_cleanup_apache_wsgi
+    fi
+    sudo rm -rf $MISTRAL_CONF_DIR
 }
 
 
@@ -180,16 +211,44 @@ function _mistral_cleanup_mistraldashboard {
     rm -f $HORIZON_DIR/openstack_dashboard/local/enabled/_50_mistral.py
 }
 
+function _mistral_cleanup_apache_wsgi {
+    sudo rm -f $(apache_site_config_for mistral-api)
+}
+
+# _config_mistral_apache_wsgi() - Set WSGI config files for Mistral
+function _config_mistral_apache_wsgi {
+    local mistral_apache_conf
+    mistral_apache_conf=$(apache_site_config_for mistral-api)
+    local mistral_ssl=""
+    local mistral_certfile=""
+    local mistral_keyfile=""
+    local mistral_api_port=$MISTRAL_SERVICE_PORT
+    local venv_path=""
+
+    sudo cp $MISTRAL_FILES_DIR/apache-mistral-api.template $mistral_apache_conf
+    sudo sed -e "
+        s|%PUBLICPORT%|$mistral_api_port|g;
+        s|%APACHE_NAME%|$APACHE_NAME|g;
+        s|%MISTRAL_BIN_DIR%|$MISTRAL_BIN_DIR|g;
+        s|%API_WORKERS%|$API_WORKERS|g;
+        s|%SSLENGINE%|$mistral_ssl|g;
+        s|%SSLCERTFILE%|$mistral_certfile|g;
+        s|%SSLKEYFILE%|$mistral_keyfile|g;
+        s|%USER%|$STACK_USER|g;
+        s|%VIRTUALENV%|$venv_path|g
+    " -i $mistral_apache_conf
+}
 
 if is_service_enabled mistral; then
     if [[ "$1" == "stack" && "$2" == "install" ]]; then
         echo_summary "Installing mistral"
         install_mistral
+        install_mistral_lib
         install_mistral_pythonclient
     elif [[ "$1" == "stack" && "$2" == "post-config" ]]; then
         echo_summary "Configuring mistral"
-        configure_mistral
         create_mistral_accounts
+        configure_mistral
     elif [[ "$1" == "stack" && "$2" == "extra" ]]; then
         echo_summary "Initializing mistral"
         init_mistral
