@@ -18,9 +18,10 @@ from mistral.actions.openstack.action_generator import base
 
 
 SUPPORTED_MODULES = [
-    'Nova', 'Glance', 'Keystone', 'Heat', 'Neutron', 'Cinder', 'Ceilometer',
-    'Trove', 'Ironic', 'Baremetal Introspection', 'Swift', 'Zaqar', 'Barbican',
-    'Mistral', 'Designate', 'Magnum', 'Murano', 'Tacker'
+    'Nova', 'Glance', 'Keystone', 'Heat', 'Neutron', 'Cinder',
+    'Trove', 'Ironic', 'Baremetal Introspection', 'Swift', 'SwiftService',
+    'Zaqar', 'Barbican', 'Mistral', 'Designate', 'Magnum', 'Murano', 'Tacker',
+    'Aodh', 'Gnocchi', 'Glare', 'Vitrage', 'Senlin', 'Zun', 'Qinling', 'Manila'
 ]
 
 
