@@ -21,9 +21,9 @@ import pecan
 from mistral.api import access_control
 from mistral import config as m_config
 from mistral import context as ctx
-from mistral import coordination
 from mistral.db.v2 import api as db_api_v2
-from mistral.engine.rpc_backend import rpc
+from mistral.rpc import base as rpc
+from mistral.service import coordination
 from mistral.services import periodic
 
 
@@ -53,14 +53,16 @@ def setup_app(config=None):
 
     db_api_v2.setup_db()
 
-    if not app_conf.pop('disable_cron_trigger_thread', False):
+    # TODO(rakhmerov): Why do we run cron triggers in the API layer?
+    # Should we move it to engine?s
+    if cfg.CONF.cron_trigger.enabled:
         periodic.setup()
 
     coordination.Service('api_group').register_membership()
 
     app = pecan.make_app(
         app_conf.pop('root'),
-        hooks=lambda: [ctx.ContextHook(), ctx.AuthHook()],
+        hooks=lambda: [ctx.AuthHook(), ctx.ContextHook()],
         logging=getattr(config, 'logging', {}),
         **app_conf
     )
@@ -68,6 +70,7 @@ def setup_app(config=None):
     # Set up access control.
     app = access_control.setup(app)
 
+    # TODO(rakhmerov): need to get rid of this call.
     # Set up RPC related flags in config
     rpc.get_transport()
 
@@ -82,3 +85,12 @@ def setup_app(config=None):
     # Create a CORS wrapper, and attach mistral-specific defaults that must be
     # included in all CORS responses.
     return cors_middleware.CORS(app, cfg.CONF)
+
+
+def init_wsgi():
+    # By default, oslo.config parses the CLI args if no args is provided.
+    # As a result, invoking this wsgi script from gunicorn leads to the error
+    # with argparse complaining that the CLI options have already been parsed.
+    m_config.parse_args(args=[])
+
+    return setup_app()
