@@ -1,4 +1,5 @@
 # Copyright 2013 - Mirantis, Inc.
+# Copyright 2018 - Extreme Networks, Inc.
 #
 #    Licensed under the Apache License, Version 2.0 (the "License");
 #    you may not use this file except in compliance with the License.
@@ -17,22 +18,37 @@ from wsme import types as wtypes
 
 from mistral.api.controllers import resource
 from mistral.api.controllers.v2 import types
+from mistral import exceptions as exc
+from mistral import utils
 from mistral.workflow import states
 
 SCOPE_TYPES = wtypes.Enum(str, 'private', 'public')
 
 
-class Workbook(resource.Resource):
+class ScopedResource(object):
+    """Utilities for scoped resources"""
+    @classmethod
+    def validate_scope(cls, scope):
+        if scope not in SCOPE_TYPES.values:
+            raise exc.InvalidModelException(
+                "Scope must be one of the following: %s; actual: "
+                "%s" % (SCOPE_TYPES.values, scope)
+            )
+
+
+class Workbook(resource.Resource, ScopedResource):
     """Workbook resource."""
 
     id = wtypes.text
     name = wtypes.text
+    namespace = wtypes.text
 
     definition = wtypes.text
     "workbook definition in Mistral v2 DSL"
     tags = [wtypes.text]
     scope = SCOPE_TYPES
     "'private' or 'public'"
+    project_id = wsme.wsattr(wtypes.text, readonly=True)
 
     created_at = wtypes.text
     updated_at = wtypes.text
@@ -45,8 +61,10 @@ class Workbook(resource.Resource):
                         'WORKBOOK DEFINITION IN MISTRAL DSL v2',
                    tags=['large', 'expensive'],
                    scope='private',
+                   project_id='a7eb669e9819420ea4bd1453e672c0a7',
                    created_at='1970-01-01T00:00:00.000000',
-                   updated_at='1970-01-01T00:00:00.000000')
+                   updated_at='1970-01-01T00:00:00.000000',
+                   namespace='')
 
 
 class Workbooks(resource.ResourceList):
@@ -64,11 +82,12 @@ class Workbooks(resource.ResourceList):
         return cls(workbooks=[Workbook.sample()])
 
 
-class Workflow(resource.Resource):
+class Workflow(resource.Resource, ScopedResource):
     """Workflow resource."""
 
     id = wtypes.text
     name = wtypes.text
+    namespace = wtypes.text
     input = wtypes.text
 
     definition = wtypes.text
@@ -92,19 +111,16 @@ class Workflow(resource.Resource):
                    scope='private',
                    project_id='a7eb669e9819420ea4bd1453e672c0a7',
                    created_at='1970-01-01T00:00:00.000000',
-                   updated_at='1970-01-01T00:00:00.000000')
+                   updated_at='1970-01-01T00:00:00.000000',
+                   namespace='')
 
     @classmethod
-    def from_dict(cls, d):
-        e = cls()
+    def _set_input(cls, obj, wf_spec):
         input_list = []
 
-        for key, val in d.items():
-            if hasattr(e, key):
-                setattr(e, key, val)
+        if wf_spec:
+            input = wf_spec.get('input', [])
 
-        if 'spec' in d:
-            input = d.get('spec', {}).get('input', [])
             for param in input:
                 if isinstance(param, dict):
                     for k, v in param.items():
@@ -112,9 +128,37 @@ class Workflow(resource.Resource):
                 else:
                     input_list.append(param)
 
-            setattr(e, 'input', ", ".join(input_list) if input_list else '')
+            setattr(obj, 'input', ", ".join(input_list) if input_list else '')
 
-        return e
+        return obj
+
+    @classmethod
+    def from_dict(cls, d):
+        obj = super(Workflow, cls).from_dict(d)
+
+        return cls._set_input(obj, d.get('spec'))
+
+    @classmethod
+    def from_db_model(cls, db_model):
+        obj = super(Workflow, cls).from_db_model(db_model)
+
+        return cls._set_input(obj, db_model.spec)
+
+    @classmethod
+    def from_tuples(cls, tuple_iterator):
+        obj = cls()
+        spec = None
+        for col_name, col_val in tuple_iterator:
+            if hasattr(obj, col_name):
+                # Convert all datetime values to strings.
+                setattr(obj, col_name, utils.datetime_to_str(col_val))
+            if col_name == 'spec':
+                spec = col_val
+
+        if spec:
+            obj = cls._set_input(obj, spec)
+
+        return obj
 
 
 class Workflows(resource.ResourceList):
@@ -139,7 +183,7 @@ class Workflows(resource.ResourceList):
         return workflows_sample
 
 
-class Action(resource.Resource):
+class Action(resource.Resource, ScopedResource):
     """Action resource.
 
     NOTE: *name* is immutable. Note that name and description get inferred
@@ -157,6 +201,7 @@ class Action(resource.Resource):
     tags = [wtypes.text]
     definition = wtypes.text
     scope = SCOPE_TYPES
+    project_id = wsme.wsattr(wtypes.text, readonly=True)
 
     created_at = wtypes.text
     updated_at = wtypes.text
@@ -169,6 +214,7 @@ class Action(resource.Resource):
             definition='HERE GOES ACTION DEFINITION IN MISTRAL DSL v2',
             tags=['large', 'expensive'],
             scope='private',
+            project_id='a7eb669e9819420ea4bd1453e672c0a7',
             created_at='1970-01-01T00:00:00.000000',
             updated_at='1970-01-01T00:00:00.000000'
         )
@@ -201,24 +247,56 @@ class Execution(resource.Resource):
     """Execution resource."""
 
     id = wtypes.text
-    "id is immutable and auto assigned."
-
-    workflow_name = wtypes.text
-    "reference to workflow definition"
+    "execution ID. It is immutable and auto assigned or determined by the API "
+    "client on execution creation. "
+    "If it's passed to POST method from a client it'll be assigned to the "
+    "newly created execution object, but only if an execution with such ID "
+    "doesn't exist. If it exists, then the endpoint will just return "
+    "execution properties in JSON."
 
     workflow_id = wtypes.text
-    "reference to workflow ID"
+    "workflow ID"
+
+    workflow_name = wtypes.text
+    "workflow name"
+
+    workflow_namespace = wtypes.text
+    """Workflow namespace. The workflow namespace is also saved
+     under params and passed to all sub-workflow executions. When looking for
+     the next sub-workflow to run, The correct workflow will be found by
+     name and namespace, where the namespace can be the workflow namespace or
+     the default namespace. Workflows in the same namespace as the top workflow
+     will be given a higher priority."""
 
     description = wtypes.text
-    "description of workflow execution."
+    "description of workflow execution"
 
     params = types.jsontype
-    "params define workflow type specific parameters. For example, reverse \
-    workflow takes one parameter 'task_name' that defines a target task."
+    """'params' define workflow type specific parameters. Specific parameters
+     are:
+     'task_name' - the name of the target task. Only for reverse workflows.
+     'env' - A string value containing the name of the stored environment
+     object or a dictionary with the environment variables used during
+     workflow execution and accessible as 'env()' from within expressions
+     (YAQL or Jinja) defined in the workflow text.
+     'evaluate_env' - If present, controls whether or not Mistral should
+     recursively find and evaluate all expressions (YAQL or Jinja) within
+     the specified environment (via 'env' parameter). 'True' - evaluate
+     all expressions recursively in the environment structure. 'False' -
+     don't evaluate expressions. 'True' by default.
+    """
 
     task_execution_id = wtypes.text
     "reference to the parent task execution"
 
+    root_execution_id = wtypes.text
+    "reference to the root execution"
+
+    source_execution_id = wtypes.text
+    """reference to a workflow execution id which will signal the api to
+    perform a lookup of a current workflow_execution and create a replica
+    based on that workflow inputs and parameters"""
+
     state = wtypes.text
     "state can be one of: IDLE, RUNNING, SUCCESS, ERROR, PAUSED"
 
@@ -226,26 +304,51 @@ class Execution(resource.Resource):
     "an optional state information string"
 
     input = types.jsontype
-    "input is a JSON structure containing workflow input values."
+    "input is a JSON structure containing workflow input values"
 
     output = types.jsontype
-    "output is a workflow output."
+    "output is a workflow output"
 
     created_at = wtypes.text
     updated_at = wtypes.text
 
+    project_id = wsme.wsattr(wtypes.text, readonly=True)
+
     @classmethod
     def sample(cls):
-        return cls(id='123e4567-e89b-12d3-a456-426655440000',
-                   workflow_name='flow',
-                   workflow_id='123e4567-e89b-12d3-a456-426655441111',
-                   description='this is the first execution.',
-                   state='SUCCESS',
-                   input={},
-                   output={},
-                   params={'env': {'k1': 'abc', 'k2': 123}},
-                   created_at='1970-01-01T00:00:00.000000',
-                   updated_at='1970-01-01T00:00:00.000000')
+        return cls(
+            id='123e4567-e89b-12d3-a456-426655440000',
+            workflow_name='flow',
+            workflow_namespace='some_namespace',
+            workflow_id='123e4567-e89b-12d3-a456-426655441111',
+            description='this is the first execution.',
+            project_id='40a908dbddfe48ad80a87fb30fa70a03',
+            state='SUCCESS',
+            input={},
+            output={},
+            params={
+                'env': {'k1': 'abc', 'k2': 123},
+                'notify': [
+                    {
+                        'type': 'webhook',
+                        'url': 'http://endpoint/of/webhook',
+                        'headers': {
+                            'Content-Type': 'application/json',
+                            'X-Auth-Token': '123456789'
+                        }
+                    },
+                    {
+                        'type': 'queue',
+                        'topic': 'failover_queue',
+                        'backend': 'rabbitmq',
+                        'host': '127.0.0.1',
+                        'port': 5432
+                    }
+                ]
+            },
+            created_at='1970-01-01T00:00:00.000000',
+            updated_at='1970-01-01T00:00:00.000000'
+        )
 
 
 class Executions(resource.ResourceList):
@@ -276,18 +379,24 @@ class Task(resource.Resource):
 
     id = wtypes.text
     name = wtypes.text
+    type = wtypes.text
 
     workflow_name = wtypes.text
+    workflow_namespace = wtypes.text
     workflow_id = wtypes.text
     workflow_execution_id = wtypes.text
 
     state = wtypes.text
-    "state can take one of the following values: \
-    IDLE, RUNNING, SUCCESS, ERROR, DELAYED"
+    """state can take one of the following values:
+     IDLE, RUNNING, SUCCESS, ERROR, DELAYED"""
 
     state_info = wtypes.text
     "an optional state information string"
 
+    project_id = wsme.wsattr(wtypes.text, readonly=True)
+
+    runtime_context = types.jsontype
+
     result = wtypes.text
     published = types.jsontype
     processed = bool
@@ -309,6 +418,15 @@ class Task(resource.Resource):
             workflow_execution_id='123e4567-e89b-12d3-a456-426655440000',
             name='task',
             state=states.SUCCESS,
+            project_id='40a908dbddfe48ad80a87fb30fa70a03',
+            runtime_context={
+                'triggered_by': [
+                    {
+                        'task_id': '123-123-123',
+                        'event': 'on-success'
+                    }
+                ]
+            },
             result='task result',
             published={'key': 'value'},
             processed=True,
@@ -339,6 +457,7 @@ class ActionExecution(resource.Resource):
     id = wtypes.text
 
     workflow_name = wtypes.text
+    workflow_namespace = wtypes.text
     task_name = wtypes.text
     task_execution_id = wtypes.text
 
@@ -348,12 +467,13 @@ class ActionExecution(resource.Resource):
     tags = [wtypes.text]
     name = wtypes.text
     description = wtypes.text
+    project_id = wsme.wsattr(wtypes.text, readonly=True)
     accepted = bool
     input = types.jsontype
     output = types.jsontype
     created_at = wtypes.text
     updated_at = wtypes.text
-    params = types.jsontype
+    params = types.jsontype  # TODO(rakhmerov): What is this??
 
     @classmethod
     def sample(cls):
@@ -368,12 +488,13 @@ class ActionExecution(resource.Resource):
             tags=['foo', 'fee'],
             name='std.echo',
             description='My running action',
+            project_id='40a908dbddfe48ad80a87fb30fa70a03',
             accepted=True,
             input={'first_name': 'John', 'last_name': 'Doe'},
             output={'some_output': 'Hello, John Doe!'},
             created_at='1970-01-01T00:00:00.000000',
             updated_at='1970-01-01T00:00:00.000000',
-            params={'save_result': True}
+            params={'save_result': True, "run_sync": False}
         )
 
 
@@ -401,6 +522,7 @@ class CronTrigger(resource.Resource):
     workflow_id = wtypes.text
     workflow_input = types.jsontype
     workflow_params = types.jsontype
+    project_id = wsme.wsattr(wtypes.text, readonly=True)
 
     scope = SCOPE_TYPES
 
@@ -421,6 +543,7 @@ class CronTrigger(resource.Resource):
             workflow_id='123e4567-e89b-12d3-a456-426655441111',
             workflow_input={},
             workflow_params={},
+            project_id='40a908dbddfe48ad80a87fb30fa70a03',
             scope='private',
             pattern='* * * * *',
             remaining_executions=42,
@@ -452,6 +575,7 @@ class Environment(resource.Resource):
     description = wtypes.text
     variables = types.jsontype
     scope = SCOPE_TYPES
+    project_id = wsme.wsattr(wtypes.text, readonly=True)
     created_at = wtypes.text
     updated_at = wtypes.text
 
@@ -468,6 +592,7 @@ class Environment(resource.Resource):
                 'verbose': True
             },
             scope='private',
+            project_id='40a908dbddfe48ad80a87fb30fa70a03',
             created_at='1970-01-01T00:00:00.000000',
             updated_at='1970-01-01T00:00:00.000000'
         )
