@@ -26,11 +26,11 @@ from mistral.api.controllers.v2 import resources
 from mistral.api.controllers.v2 import types
 from mistral import context
 from mistral.db.v2 import api as db_api
-from mistral.engine.rpc_backend import rpc
 from mistral import exceptions as exc
+from mistral.lang import parser as spec_parser
+from mistral.rpc import clients as rpc
 from mistral.utils import filter_utils
 from mistral.utils import rest_utils
-from mistral.workbook import parser as spec_parser
 from mistral.workflow import data_flow
 from mistral.workflow import states
 
@@ -41,13 +41,15 @@ STATE_TYPES = wtypes.Enum(str, states.IDLE, states.RUNNING, states.SUCCESS,
 
 
 def _get_task_resource_with_result(task_ex):
-    task = resources.Task.from_dict(task_ex.to_dict())
+    task = resources.Task.from_db_model(task_ex)
+
     task.result = json.dumps(data_flow.get_task_execution_result(task_ex))
 
     return task
 
 
 class TaskExecutionsController(rest.RestController):
+    @rest_utils.wrap_wsme_controller_exception
     @wsme_pecan.wsexpose(resources.Executions, types.uuid, types.uuid, int,
                          types.uniquelist, types.list, types.uniquelist,
                          wtypes.text, types.uuid, wtypes.text, types.jsontype,
@@ -108,7 +110,7 @@ class TaskExecutionsController(rest.RestController):
             description=description
         )
 
-        LOG.info(
+        LOG.debug(
             "Fetch executions. marker=%s, limit=%s, sort_keys=%s, "
             "sort_dirs=%s, filters=%s", marker, limit, sort_keys, sort_dirs,
             filters
@@ -128,6 +130,15 @@ class TaskExecutionsController(rest.RestController):
         )
 
 
+# Use retries to prevent possible failures.
+@rest_utils.rest_retry_on_db_error
+def _get_task_execution(id):
+    with db_api.transaction():
+        task_ex = db_api.get_task_execution(id)
+
+        return _get_task_resource_with_result(task_ex)
+
+
 class TasksController(rest.RestController):
     action_executions = action_execution.TasksActionExecutionController()
     workflow_executions = TaskExecutionsController()
@@ -135,14 +146,16 @@ class TasksController(rest.RestController):
     @rest_utils.wrap_wsme_controller_exception
     @wsme_pecan.wsexpose(resources.Task, wtypes.text)
     def get(self, id):
-        """Return the specified task."""
+        """Return the specified task.
+
+        :param id: UUID of task to retrieve
+        """
         acl.enforce('tasks:get', context.ctx())
-        LOG.info("Fetch task [id=%s]" % id)
+        LOG.debug("Fetch task [id=%s]", id)
 
-        task_ex = db_api.get_task_execution(id)
-
-        return _get_task_resource_with_result(task_ex)
+        return _get_task_execution(id)
 
+    @rest_utils.wrap_wsme_controller_exception
     @wsme_pecan.wsexpose(resources.Tasks, types.uuid, int, types.uniquelist,
                          types.list, types.uniquelist, wtypes.text,
                          wtypes.text, types.uuid, types.uuid, STATE_TYPES,
@@ -213,7 +226,7 @@ class TasksController(rest.RestController):
             env=env
         )
 
-        LOG.info(
+        LOG.debug(
             "Fetch tasks. marker=%s, limit=%s, sort_keys=%s, sort_dirs=%s,"
             " filters=%s", marker, limit, sort_keys, sort_dirs, filters
         )
@@ -241,18 +254,28 @@ class TasksController(rest.RestController):
         """
         acl.enforce('tasks:update', context.ctx())
 
-        LOG.info("Update task execution [id=%s, task=%s]" % (id, task))
+        LOG.debug("Update task execution [id=%s, task=%s]", id, task)
 
-        task_ex = db_api.get_task_execution(id)
-        task_spec = spec_parser.get_task_spec(task_ex.spec)
-        task_name = task.name or None
-        reset = task.reset
-        env = task.env or None
+        @rest_utils.rest_retry_on_db_error
+        def _read_task_params(id, task):
+            with db_api.transaction():
+                task_ex = db_api.get_task_execution(id)
+                task_spec = spec_parser.get_task_spec(task_ex.spec)
+                task_name = task.name or None
+                reset = task.reset
+                env = task.env or None
 
-        if task_name and task_name != task_ex.name:
-            raise exc.WorkflowException('Task name does not match.')
+                if task_name and task_name != task_ex.name:
+                    raise exc.WorkflowException('Task name does not match.')
+
+                wf_ex = db_api.get_workflow_execution(
+                    task_ex.workflow_execution_id
+                )
+
+                return env, reset, task_ex, task_spec, wf_ex
+
+        env, reset, task_ex, task_spec, wf_ex = _read_task_params(id, task)
 
-        wf_ex = db_api.get_workflow_execution(task_ex.workflow_execution_id)
         wf_name = task.workflow_name or None
 
         if wf_name and wf_name != wf_ex.name:
@@ -260,7 +283,8 @@ class TasksController(rest.RestController):
 
         if task.state != states.RUNNING:
             raise exc.WorkflowException(
-                'Invalid task state. Only updating task to rerun is supported.'
+                'Invalid task state. '
+                'Only updating task to rerun is supported.'
             )
 
         if task_ex.state != states.ERROR:
@@ -280,12 +304,18 @@ class TasksController(rest.RestController):
             env=env
         )
 
-        task_ex = db_api.get_task_execution(id)
+        @rest_utils.rest_retry_on_db_error
+        def _retrieve_task():
+            with db_api.transaction():
+                task_ex = db_api.get_task_execution(id)
 
-        return _get_task_resource_with_result(task_ex)
+                return _get_task_resource_with_result(task_ex)
+
+        return _retrieve_task()
 
 
 class ExecutionTasksController(rest.RestController):
+    @rest_utils.wrap_wsme_controller_exception
     @wsme_pecan.wsexpose(resources.Tasks, types.uuid, types.uuid, int,
                          types.uniquelist, types.list, types.uniquelist,
                          wtypes.text, wtypes.text, types.uuid, STATE_TYPES,
@@ -356,7 +386,7 @@ class ExecutionTasksController(rest.RestController):
             env=env
         )
 
-        LOG.info(
+        LOG.debug(
             "Fetch tasks. workflow_execution_id=%s, marker=%s, limit=%s, "
             "sort_keys=%s, sort_dirs=%s, filters=%s",
             workflow_execution_id, marker, limit, sort_keys, sort_dirs,
