@@ -27,10 +27,10 @@ from mistral.api.controllers.v2 import validation
 from mistral.api.hooks import content_type as ct_hook
 from mistral import context
 from mistral.db.v2 import api as db_api
+from mistral.lang import parser as spec_parser
 from mistral.services import workbooks
 from mistral.utils import filter_utils
 from mistral.utils import rest_utils
-from mistral.workbook import parser as spec_parser
 
 
 LOG = logging.getLogger(__name__)
@@ -43,64 +43,107 @@ class WorkbooksController(rest.RestController, hooks.HookController):
         spec_parser.get_workbook_spec_from_yaml)
 
     @rest_utils.wrap_wsme_controller_exception
-    @wsme_pecan.wsexpose(resources.Workbook, wtypes.text)
-    def get(self, name):
-        """Return the named workbook."""
+    @wsme_pecan.wsexpose(resources.Workbook, wtypes.text, wtypes.text)
+    def get(self, name, namespace=''):
+        """Return the named workbook.
+
+        :param name: Name of workbook to retrieve.
+        :param namespace: Optional. Namespace of workbook to retrieve.
+        """
         acl.enforce('workbooks:get', context.ctx())
 
-        LOG.info("Fetch workbook [name=%s]" % name)
+        LOG.debug("Fetch workbook [name=%s, namespace=%s]", name, namespace)
 
-        db_model = db_api.get_workbook(name)
+        # Use retries to prevent possible failures.
+        r = rest_utils.create_db_retry_object()
+        db_model = r.call(db_api.get_workbook,
+                          name,
+                          namespace=namespace)
 
-        return resources.Workbook.from_dict(db_model.to_dict())
+        return resources.Workbook.from_db_model(db_model)
 
     @rest_utils.wrap_pecan_controller_exception
     @pecan.expose(content_type="text/plain")
-    def put(self):
-        """Update a workbook."""
+    def put(self, namespace=''):
+        """Update a workbook.
+
+        :param namespace: Optional. Namespace of workbook to update.
+        """
+
         acl.enforce('workbooks:update', context.ctx())
 
         definition = pecan.request.text
+        scope = pecan.request.GET.get('scope', 'private')
 
-        LOG.info("Update workbook [definition=%s]" % definition)
+        resources.Workbook.validate_scope(scope)
 
-        wb_db = workbooks.update_workbook_v2(definition)
+        LOG.debug("Update workbook [definition=%s]", definition)
 
-        return resources.Workbook.from_dict(wb_db.to_dict()).to_json()
+        wb_db = rest_utils.rest_retry_on_db_error(
+            workbooks.update_workbook_v2)(
+            definition,
+            namespace=namespace,
+            scope=scope
+        )
+
+        return resources.Workbook.from_db_model(wb_db).to_json()
 
     @rest_utils.wrap_pecan_controller_exception
     @pecan.expose(content_type="text/plain")
-    def post(self):
-        """Create a new workbook."""
+    def post(self, namespace=''):
+        """Create a new workbook.
+
+        :param namespace: Optional. The namespace to create the workbook
+            in. Workbooks with the same name can be added to a given
+            project if they are in two different namespaces.
+        """
         acl.enforce('workbooks:create', context.ctx())
 
         definition = pecan.request.text
+        scope = pecan.request.GET.get('scope', 'private')
 
-        LOG.info("Create workbook [definition=%s]" % definition)
+        resources.Workbook.validate_scope(scope)
+
+        LOG.debug("Create workbook [definition=%s]", definition)
+
+        wb_db = rest_utils.rest_retry_on_db_error(
+            workbooks.create_workbook_v2)(
+            definition,
+            namespace=namespace,
+            scope=scope
+        )
 
-        wb_db = workbooks.create_workbook_v2(definition)
         pecan.response.status = 201
 
-        return resources.Workbook.from_dict(wb_db.to_dict()).to_json()
+        return resources.Workbook.from_db_model(wb_db).to_json()
 
     @rest_utils.wrap_wsme_controller_exception
-    @wsme_pecan.wsexpose(None, wtypes.text, status_code=204)
-    def delete(self, name):
-        """Delete the named workbook."""
+    @wsme_pecan.wsexpose(None, wtypes.text, wtypes.text, status_code=204)
+    def delete(self, name, namespace=''):
+        """Delete the named workbook.
+
+        :param name: Name of workbook to delete.
+        :param namespace: Optional. Namespace of workbook to delete.
+        """
         acl.enforce('workbooks:delete', context.ctx())
 
-        LOG.info("Delete workbook [name=%s]" % name)
+        LOG.debug("Delete workbook [name=%s, namespace=%s]", name, namespace)
 
-        db_api.delete_workbook(name)
+        rest_utils.rest_retry_on_db_error(db_api.delete_workbook)(
+            name,
+            namespace
+        )
 
+    @rest_utils.wrap_wsme_controller_exception
     @wsme_pecan.wsexpose(resources.Workbooks, types.uuid, int,
                          types.uniquelist, types.list, types.uniquelist,
                          wtypes.text, wtypes.text, wtypes.text,
-                         resources.SCOPE_TYPES, wtypes.text, wtypes.text)
+                         resources.SCOPE_TYPES, wtypes.text,
+                         wtypes.text, wtypes.text)
     def get_all(self, marker=None, limit=None, sort_keys='created_at',
                 sort_dirs='asc', fields='', created_at=None,
                 definition=None, name=None, scope=None, tags=None,
-                updated_at=None):
+                updated_at=None, namespace=None):
         """Return a list of workbooks.
 
         :param marker: Optional. Pagination marker for large data sets.
@@ -125,9 +168,8 @@ class WorkbooksController(rest.RestController, hooks.HookController):
                            time and date.
         :param updated_at: Optional. Keep only resources with specific latest
                            update time and date.
-
-        Where project_id is the same as the requester or
-        project_id is different but the scope is public.
+        :param namespace: Optional. Keep only resources with specific
+                          namespace.
         """
         acl.enforce('workbooks:list', context.ctx())
 
@@ -137,12 +179,13 @@ class WorkbooksController(rest.RestController, hooks.HookController):
             name=name,
             scope=scope,
             tags=tags,
-            updated_at=updated_at
+            updated_at=updated_at,
+            namespace=namespace
         )
 
-        LOG.info("Fetch workbooks. marker=%s, limit=%s, sort_keys=%s, "
-                 "sort_dirs=%s, fields=%s, filters=%s", marker, limit,
-                 sort_keys, sort_dirs, fields, filters)
+        LOG.debug("Fetch workbooks. marker=%s, limit=%s, sort_keys=%s, "
+                  "sort_dirs=%s, fields=%s, filters=%s", marker, limit,
+                  sort_keys, sort_dirs, fields, filters)
 
         return rest_utils.get_all(
             resources.Workbooks,
