@@ -31,10 +31,10 @@ from mistral.api.hooks import content_type as ct_hook
 from mistral import context
 from mistral.db.v2 import api as db_api
 from mistral import exceptions as exc
+from mistral.lang import parser as spec_parser
 from mistral.services import workflows
 from mistral.utils import filter_utils
 from mistral.utils import rest_utils
-from mistral.workbook import parser as spec_parser
 
 
 LOG = logging.getLogger(__name__)
@@ -51,7 +51,7 @@ class WorkflowsController(rest.RestController, hooks.HookController):
 
     @pecan.expose()
     def _lookup(self, identifier, sub_resource, *remainder):
-        LOG.info(
+        LOG.debug(
             "Lookup subcontrollers of WorkflowsController, "
             "sub_resource: %s, remainder: %s.",
             sub_resource,
@@ -76,50 +76,61 @@ class WorkflowsController(rest.RestController, hooks.HookController):
         )
 
     @rest_utils.wrap_wsme_controller_exception
-    @wsme_pecan.wsexpose(resources.Workflow, wtypes.text)
-    def get(self, identifier):
-        """Return the named workflow."""
+    @wsme_pecan.wsexpose(resources.Workflow, wtypes.text, wtypes.text)
+    def get(self, identifier, namespace=''):
+        """Return the named workflow.
+
+        :param identifier: Name or UUID of the workflow to retrieve.
+        :param namespace: Optional. Namespace of the workflow to retrieve.
+        """
         acl.enforce('workflows:get', context.ctx())
 
-        LOG.info("Fetch workflow [identifier=%s]" % identifier)
+        LOG.debug("Fetch workflow [identifier=%s]", identifier)
 
-        db_model = db_api.get_workflow_definition(identifier)
+        # Use retries to prevent possible failures.
+        r = rest_utils.create_db_retry_object()
+        db_model = r.call(
+            db_api.get_workflow_definition,
+            identifier,
+            namespace=namespace
+        )
 
-        return resources.Workflow.from_dict(db_model.to_dict())
+        return resources.Workflow.from_db_model(db_model)
 
     @rest_utils.wrap_pecan_controller_exception
     @pecan.expose(content_type="text/plain")
-    def put(self, identifier=None):
+    def put(self, identifier=None, namespace=''):
         """Update one or more workflows.
 
         :param identifier: Optional. If provided, it's UUID of a workflow.
             Only one workflow can be updated with identifier param.
+        :param namespace: Optional. If provided, it's the namespace of the
+                          workflow/workflows. Currently, namespace cannot be
+                          changed.
 
-        The text is allowed to have definitions of multiple workflows. In this
-        case they all will be updated.
+        The text is allowed to have definitions of multiple workflows. In such
+        case, they all will be updated.
         """
         acl.enforce('workflows:update', context.ctx())
 
         definition = pecan.request.text
         scope = pecan.request.GET.get('scope', 'private')
 
-        if scope not in resources.SCOPE_TYPES.values:
-            raise exc.InvalidModelException(
-                "Scope must be one of the following: %s; actual: "
-                "%s" % (resources.SCOPE_TYPES.values, scope)
-            )
+        resources.Workflow.validate_scope(scope)
+        if scope == 'public':
+            acl.enforce('workflows:publicize', context.ctx())
 
-        LOG.info("Update workflow(s) [definition=%s]" % definition)
+        LOG.debug("Update workflow(s) [definition=%s]", definition)
 
-        db_wfs = workflows.update_workflows(
+        db_wfs = rest_utils.rest_retry_on_db_error(workflows.update_workflows)(
             definition,
             scope=scope,
-            identifier=identifier
+            identifier=identifier,
+            namespace=namespace
         )
 
-        models_dicts = [db_wf.to_dict() for db_wf in db_wfs]
         workflow_list = [
-            resources.Workflow.from_dict(wf) for wf in models_dicts
+            resources.Workflow.from_db_model(db_wf) for db_wf in db_wfs
         ]
 
         return (workflow_list[0].to_json() if identifier
@@ -127,11 +138,15 @@ class WorkflowsController(rest.RestController, hooks.HookController):
 
     @rest_utils.wrap_pecan_controller_exception
     @pecan.expose(content_type="text/plain")
-    def post(self):
+    def post(self, namespace=''):
         """Create a new workflow.
 
-        NOTE: The text is allowed to have definitions
-            of multiple workflows. In this case they all will be created.
+        :param namespace: Optional. The namespace to create the workflow
+            in. Workflows with the same name can be added to a given
+            project if they are in two different namespaces.
+
+        The text is allowed to have definitions of multiple workflows.
+        In such case, they all will be created.
         """
         acl.enforce('workflows:create', context.ctx())
 
@@ -139,43 +154,55 @@ class WorkflowsController(rest.RestController, hooks.HookController):
         scope = pecan.request.GET.get('scope', 'private')
         pecan.response.status = 201
 
-        if scope not in resources.SCOPE_TYPES.values:
-            raise exc.InvalidModelException(
-                "Scope must be one of the following: %s; actual: "
-                "%s" % (resources.SCOPE_TYPES.values, scope)
-            )
+        resources.Workflow.validate_scope(scope)
+        if scope == 'public':
+            acl.enforce('workflows:publicize', context.ctx())
 
-        LOG.info("Create workflow(s) [definition=%s]" % definition)
+        LOG.debug("Create workflow(s) [definition=%s]", definition)
 
-        db_wfs = workflows.create_workflows(definition, scope=scope)
-        models_dicts = [db_wf.to_dict() for db_wf in db_wfs]
+        db_wfs = rest_utils.rest_retry_on_db_error(workflows.create_workflows)(
+            definition,
+            scope=scope,
+            namespace=namespace
+        )
 
         workflow_list = [
-            resources.Workflow.from_dict(wf) for wf in models_dicts
+            resources.Workflow.from_db_model(db_wf) for db_wf in db_wfs
         ]
 
         return resources.Workflows(workflows=workflow_list).to_json()
 
     @rest_utils.wrap_wsme_controller_exception
-    @wsme_pecan.wsexpose(None, wtypes.text, status_code=204)
-    def delete(self, identifier):
-        """Delete a workflow."""
-        acl.enforce('workflows:delete', context.ctx())
-        LOG.info("Delete workflow [identifier=%s]" % identifier)
+    @wsme_pecan.wsexpose(None, wtypes.text, wtypes.text, status_code=204)
+    def delete(self, identifier, namespace=''):
+        """Delete a workflow.
 
-        with db_api.transaction():
-            db_api.delete_workflow_definition(identifier)
+        :param identifier: Name or ID of workflow to delete.
+        :param namespace: Optional. Namespace of the workflow to delete.
+        """
+        acl.enforce('workflows:delete', context.ctx())
+
+        LOG.debug("Delete workflow [identifier=%s, namespace=%s]",
+                  identifier, namespace)
+
+        @rest_utils.rest_retry_on_db_error
+        def _delete_workflow_definition():
+            with db_api.transaction():
+                db_api.delete_workflow_definition(identifier, namespace)
+
+        _delete_workflow_definition()
 
     @rest_utils.wrap_wsme_controller_exception
     @wsme_pecan.wsexpose(resources.Workflows, types.uuid, int,
                          types.uniquelist, types.list, types.uniquelist,
                          wtypes.text, wtypes.text, wtypes.text, wtypes.text,
                          resources.SCOPE_TYPES, types.uuid, wtypes.text,
-                         wtypes.text)
+                         wtypes.text, bool, wtypes.text)
     def get_all(self, marker=None, limit=None, sort_keys='created_at',
                 sort_dirs='asc', fields='', name=None, input=None,
                 definition=None, tags=None, scope=None,
-                project_id=None, created_at=None, updated_at=None):
+                project_id=None, created_at=None, updated_at=None,
+                all_projects=False, namespace=None):
         """Return a list of workflows.
 
         :param marker: Optional. Pagination marker for large data sets.
@@ -192,6 +219,8 @@ class WorkflowsController(rest.RestController, hooks.HookController):
                        fields if it's provided, since it will be used when
                        constructing 'next' link.
         :param name: Optional. Keep only resources with a specific name.
+        :param namespace: Optional. Keep only resources with a specific
+                          namespace
         :param input: Optional. Keep only resources with a specific input.
         :param definition: Optional. Keep only resources with a specific
                            definition.
@@ -203,9 +232,13 @@ class WorkflowsController(rest.RestController, hooks.HookController):
                            time and date.
         :param updated_at: Optional. Keep only resources with specific latest
                            update time and date.
+        :param all_projects: Optional. Get resources of all projects.
         """
         acl.enforce('workflows:list', context.ctx())
 
+        if all_projects:
+            acl.enforce('workflows:list:all_projects', context.ctx())
+
         filters = filter_utils.create_filters_from_request_params(
             created_at=created_at,
             name=name,
@@ -214,12 +247,14 @@ class WorkflowsController(rest.RestController, hooks.HookController):
             updated_at=updated_at,
             input=input,
             definition=definition,
-            project_id=project_id
+            project_id=project_id,
+            namespace=namespace
         )
 
-        LOG.info("Fetch workflows. marker=%s, limit=%s, sort_keys=%s, "
-                 "sort_dirs=%s, fields=%s, filters=%s", marker, limit,
-                 sort_keys, sort_dirs, fields, filters)
+        LOG.debug("Fetch workflows. marker=%s, limit=%s, sort_keys=%s, "
+                  "sort_dirs=%s, fields=%s, filters=%s, all_projects=%s",
+                  marker, limit, sort_keys, sort_dirs, fields, filters,
+                  all_projects)
 
         return rest_utils.get_all(
             resources.Workflows,
@@ -231,5 +266,6 @@ class WorkflowsController(rest.RestController, hooks.HookController):
             sort_keys=sort_keys,
             sort_dirs=sort_dirs,
             fields=fields,
+            all_projects=all_projects,
             **filters
         )
