@@ -1,5 +1,6 @@
 #!/usr/bin/env python
 # Copyright 2016 - Brocade Communications Systems, Inc.
+# Copyright 2018 - Extreme Networks, Inc.
 #
 #    Licensed under the Apache License, Version 2.0 (the "License");
 #    you may not use this file except in compliance with the License.
@@ -39,114 +40,87 @@ if os.path.exists(os.path.join(POSSIBLE_TOPDIR, 'mistral', '__init__.py')):
 
 from oslo_config import cfg
 from oslo_log import log as logging
+from oslo_service import service
 
-from mistral.api import service as mistral_service
+from mistral.api import service as api_service
 from mistral import config
-from mistral.db.v2 import api as db_api
-from mistral.engine import default_engine as def_eng
-from mistral.engine import default_executor as def_executor
-from mistral.engine.rpc_backend import rpc
-from mistral.services import event_engine
-from mistral.services import expiration_policy
-from mistral.services import scheduler
-from mistral.utils import profiler
-from mistral.utils import rpc_utils
+from mistral.engine import engine_server
+from mistral.event_engine import event_engine_server
+from mistral.executors import executor_server
+from mistral.notifiers import notification_server
+from mistral.rpc import base as rpc
 from mistral import version
 
 
 CONF = cfg.CONF
+SERVER_THREAD_MANAGER = None
+SERVER_PROCESS_MANAGER = None
+
+LOG = logging.getLogger(__name__)
+
+
+def launch_thread(server, workers=1):
+    try:
+        global SERVER_THREAD_MANAGER
+
+        if not SERVER_THREAD_MANAGER:
+            SERVER_THREAD_MANAGER = service.ServiceLauncher(
+                CONF, restart_method='mutate')
+
+        SERVER_THREAD_MANAGER.launch_service(server, workers=workers)
+    except Exception as e:
+        sys.stderr.write("ERROR: %s\n" % e)
+        sys.exit(1)
+
+
+def launch_process(server, workers=1):
+    try:
+        global SERVER_PROCESS_MANAGER
+
+        if not SERVER_PROCESS_MANAGER:
+            SERVER_PROCESS_MANAGER = service.ProcessLauncher(
+                CONF, restart_method='mutate')
+
+        SERVER_PROCESS_MANAGER.launch_service(server, workers=workers)
+    except Exception as e:
+        sys.stderr.write("ERROR: %s\n" % e)
+        sys.exit(1)
 
 
 def launch_executor():
-    profiler.setup('mistral-executor', cfg.CONF.executor.host)
-
-    executor_v2 = def_executor.DefaultExecutor(rpc.get_engine_client())
-    executor_endpoint = rpc.ExecutorServer(executor_v2)
-
-    executor_server = rpc.get_rpc_server_driver()(
-        rpc_utils.get_rpc_info_from_oslo(CONF.executor)
-    )
-    executor_server.register_endpoint(executor_endpoint)
-
-    executor_v2.register_membership()
-
-    try:
-        executor_server.run(executor='threading')
-    except (KeyboardInterrupt, SystemExit):
-        pass
-    finally:
-        print("Stopping executor service...")
+    launch_thread(executor_server.get_oslo_service())
 
 
 def launch_engine():
-    profiler.setup('mistral-engine', cfg.CONF.engine.host)
-
-    engine_v2 = def_eng.DefaultEngine(rpc.get_engine_client())
-
-    engine_endpoint = rpc.EngineServer(engine_v2)
-
-    # Setup scheduler in engine.
-    db_api.setup_db()
-    scheduler.setup()
-
-    # Setup expiration policy
-    expiration_policy.setup()
-
-    engine_server = rpc.get_rpc_server_driver()(
-        rpc_utils.get_rpc_info_from_oslo(CONF.engine)
-    )
-    engine_server.register_endpoint(engine_endpoint)
-
-    engine_v2.register_membership()
-
-    try:
-        # Note(ddeja): Engine needs to be run in default (blocking) mode
-        # since using another mode may lead to deadlock.
-        # See https://review.openstack.org/#/c/356343/
-        # for more info.
-        engine_server.run()
-    except (KeyboardInterrupt, SystemExit):
-        pass
-    finally:
-        print("Stopping engine service...")
+    launch_thread(engine_server.get_oslo_service())
 
 
 def launch_event_engine():
-    profiler.setup('mistral-event-engine', cfg.CONF.event_engine.host)
+    launch_thread(event_engine_server.get_oslo_service())
 
-    event_eng = event_engine.EventEngine(rpc.get_engine_client())
-    endpoint = rpc.EventEngineServer(event_eng)
 
-    event_engine_server = rpc.get_rpc_server_driver()(
-        rpc_utils.get_rpc_info_from_oslo(CONF.event_engine)
-    )
-    event_engine_server.register_endpoint(endpoint)
-
-    event_eng.register_membership()
-
-    try:
-        event_engine_server.run()
-    except (KeyboardInterrupt, SystemExit):
-        pass
-    finally:
-        print("Stopping event_engine service...")
+def launch_notifier():
+    launch_thread(notification_server.get_oslo_service())
 
 
 def launch_api():
-    launcher = mistral_service.process_launcher()
-    server = mistral_service.WSGIService('mistral_api')
-    launcher.launch_service(server, workers=server.workers)
-    launcher.wait()
+    server = api_service.WSGIService('mistral_api')
+
+    launch_process(server, workers=server.workers)
 
 
 def launch_any(options):
-    # Launch the servers on different threads.
-    threads = [eventlet.spawn(LAUNCH_OPTIONS[option])
-               for option in options]
+    for option in options:
+        LAUNCH_OPTIONS[option]()
 
-    print('Server started.')
+    global SERVER_PROCESS_MANAGER
+    global SERVER_THREAD_MANAGER
 
-    [thread.wait() for thread in threads]
+    if SERVER_PROCESS_MANAGER:
+        SERVER_PROCESS_MANAGER.wait()
+
+    if SERVER_THREAD_MANAGER:
+        SERVER_THREAD_MANAGER.wait()
 
 
 # Map cli options to appropriate functions. The cli options are
@@ -155,17 +129,18 @@ LAUNCH_OPTIONS = {
     'api': launch_api,
     'engine': launch_engine,
     'executor': launch_executor,
-    'event-engine': launch_event_engine
+    'event-engine': launch_event_engine,
+    'notifier': launch_notifier
 }
 
 
 MISTRAL_TITLE = """
-|\\\    //| //   // |||||| |||\\\       /\      ||
-||\\\  //||    //     ||   ||  ||     //\\\     ||
-|| \\\// || || ||     ||   || //     //  \\\    ||
-||  \/  || ||  \\\    ||   || \\\    //-||-\\\   ||
-||      || ||   ||   ||   ||  ||  //      \\\  ||
-||      || || _//    ||   ||  || //        \\\ |||||
+|\\\    //|           ||                       ||
+||\\\  //||      __   ||      __      __       ||
+|| \\\// || ||  //  ||||||  ||  \\\  //  \\\     ||
+||  \\/  ||     \\\    ||    ||     ||    \\\    ||
+||      || ||   \\\   ||    ||     ||    /\\\   ||
+||      || || __//   ||_// ||      \\\__// \\\_ ||
 
 Mistral Workflow Service, version %s
 """ % version.version_string()
@@ -174,8 +149,8 @@ Mistral Workflow Service, version %s
 def print_server_info():
     print(MISTRAL_TITLE)
 
-    comp_str = ("[%s]" % ','.join(LAUNCH_OPTIONS)
-                if cfg.CONF.server == ['all'] else cfg.CONF.server)
+    comp_str = ("[%s]" % ','.join(LAUNCH_OPTIONS
+                if cfg.CONF.server == ['all'] else cfg.CONF.server))
 
     print('Launching server components %s...' % comp_str)
 
@@ -205,12 +180,28 @@ def get_properly_ordered_parameters():
     return args
 
 
+def override_keystone_options():
+    # TODO(wxy): This function is used for keeping backward compatibility.
+    # Remove it in Stein.
+    auth_opts = CONF['keystone_authtoken']
+    for opt, value in auth_opts.items():
+        if opt in CONF['keystone']:
+            default_value = auth_opts._group._opts[opt]['opt'].default
+            if default_value != value != CONF['keystone'][opt]:
+                LOG.warning("The config option '%s' in section "
+                            "[keystone_authtoken] has the same copy in "
+                            "[keystone]. Please add the same option to the "
+                            "[keystone] section to keep using it.", opt)
+                CONF.set_override(opt, value, group='keystone')
+
+
 def main():
     try:
         config.parse_args(get_properly_ordered_parameters())
         print_server_info()
 
         logging.setup(CONF, 'Mistral')
+        override_keystone_options()
 
         # Please refer to the oslo.messaging documentation for transport
         # configuration. The default transport for oslo.messaging is
@@ -236,7 +227,7 @@ def main():
             # Validate launch option.
             if set(cfg.CONF.server) - set(LAUNCH_OPTIONS.keys()):
                 raise Exception('Valid options are all or any combination of '
-                                'api, engine, and executor.')
+                                ', '.join(LAUNCH_OPTIONS.keys()))
 
             # Launch distinct set of server(s).
             launch_any(set(cfg.CONF.server))
@@ -246,5 +237,25 @@ def main():
         sys.exit(1)
 
 
+# Helper method used in unit tests to reset the service launchers.
+def reset_server_managers():
+    global SERVER_THREAD_MANAGER
+    global SERVER_PROCESS_MANAGER
+    SERVER_THREAD_MANAGER = None
+    SERVER_PROCESS_MANAGER = None
+
+
+# Helper method used in unit tests to access the service launcher.
+def get_server_thread_manager():
+    global SERVER_THREAD_MANAGER
+    return SERVER_THREAD_MANAGER
+
+
+# Helper method used in unit tests to access the process launcher.
+def get_server_process_manager():
+    global SERVER_PROCESS_MANAGER
+    return SERVER_PROCESS_MANAGER
+
+
 if __name__ == '__main__':
-    main()
+    sys.exit(main())
