@@ -53,11 +53,15 @@ def end_tx():
 
 
 @contextlib.contextmanager
-def transaction():
-    with IMPL.transaction():
+def transaction(read_only=False):
+    with IMPL.transaction(read_only):
         yield
 
 
+def refresh(model):
+    IMPL.refresh(model)
+
+
 # Locking.
 
 
@@ -67,13 +71,13 @@ def acquire_lock(model, id):
 
 # Workbooks.
 
-def get_workbook(name):
-    return IMPL.get_workbook(name)
+def get_workbook(name, namespace, fields=()):
+    return IMPL.get_workbook(name, namespace=namespace, fields=fields)
 
 
-def load_workbook(name):
+def load_workbook(name, namespace, fields=()):
     """Unlike get_workbook this method is allowed to return None."""
-    return IMPL.load_workbook(name)
+    return IMPL.load_workbook(name, namespace=namespace, fields=fields)
 
 
 def get_workbooks(limit=None, marker=None, sort_keys=None,
@@ -100,8 +104,8 @@ def create_or_update_workbook(name, values):
     return IMPL.create_or_update_workbook(name, values)
 
 
-def delete_workbook(name):
-    IMPL.delete_workbook(name)
+def delete_workbook(name, namespace=None):
+    IMPL.delete_workbook(name, namespace)
 
 
 def delete_workbooks(**kwargs):
@@ -110,17 +114,21 @@ def delete_workbooks(**kwargs):
 
 # Workflow definitions.
 
-def get_workflow_definition(identifier):
-    return IMPL.get_workflow_definition(identifier)
+def get_workflow_definition(identifier, namespace='', fields=()):
+    return IMPL.get_workflow_definition(
+        identifier,
+        namespace=namespace,
+        fields=fields
+    )
 
 
-def get_workflow_definition_by_id(id):
-    return IMPL.get_workflow_definition_by_id(id)
+def get_workflow_definition_by_id(id, fields=()):
+    return IMPL.get_workflow_definition_by_id(id, fields=fields)
 
 
-def load_workflow_definition(name):
+def load_workflow_definition(name, namespace='', fields=()):
     """Unlike get_workflow_definition this method is allowed to return None."""
-    return IMPL.load_workflow_definition(name)
+    return IMPL.load_workflow_definition(name, namespace, fields=fields)
 
 
 def get_workflow_definitions(limit=None, marker=None, sort_keys=None,
@@ -147,8 +155,8 @@ def create_or_update_workflow_definition(name, values):
     return IMPL.create_or_update_workflow_definition(name, values)
 
 
-def delete_workflow_definition(identifier):
-    IMPL.delete_workflow_definition(identifier)
+def delete_workflow_definition(identifier, namespace=''):
+    IMPL.delete_workflow_definition(identifier, namespace)
 
 
 def delete_workflow_definitions(**kwargs):
@@ -157,20 +165,20 @@ def delete_workflow_definitions(**kwargs):
 
 # Action definitions.
 
-def get_action_definition_by_id(id):
-    return IMPL.get_action_definition_by_id(id)
+def get_action_definition_by_id(id, fields=()):
+    return IMPL.get_action_definition_by_id(id, fields=fields)
 
 
-def get_action_definition(name):
-    return IMPL.get_action_definition(name)
+def get_action_definition(name, fields=()):
+    return IMPL.get_action_definition(name, fields=fields)
 
 
-def load_action_definition(name):
+def load_action_definition(name, fields=()):
     """Unlike get_action_definition this method is allowed to return None."""
-    return IMPL.load_action_definition(name)
+    return IMPL.load_action_definition(name, fields=fields)
 
 
-def get_action_definitions(limit=None, marker=None, sort_keys=['name'],
+def get_action_definitions(limit=None, marker=None, sort_keys=None,
                            sort_dirs=None, **kwargs):
     return IMPL.get_action_definitions(
         limit=limit,
@@ -203,29 +211,25 @@ def delete_action_definitions(**kwargs):
 
 # Action executions.
 
-def get_action_execution(id):
-    return IMPL.get_action_execution(id)
+def get_action_execution(id, fields=(), insecure=False):
+    return IMPL.get_action_execution(id, fields=fields, insecure=insecure)
 
 
-def load_action_execution(name):
+def load_action_execution(name, fields=()):
     """Unlike get_action_execution this method is allowed to return None."""
-    return IMPL.load_action_execution(name)
+    return IMPL.load_action_execution(name, fields=fields)
 
 
 def get_action_executions(**kwargs):
     return IMPL.get_action_executions(**kwargs)
 
 
-def ensure_action_execution_exists(id):
-    return IMPL.ensure_action_execution_exists(id)
-
-
 def create_action_execution(values):
     return IMPL.create_action_execution(values)
 
 
-def update_action_execution(id, values):
-    return IMPL.update_action_execution(id, values)
+def update_action_execution(id, values, insecure=False):
+    return IMPL.update_action_execution(id, values, insecure)
 
 
 def create_or_update_action_execution(id, values):
@@ -242,16 +246,16 @@ def delete_action_executions(**kwargs):
 
 # Workflow executions.
 
-def get_workflow_execution(id):
-    return IMPL.get_workflow_execution(id)
+def get_workflow_execution(id, fields=()):
+    return IMPL.get_workflow_execution(id, fields=fields)
 
 
-def load_workflow_execution(name):
+def load_workflow_execution(name, fields=()):
     """Unlike get_workflow_execution this method is allowed to return None."""
-    return IMPL.load_workflow_execution(name)
+    return IMPL.load_workflow_execution(name, fields=fields)
 
 
-def get_workflow_executions(limit=None, marker=None, sort_keys=['created_at'],
+def get_workflow_executions(limit=None, marker=None, sort_keys=None,
                             sort_dirs=None, **kwargs):
     return IMPL.get_workflow_executions(
         limit=limit,
@@ -262,10 +266,6 @@ def get_workflow_executions(limit=None, marker=None, sort_keys=['created_at'],
     )
 
 
-def ensure_workflow_execution_exists(id):
-    return IMPL.ensure_workflow_execution_exists(id)
-
-
 def create_workflow_execution(values):
     return IMPL.create_workflow_execution(values)
 
@@ -286,18 +286,22 @@ def delete_workflow_executions(**kwargs):
     IMPL.delete_workflow_executions(**kwargs)
 
 
+def update_workflow_execution_state(**kwargs):
+    return IMPL.update_workflow_execution_state(**kwargs)
+
+
 # Tasks executions.
 
-def get_task_execution(id):
-    return IMPL.get_task_execution(id)
+def get_task_execution(id, fields=()):
+    return IMPL.get_task_execution(id, fields=fields)
 
 
-def load_task_execution(id):
+def load_task_execution(id, fields=()):
     """Unlike get_task_execution this method is allowed to return None."""
-    return IMPL.load_task_execution(id)
+    return IMPL.load_task_execution(id, fields=fields)
 
 
-def get_task_executions(limit=None, marker=None, sort_keys=['created_at'],
+def get_task_executions(limit=None, marker=None, sort_keys=None,
                         sort_dirs=None, **kwargs):
     return IMPL.get_task_executions(
         limit=limit,
@@ -308,6 +312,14 @@ def get_task_executions(limit=None, marker=None, sort_keys=['created_at'],
     )
 
 
+def get_completed_task_executions(**kwargs):
+    return IMPL.get_completed_task_executions(**kwargs)
+
+
+def get_completed_task_executions_as_batches(**kwargs):
+    return IMPL.get_completed_task_executions_as_batches(**kwargs)
+
+
 def get_incomplete_task_executions(**kwargs):
     return IMPL.get_incomplete_task_executions(**kwargs)
 
@@ -336,10 +348,14 @@ def delete_task_executions(**kwargs):
     return IMPL.delete_task_executions(**kwargs)
 
 
+def update_task_execution_state(**kwargs):
+    return IMPL.update_task_execution_state(**kwargs)
+
+
 # Delayed calls.
 
-def get_delayed_calls_to_start(time):
-    return IMPL.get_delayed_calls_to_start(time)
+def get_delayed_calls_to_start(time, batch_size=None):
+    return IMPL.get_delayed_calls_to_start(time, batch_size)
 
 
 def create_delayed_call(values):
@@ -368,13 +384,17 @@ def delete_delayed_calls(**kwargs):
 
 # Cron triggers.
 
-def get_cron_trigger(name):
-    return IMPL.get_cron_trigger(name)
+def get_cron_trigger(identifier):
+    return IMPL.get_cron_trigger(identifier)
 
 
-def load_cron_trigger(name):
+def get_cron_trigger_by_id(id):
+    return IMPL.get_cron_trigger_by_id(id)
+
+
+def load_cron_trigger(identifier):
     """Unlike get_cron_trigger this method is allowed to return None."""
-    return IMPL.load_cron_trigger(name)
+    return IMPL.load_cron_trigger(identifier)
 
 
 def get_cron_triggers(**kwargs):
@@ -385,24 +405,42 @@ def get_next_cron_triggers(time):
     return IMPL.get_next_cron_triggers(time)
 
 
-def get_expired_executions(time):
-    return IMPL.get_expired_executions(time)
+def get_expired_executions(expiration_time, limit=None, columns=()):
+    return IMPL.get_expired_executions(
+        expiration_time,
+        limit,
+        columns
+    )
+
+
+def get_running_expired_sync_actions(expiration_time, session=None):
+    return IMPL.get_running_expired_sync_actions(expiration_time)
+
+
+def get_superfluous_executions(max_finished_executions, limit=None,
+                               columns=()):
+    return IMPL.get_superfluous_executions(
+        max_finished_executions,
+        limit,
+        columns
+    )
 
 
 def create_cron_trigger(values):
     return IMPL.create_cron_trigger(values)
 
 
-def update_cron_trigger(name, values, query_filter=None):
-    return IMPL.update_cron_trigger(name, values, query_filter=query_filter)
+def update_cron_trigger(identifier, values, query_filter=None):
+    return IMPL.update_cron_trigger(identifier, values,
+                                    query_filter=query_filter)
 
 
-def create_or_update_cron_trigger(name, values):
-    return IMPL.create_or_update_cron_trigger(name, values)
+def create_or_update_cron_trigger(identifier, values):
+    return IMPL.create_or_update_cron_trigger(identifier, values)
 
 
-def delete_cron_trigger(name):
-    return IMPL.delete_cron_trigger(name)
+def delete_cron_trigger(identifier):
+    return IMPL.delete_cron_trigger(identifier)
 
 
 def delete_cron_triggers(**kwargs):
@@ -420,7 +458,7 @@ def load_environment(name):
     return IMPL.load_environment(name)
 
 
-def get_environments(limit=None, marker=None, sort_keys=['name'],
+def get_environments(limit=None, marker=None, sort_keys=None,
                      sort_dirs=None, **kwargs):
 
     return IMPL.get_environments(
@@ -490,10 +528,14 @@ def get_event_trigger(id, insecure=False):
     return IMPL.get_event_trigger(id, insecure)
 
 
+def load_event_trigger(id, insecure=False):
+    return IMPL.load_event_trigger(id, insecure)
+
+
 def get_event_triggers(insecure=False, limit=None, marker=None, sort_keys=None,
                        sort_dirs=None, fields=None, **kwargs):
     return IMPL.get_event_triggers(
-        insecure=False,
+        insecure=insecure,
         limit=limit,
         marker=marker,
         sort_keys=sort_keys,
@@ -519,10 +561,6 @@ def delete_event_triggers(**kwargs):
     return IMPL.delete_event_triggers(**kwargs)
 
 
-def ensure_event_trigger_exists(id):
-    return IMPL.ensure_event_trigger_exists(id)
-
-
 # Locks.
 
 def create_named_lock(name):
