@@ -13,16 +13,17 @@
 #    See the License for the specific language governing permissions and
 #    limitations under the License.
 
+import datetime
 import hashlib
 import json
-import sqlalchemy as sa
-from sqlalchemy import event
-from sqlalchemy.orm import backref
-from sqlalchemy.orm import relationship
 import sys
 
 from oslo_config import cfg
 from oslo_log import log as logging
+import sqlalchemy as sa
+from sqlalchemy import event
+from sqlalchemy.orm import backref
+from sqlalchemy.orm import relationship
 
 from mistral.db.sqlalchemy import model_base as mb
 from mistral.db.sqlalchemy import types as st
@@ -33,6 +34,7 @@ from mistral import utils
 
 # Definition objects.
 
+CONF = cfg.CONF
 LOG = logging.getLogger(__name__)
 
 
@@ -65,18 +67,20 @@ def validate_long_type_length(cls, field_name, value):
         size_kb = int(sys.getsizeof(str(value)) / 1024)
 
         if size_kb > size_limit_kb:
-            LOG.error(
-                "Size limit %dKB exceed for class [%s], "
-                "field %s of size %dKB.",
-                size_limit_kb, str(cls), field_name, size_kb
-            )
-
-            raise exc.SizeLimitExceededException(
+            msg = (
+                "Field size limit exceeded"
+                " [class={}, field={}, size={}KB, limit={}KB]"
+            ).format(
+                cls.__name__,
                 field_name,
                 size_kb,
                 size_limit_kb
             )
 
+            LOG.error(msg)
+
+            raise exc.SizeLimitExceededException(msg)
+
 
 def register_length_validator(attr_name):
     """Register an event listener on the attribute.
@@ -109,9 +113,14 @@ class Workbook(Definition):
     """Contains info about workbook (including definition in Mistral DSL)."""
 
     __tablename__ = 'workbooks_v2'
+    namespace = sa.Column(sa.String(255), nullable=True)
 
     __table_args__ = (
-        sa.UniqueConstraint('name', 'project_id'),
+        sa.UniqueConstraint(
+            'name',
+            'namespace',
+            'project_id'
+        ),
         sa.Index('%s_project_id' % __tablename__, 'project_id'),
         sa.Index('%s_scope' % __tablename__, 'scope'),
     )
@@ -121,9 +130,13 @@ class WorkflowDefinition(Definition):
     """Contains info about workflow (including definition in Mistral DSL)."""
 
     __tablename__ = 'workflow_definitions_v2'
-
+    namespace = sa.Column(sa.String(255), nullable=True)
     __table_args__ = (
-        sa.UniqueConstraint('name', 'project_id'),
+        sa.UniqueConstraint(
+            'name',
+            'namespace',
+            'project_id'
+        ),
         sa.Index('%s_is_system' % __tablename__, 'is_system'),
         sa.Index('%s_project_id' % __tablename__, 'project_id'),
         sa.Index('%s_scope' % __tablename__, 'scope'),
@@ -162,6 +175,7 @@ class Execution(mb.MistralSecureModelBase):
     name = sa.Column(sa.String(255))
     description = sa.Column(sa.String(255), nullable=True)
     workflow_name = sa.Column(sa.String(255))
+    workflow_namespace = sa.Column(sa.String(255))
     workflow_id = sa.Column(sa.String(80))
     spec = sa.Column(st.JsonMediumDictType())
     state = sa.Column(sa.String(20))
@@ -188,6 +202,13 @@ class ActionExecution(Execution):
     accepted = sa.Column(sa.Boolean(), default=False)
     input = sa.Column(st.JsonLongDictType(), nullable=True)
     output = sa.orm.deferred(sa.Column(st.JsonLongDictType(), nullable=True))
+    last_heartbeat = sa.Column(
+        sa.DateTime,
+        default=lambda: utils.utc_now_sec() + datetime.timedelta(
+            seconds=CONF.action_heartbeat.first_heartbeat_timeout
+        )
+    )
+    is_sync = sa.Column(sa.Boolean(), default=None, nullable=True)
 
 
 class WorkflowExecution(Execution):
@@ -232,7 +253,8 @@ class TaskExecution(Execution):
 
     # Main properties.
     action_spec = sa.Column(st.JsonLongDictType())
-    unique_key = sa.Column(sa.String(250), nullable=True)
+    unique_key = sa.Column(sa.String(255), nullable=True)
+    type = sa.Column(sa.String(10))
 
     # Whether the task is fully processed (publishing and calculating commands
     # after it). It allows to simplify workflow controller implementations
@@ -255,9 +277,16 @@ class TaskExecution(Execution):
 for cls in utils.iter_subclasses(Execution):
     event.listen(
         # Catch and trim Execution.state_info to always fit allocated size.
+        # Note that the limit is 65500 which is less than 65535 (2^16 -1).
+        # The reason is that utils.cut() is not exactly accurate in case if
+        # the value is not a string, but, for example, a dictionary. If we
+        # limit it exactly to 65535 then once in a while it may go slightly
+        # beyond the allowed maximum size. It may depend on the order of
+        # keys in a string representation and other things that are hidden
+        # inside utils.cut_dict() method.
         cls.state_info,
         'set',
-        lambda t, v, o, i: utils.cut(v, 65532),
+        lambda t, v, o, i: utils.cut(v, 65500),
         retval=True
     )
 
@@ -275,7 +304,8 @@ TaskExecution.action_executions = relationship(
     backref=backref('task_execution', remote_side=[TaskExecution.id]),
     cascade='all, delete-orphan',
     foreign_keys=ActionExecution.task_execution_id,
-    lazy='select'
+    lazy='select',
+    passive_deletes=True
 )
 
 sa.Index(
@@ -296,7 +326,8 @@ TaskExecution.workflow_executions = relationship(
     backref=backref('task_execution', remote_side=[TaskExecution.id]),
     cascade='all, delete-orphan',
     foreign_keys=WorkflowExecution.task_execution_id,
-    lazy='select'
+    lazy='select',
+    passive_deletes=True
 )
 
 sa.Index(
@@ -304,6 +335,14 @@ sa.Index(
     'task_execution_id'
 )
 
+# Many-to-one for 'WorkflowExecution' and 'WorkflowExecution'
+
+WorkflowExecution.root_execution_id = sa.Column(
+    sa.String(36),
+    sa.ForeignKey(WorkflowExecution.id, ondelete='SET NULL'),
+    nullable=True
+)
+
 # Many-to-one for 'TaskExecution' and 'WorkflowExecution'.
 
 TaskExecution.workflow_execution_id = sa.Column(
@@ -316,7 +355,8 @@ WorkflowExecution.task_executions = relationship(
     backref=backref('workflow_execution', remote_side=[WorkflowExecution.id]),
     cascade='all, delete-orphan',
     foreign_keys=TaskExecution.workflow_execution_id,
-    lazy='select'
+    lazy='select',
+    passive_deletes=True
 )
 
 sa.Index(
@@ -324,6 +364,11 @@ sa.Index(
     TaskExecution.workflow_execution_id
 )
 
+sa.Index(
+    '%s_workflow_execution_id_name' % TaskExecution.__tablename__,
+    TaskExecution.workflow_execution_id, TaskExecution.name
+)
+
 
 # Other objects.
 
@@ -366,7 +411,7 @@ class Environment(mb.MistralSecureModelBase):
     id = mb.id_column()
     name = sa.Column(sa.String(200))
     description = sa.Column(sa.Text())
-    variables = sa.Column(st.JsonDictType())
+    variables = sa.Column(st.JsonLongDictType())
 
 
 class CronTrigger(mb.MistralSecureModelBase):
@@ -424,8 +469,8 @@ class CronTrigger(mb.MistralSecureModelBase):
     def to_dict(self):
         d = super(CronTrigger, self).to_dict()
 
-        mb.datetime_to_str(d, 'first_execution_time')
-        mb.datetime_to_str(d, 'next_execution_time')
+        utils.datetime_to_str_in_dict(d, 'first_execution_time')
+        utils.datetime_to_str_in_dict(d, 'next_execution_time')
 
         return d
 
@@ -484,6 +529,7 @@ class EventTrigger(mb.MistralSecureModelBase):
         sa.String(36),
         sa.ForeignKey(WorkflowDefinition.id)
     )
+    workflow = relationship('WorkflowDefinition', lazy='joined')
     workflow_params = sa.Column(st.JsonDictType())
     workflow_input = sa.Column(st.JsonDictType())
 
@@ -523,7 +569,7 @@ class NamedLock(mb.MistralModelBase):
     sa.UniqueConstraint('name')
 
     id = mb.id_column()
-    name = sa.Column(sa.String(250))
+    name = sa.Column(sa.String(255))
 
 
 sa.UniqueConstraint(NamedLock.name)
