@@ -1,5 +1,6 @@
 # Copyright 2016 - Nokia Networks.
 # Copyright 2016 - Brocade Communications Systems, Inc.
+# Copyright 2018 - Extreme Networks, Inc.
 #
 #    Licensed under the Apache License, Version 2.0 (the "License");
 #    you may not use this file except in compliance with the License.
@@ -15,26 +16,29 @@
 
 import abc
 from oslo_config import cfg
+from oslo_log import log as logging
 from osprofiler import profiler
 import six
 
 from mistral.db.v2 import api as db_api
-from mistral.engine.rpc_backend import rpc
-from mistral.engine import utils as e_utils
+from mistral.engine import action_queue
+from mistral.engine import utils as engine_utils
 from mistral.engine import workflow_handler as wf_handler
 from mistral import exceptions as exc
+from mistral.executors import base as exe
 from mistral import expressions as expr
+from mistral.lang import parser as spec_parser
 from mistral.services import action_manager as a_m
-from mistral.services import scheduler
 from mistral.services import security
 from mistral import utils
 from mistral.utils import wf_trace
-from mistral.workbook import parser as spec_parser
+from mistral.workflow import data_flow
+from mistral.workflow import lookup_utils
 from mistral.workflow import states
-from mistral.workflow import utils as wf_utils
+from mistral_lib import actions as ml_actions
 
 
-_RUN_EXISTING_ACTION_PATH = 'mistral.engine.actions._run_existing_action'
+LOG = logging.getLogger(__name__)
 
 
 @six.add_metaclass(abc.ABCMeta)
@@ -71,8 +75,26 @@ class Action(object):
         self.action_ex.state = states.ERROR
         self.action_ex.output = {'result': msg}
 
+    def update(self, state):
+        assert self.action_ex
+
+        if state == states.PAUSED and self.is_sync(self.action_ex.input):
+            raise exc.InvalidStateTransitionException(
+                'Transition to the PAUSED state is only supported '
+                'for asynchronous action execution.'
+            )
+
+        if not states.is_valid_transition(self.action_ex.state, state):
+            raise exc.InvalidStateTransitionException(
+                'Invalid state transition from %s to %s.' %
+                (self.action_ex.state, state)
+            )
+
+        self.action_ex.state = state
+
     @abc.abstractmethod
-    def schedule(self, input_dict, target, index=0, desc='', safe_rerun=False):
+    def schedule(self, input_dict, target, index=0, desc='', safe_rerun=False,
+                 timeout=None):
         """Schedule action run.
 
         This method is needed to schedule action run so its result can
@@ -81,6 +103,8 @@ class Action(object):
         executor asynchrony when executor doesn't immediately send a
         result).
 
+        :param timeout: a period of time in seconds after which execution of
+            action will be interrupted
         :param input_dict: Action input.
         :param target: Target (group of action executors).
         :param index: Action execution index. Makes sense for some types.
@@ -92,13 +116,15 @@ class Action(object):
 
     @abc.abstractmethod
     def run(self, input_dict, target, index=0, desc='', save=True,
-            safe_rerun=False):
+            safe_rerun=False, timeout=None):
         """Immediately run action.
 
         This method runs method w/o scheduling its run for a later time.
         From engine perspective action will be processed in synchronous
         mode.
 
+        :param timeout: a period of time in seconds after which execution of
+            action will be interrupted
         :param input_dict: Action input.
         :param target: Target (group of action executors).
         :param index: Action execution index. Makes sense for some types.
@@ -124,7 +150,7 @@ class Action(object):
         """
         return True
 
-    def _create_action_execution(self, input_dict, runtime_ctx,
+    def _create_action_execution(self, input_dict, runtime_ctx, is_sync,
                                  desc='', action_ex_id=None):
         action_ex_id = action_ex_id or utils.generate_unicode_uuid()
 
@@ -135,13 +161,15 @@ class Action(object):
             'state': states.RUNNING,
             'input': input_dict,
             'runtime_context': runtime_ctx,
-            'description': desc
+            'description': desc,
+            'is_sync': is_sync
         }
 
         if self.task_ex:
             values.update({
                 'task_execution_id': self.task_ex.id,
                 'workflow_name': self.task_ex.workflow_name,
+                'workflow_namespace': self.task_ex.workflow_namespace,
                 'workflow_id': self.task_ex.workflow_id,
                 'project_id': self.task_ex.project_id,
             })
@@ -157,11 +185,7 @@ class Action(object):
             # state within the current session.
             self.task_ex.action_executions.append(self.action_ex)
 
-    def _inject_action_ctx_for_validating(self, input_dict):
-        if a_m.has_action_context(
-                self.action_def.action_class, self.action_def.attributes):
-            input_dict.update(a_m.get_empty_action_context())
-
+    @profiler.trace('action-log-result', hide_args=True)
     def _log_result(self, prev_state, result):
         state = self.action_ex.state
 
@@ -187,12 +211,13 @@ class Action(object):
 class PythonAction(Action):
     """Regular Python action."""
 
-    @profiler.trace('action-complete')
+    @profiler.trace('action-complete', hide_args=True)
     def complete(self, result):
         assert self.action_ex
 
         if states.is_completed(self.action_ex.state):
-            return
+            raise ValueError(
+                "Action {} is already completed".format(self.action_ex.id))
 
         prev_state = self.action_ex.state
 
@@ -208,8 +233,9 @@ class PythonAction(Action):
 
         self._log_result(prev_state, result)
 
-    @profiler.trace('action-schedule')
-    def schedule(self, input_dict, target, index=0, desc='', safe_rerun=False):
+    @profiler.trace('action-schedule', hide_args=True)
+    def schedule(self, input_dict, target, index=0, desc='', safe_rerun=False,
+                 timeout=None):
         assert not self.action_ex
 
         # Assign the action execution ID here to minimize database calls.
@@ -218,26 +244,27 @@ class PythonAction(Action):
         # DB object is created.
         action_ex_id = utils.generate_unicode_uuid()
 
-        self._insert_action_context(action_ex_id, input_dict)
-
         self._create_action_execution(
             self._prepare_input(input_dict),
             self._prepare_runtime_context(index, safe_rerun),
+            self.is_sync(input_dict),
             desc=desc,
             action_ex_id=action_ex_id
         )
 
-        scheduler.schedule_call(
-            None,
-            _RUN_EXISTING_ACTION_PATH,
-            0,
-            action_ex_id=self.action_ex.id,
-            target=target
+        execution_context = self._prepare_execution_context()
+
+        action_queue.schedule_run_action(
+            self.action_ex,
+            self.action_def,
+            target,
+            execution_context,
+            timeout=timeout
         )
 
-    @profiler.trace('action-run')
+    @profiler.trace('action-run', hide_args=True)
     def run(self, input_dict, target, index=0, desc='', save=True,
-            safe_rerun=False):
+            safe_rerun=False, timeout=None):
         assert not self.action_ex
 
         input_dict = self._prepare_input(input_dict)
@@ -249,24 +276,29 @@ class PythonAction(Action):
         # DB object is created.
         action_ex_id = utils.generate_unicode_uuid()
 
-        self._insert_action_context(action_ex_id, input_dict, save=save)
-
         if save:
             self._create_action_execution(
                 input_dict,
                 runtime_ctx,
+                self.is_sync(input_dict),
                 desc=desc,
                 action_ex_id=action_ex_id
             )
 
-        result = rpc.get_executor_client().run_action(
+        executor = exe.get_executor(cfg.CONF.executor.type)
+
+        execution_context = self._prepare_execution_context()
+
+        result = executor.run_action(
             self.action_ex.id if self.action_ex else None,
             self.action_def.action_class,
             self.action_def.attributes or {},
             input_dict,
-            target,
-            async=False,
-            safe_rerun=safe_rerun
+            safe_rerun,
+            execution_context,
+            target=target,
+            async_=False,
+            timeout=timeout
         )
 
         return self._prepare_output(result)
@@ -279,13 +311,37 @@ class PythonAction(Action):
         return a.is_sync()
 
     def validate_input(self, input_dict):
-        if self.action_def.action_class:
-            self._inject_action_ctx_for_validating(input_dict)
 
-        # NOTE(xylan): Don't validate action input if action initialization
+        # NOTE(kong): Don't validate action input if action initialization
         # method contains ** argument.
-        if '**' not in self.action_def.input:
-            e_utils.validate_input(self.action_def, input_dict)
+        if '**' in self.action_def.input:
+            return
+
+        expected_input = utils.get_dict_from_string(self.action_def.input)
+
+        engine_utils.validate_input(
+            expected_input,
+            input_dict,
+            self.action_def.name,
+            self.action_def.action_class
+        )
+
+    def _prepare_execution_context(self):
+
+        exc_ctx = {}
+
+        if self.task_ex:
+            wf_ex = self.task_ex.workflow_execution
+            exc_ctx['workflow_execution_id'] = wf_ex.id
+            exc_ctx['task_execution_id'] = self.task_ex.id
+            exc_ctx['workflow_name'] = wf_ex.name
+
+        if self.action_ex:
+            exc_ctx['action_execution_id'] = self.action_ex.id
+            callback_url = '/v2/action_executions/%s' % self.action_ex.id
+            exc_ctx['callback_url'] = callback_url
+
+        return exc_ctx
 
     def _prepare_input(self, input_dict):
         """Template method to do manipulations with input parameters.
@@ -309,36 +365,27 @@ class PythonAction(Action):
         """
         return {'index': index, 'safe_rerun': safe_rerun}
 
-    def _insert_action_context(self, action_ex_id, input_dict, save=True):
-        """Template method to prepare action context.
-
-        It inserts the action context in the input if required
-        runtime context.
-        """
-        # we need to push action context to all actions. It's related to
-        # https://blueprints.launchpad.net/mistral/+spec/mistral-custom-actions-api
-        has_action_context = a_m.has_action_context(
-            self.action_def.action_class,
-            self.action_def.attributes or {}
-        )
-
-        if has_action_context:
-            input_dict.update(
-                a_m.get_action_context(self.task_ex, action_ex_id, save=save)
-            )
-
 
 class AdHocAction(PythonAction):
     """Ad-hoc action."""
 
-    def __init__(self, action_def, action_ex=None, task_ex=None):
+    def __init__(self, action_def, action_ex=None, task_ex=None, task_ctx=None,
+                 wf_ctx=None):
         self.action_spec = spec_parser.get_action_spec(action_def.spec)
 
-        base_action_def = db_api.get_action_definition(
+        base_action_def = lookup_utils.find_action_definition_by_name(
             self.action_spec.get_base()
         )
+
+        if not base_action_def:
+            raise exc.InvalidActionException(
+                "Failed to find action [action_name=%s]" %
+                self.action_spec.get_base()
+            )
+
         base_action_def = self._gather_base_actions(
-            action_def, base_action_def
+            action_def,
+            base_action_def
         )
 
         super(AdHocAction, self).__init__(
@@ -348,12 +395,17 @@ class AdHocAction(PythonAction):
         )
 
         self.adhoc_action_def = action_def
+        self.task_ctx = task_ctx or {}
+        self.wf_ctx = wf_ctx or {}
 
     def validate_input(self, input_dict):
-        e_utils.validate_input(
-            self.adhoc_action_def,
+        expected_input = self.action_spec.get_input()
+
+        engine_utils.validate_input(
+            expected_input,
             input_dict,
-            self.action_spec
+            self.adhoc_action_def.name,
+            self.action_spec.__class__.__name__
         )
 
         super(AdHocAction, self).validate_input(
@@ -365,11 +417,28 @@ class AdHocAction(PythonAction):
 
         for action_def in self.adhoc_action_defs:
             action_spec = spec_parser.get_action_spec(action_def.spec)
+            for k, v in action_spec.get_input().items():
+                if (k not in base_input_dict or
+                        base_input_dict[k] is utils.NotDefined):
+                    base_input_dict[k] = v
+
             base_input_expr = action_spec.get_base_input()
+
             if base_input_expr:
+                wf_ex = (
+                    self.task_ex.workflow_execution if self.task_ex else None
+                )
+
+                ctx_view = data_flow.ContextView(
+                    base_input_dict,
+                    self.task_ctx,
+                    data_flow.get_workflow_environment_dict(wf_ex),
+                    self.wf_ctx
+                )
+
                 base_input_dict = expr.evaluate_recursively(
                     base_input_expr,
-                    base_input_dict
+                    ctx_view
                 )
             else:
                 base_input_dict = {}
@@ -379,18 +448,22 @@ class AdHocAction(PythonAction):
     def _prepare_output(self, result):
         # In case of error, we don't transform a result.
         if not result.is_error():
-            adhoc_action_spec = spec_parser.get_action_spec(
-                self.adhoc_action_def.spec
-            )
-
-            transformer = adhoc_action_spec.get_output()
-
-            if transformer is not None:
-                result = wf_utils.Result(
-                    data=expr.evaluate_recursively(transformer, result.data),
-                    error=result.error
+            for action_def in reversed(self.adhoc_action_defs):
+                adhoc_action_spec = spec_parser.get_action_spec(
+                    action_def.spec
                 )
 
+                transformer = adhoc_action_spec.get_output()
+
+                if transformer is not None:
+                    result = ml_actions.Result(
+                        data=expr.evaluate_recursively(
+                            transformer,
+                            result.data
+                        ),
+                        error=result.error
+                    )
+
         return result
 
     def _prepare_runtime_context(self, index, safe_rerun):
@@ -412,24 +485,32 @@ class AdHocAction(PythonAction):
         An ad-hoc action may be based on another ad-hoc action (and this
         recursively). Using twice the same base action is not allowed to
         avoid infinite loops. It stores the list of ad-hoc actions.
+
         :param action_def: Action definition
         :type action_def: ActionDefinition
         :param base_action_def: Original base action definition
         :type base_action_def: ActionDefinition
-        :return; The definition of the base system action
-        :rtype; ActionDefinition
+        :return: The definition of the base system action
+        :rtype: ActionDefinition
         """
+
         self.adhoc_action_defs = [action_def]
         original_base_name = self.action_spec.get_name()
         action_names = set([original_base_name])
 
         base = base_action_def
+
         while not base.is_system and base.name not in action_names:
             action_names.add(base.name)
             self.adhoc_action_defs.append(base)
 
             base_name = base.spec['base']
-            base = db_api.get_action_definition(base_name)
+            try:
+                base = db_api.get_action_definition(base_name)
+            except exc.DBEntityNotFoundError:
+                raise exc.InvalidActionException(
+                    "Failed to find action [action_name=%s]" % base_name
+                )
 
         # if the action is repeated
         if base.name in action_names:
@@ -444,13 +525,18 @@ class AdHocAction(PythonAction):
 class WorkflowAction(Action):
     """Workflow action."""
 
-    @profiler.trace('action-complete')
+    def __init__(self, wf_name, **kwargs):
+        super(WorkflowAction, self).__init__(None, **kwargs)
+        self.wf_name = wf_name
+
+    @profiler.trace('workflow-action-complete', hide_args=True)
     def complete(self, result):
         # No-op because in case of workflow result is already processed.
         pass
 
-    @profiler.trace('action-schedule')
-    def schedule(self, input_dict, target, index=0, desc='', safe_rerun=False):
+    @profiler.trace('workflkow-action-schedule', hide_args=True)
+    def schedule(self, input_dict, target, index=0, desc='', safe_rerun=False,
+                 timeout=None):
         assert not self.action_ex
 
         parent_wf_ex = self.task_ex.workflow_execution
@@ -458,14 +544,11 @@ class WorkflowAction(Action):
             parent_wf_ex.id
         )
 
-        task_spec = spec_parser.get_task_spec(self.task_ex.spec)
-
-        wf_spec_name = task_spec.get_workflow_name()
-
-        wf_def = e_utils.resolve_workflow_definition(
+        wf_def = engine_utils.resolve_workflow_definition(
             parent_wf_ex.workflow_name,
             parent_wf_spec.get_name(),
-            wf_spec_name
+            namespace=parent_wf_ex.params['namespace'],
+            wf_spec_name=self.wf_name
         )
 
         wf_spec = spec_parser.get_workflow_spec_by_definition_id(
@@ -473,13 +556,20 @@ class WorkflowAction(Action):
             wf_def.updated_at
         )
 
+        # If the parent has a root_execution_id, it must be a sub-workflow. So
+        # we should propogate that ID down. Otherwise the parent must be the
+        # root execution and we should use the parents ID.
+        root_execution_id = parent_wf_ex.root_execution_id or parent_wf_ex.id
+
         wf_params = {
+            'root_execution_id': root_execution_id,
             'task_execution_id': self.task_ex.id,
-            'index': index
+            'index': index,
+            'namespace': parent_wf_ex.params['namespace']
         }
 
-        if 'env' in parent_wf_ex.params:
-            wf_params['env'] = parent_wf_ex.params['env']
+        if 'notify' in parent_wf_ex.params:
+            wf_params['notify'] = parent_wf_ex.params['notify']
 
         for k, v in list(input_dict.items()):
             if k not in wf_spec.get_input():
@@ -488,14 +578,16 @@ class WorkflowAction(Action):
 
         wf_handler.start_workflow(
             wf_def.id,
+            wf_def.namespace,
+            None,
             input_dict,
             "sub-workflow execution",
             wf_params
         )
 
-    @profiler.trace('action-run')
+    @profiler.trace('workflow-action-run', hide_args=True)
     def run(self, input_dict, target, index=0, desc='', save=True,
-            safe_rerun=True):
+            safe_rerun=True, timeout=None):
         raise NotImplementedError('Does not apply to this WorkflowAction.')
 
     def is_sync(self, input_dict):
@@ -507,22 +599,6 @@ class WorkflowAction(Action):
         pass
 
 
-def _run_existing_action(action_ex_id, target):
-    action_ex = db_api.get_action_execution(action_ex_id)
-    action_def = db_api.get_action_definition(action_ex.name)
-
-    result = rpc.get_executor_client().run_action(
-        action_ex_id,
-        action_def.action_class,
-        action_def.attributes or {},
-        action_ex.input,
-        target,
-        safe_rerun=action_ex.runtime_context.get('safe_rerun', False)
-    )
-
-    return result.to_dict() if result else None
-
-
 def resolve_action_definition(action_spec_name, wf_name=None,
                               wf_spec_name=None):
     """Resolve action definition accounting for ad-hoc action namespacing.
@@ -532,6 +608,7 @@ def resolve_action_definition(action_spec_name, wf_name=None,
     :param wf_spec_name: Workflow name according to a spec.
     :return: Action definition (python or ad-hoc).
     """
+
     action_db = None
 
     if wf_name and wf_name != wf_spec_name:
@@ -544,10 +621,14 @@ def resolve_action_definition(action_spec_name, wf_name=None,
 
         action_full_name = "%s.%s" % (wb_name, action_spec_name)
 
-        action_db = db_api.load_action_definition(action_full_name)
+        action_db = lookup_utils.find_action_definition_by_name(
+            action_full_name
+        )
 
     if not action_db:
-        action_db = db_api.load_action_definition(action_spec_name)
+        action_db = lookup_utils.find_action_definition_by_name(
+            action_spec_name
+        )
 
     if not action_db:
         raise exc.InvalidActionException(
