@@ -1,6 +1,5 @@
-# -*- coding: utf-8 -*-
-#
 # Copyright 2014 - Mirantis, Inc.
+# Copyright 2017 - Brocade Communications Systems, Inc.
 #
 #    Licensed under the Apache License, Version 2.0 (the "License");
 #    you may not use this file except in compliance with the License.
@@ -29,13 +28,16 @@ class Engine(object):
     """Engine interface."""
 
     @abc.abstractmethod
-    def start_workflow(self, wf_identifier, wf_input, description='',
-                       **params):
+    def start_workflow(self, wf_identifier, wf_namespace='', wf_ex_id=None,
+                       wf_input=None, description='', **params):
         """Starts the specified workflow.
 
         :param wf_identifier: Workflow ID or name. Workflow ID is recommended,
             workflow name will be deprecated since Mitaka.
+        :param wf_namespace: Workflow namespace.
         :param wf_input: Workflow input data as a dictionary.
+        :param wf_ex_id: Workflow execution id. If passed, it will be set
+            in the new execution object.
         :param description: Execution description.
         :param params: Additional workflow type specific parameters.
         :return: Workflow execution object.
@@ -56,11 +58,13 @@ class Engine(object):
         raise NotImplementedError
 
     @abc.abstractmethod
-    def on_action_complete(self, action_ex_id, result, wf_action=False):
+    def on_action_complete(self, action_ex_id, result, wf_action=False,
+                           async_=False):
         """Accepts action result and continues the workflow.
 
         Action execution result here is a result which comes from an
         action/workflow associated which the task.
+
         :param action_ex_id: Action execution id.
         :param result: Action/workflow result. Instance of
             mistral.workflow.base.Result
@@ -68,6 +72,8 @@ class Engine(object):
             a workflow execution rather than action execution. It happens
             when a nested workflow execution sends its result to a parent
             workflow.
+        :param async: If True, run action in asynchronous mode (w/o waiting
+            for completion).
         :return: Action(or workflow if wf_action=True) execution object.
         """
         raise NotImplementedError
@@ -125,38 +131,13 @@ class Engine(object):
         """
         raise NotImplementedError
 
-
-@six.add_metaclass(abc.ABCMeta)
-class Executor(object):
-    """Action executor interface."""
-
     @abc.abstractmethod
-    def run_action(self, action_ex_id, action_class_str, attributes,
-                   action_params, safe_rerun, redelivered=False):
-        """Runs action.
+    def report_running_actions(self, action_ex_ids):
+        """Receives the heartbeat about the running actions.
 
-        :param action_ex_id: Corresponding action execution id.
-        :param action_class_str: Path to action class in dot notation.
-        :param attributes: Attributes of action class which will be set to.
-        :param action_params: Action parameters.
-        :param safe_rerun: Tells if given action can be safely rerun.
-        :param redelivered: Tells if given action was run before on another
-            executor.
+        :param action_ex_ids: The action execution ids.
         """
-        raise NotImplementedError()
-
-
-@six.add_metaclass(abc.ABCMeta)
-class EventEngine(object):
-    """Action event trigger interface."""
-
-    @abc.abstractmethod
-    def create_event_trigger(self, trigger, events):
-        raise NotImplementedError()
-
-    @abc.abstractmethod
-    def delete_event_trigger(self, trigger, events):
-        raise NotImplementedError()
+        raise NotImplementedError
 
 
 @six.add_metaclass(abc.ABCMeta)
@@ -179,6 +160,7 @@ class TaskPolicy(object):
 
         ctx_view = data_flow.ContextView(
             task_ex.in_context,
+            data_flow.get_workflow_environment_dict(wf_ex),
             wf_ex.context,
             wf_ex.input
         )
@@ -197,6 +179,7 @@ class TaskPolicy(object):
 
         ctx_view = data_flow.ContextView(
             task_ex.in_context,
+            data_flow.get_workflow_environment_dict(wf_ex),
             wf_ex.context,
             wf_ex.input
         )
@@ -215,5 +198,5 @@ class TaskPolicy(object):
             raise exc.InvalidModelException(
                 "Invalid data type in %s: %s. Value(s) can be shown after "
                 "YAQL evaluating. If you use YAQL here, please correct it."
-                % (self.__class__.__name__, e.message)
+                % (self.__class__.__name__, str(e))
             )
