@@ -13,8 +13,11 @@
 #    See the License for the specific language governing permissions and
 #    limitations under the License.
 
+from mistral.db import utils as db_utils
 from mistral.db.v2 import api as db_api
+from mistral.engine import action_queue
 from mistral.engine import base
+from mistral.engine import workflow_handler as wf_handler
 from mistral import expressions
 from mistral.services import scheduler
 from mistral.utils import wf_trace
@@ -25,6 +28,9 @@ import six
 
 _CONTINUE_TASK_PATH = 'mistral.engine.policies._continue_task'
 _COMPLETE_TASK_PATH = 'mistral.engine.policies._complete_task'
+_FAIL_IF_INCOMPLETE_TASK_PATH = (
+    'mistral.engine.policies._fail_task_if_incomplete'
+)
 
 
 def _log_task_delay(task_ex, delay_sec):
@@ -77,7 +83,7 @@ def build_wait_before_policy(policies_spec):
 
     wait_before = policies_spec.get_wait_before()
 
-    if (isinstance(wait_before, six.string_types) or wait_before > 0):
+    if isinstance(wait_before, six.string_types) or wait_before > 0:
         return WaitBeforePolicy(wait_before)
     else:
         return None
@@ -89,7 +95,7 @@ def build_wait_after_policy(policies_spec):
 
     wait_after = policies_spec.get_wait_after()
 
-    if (isinstance(wait_after, six.string_types) or wait_after > 0):
+    if isinstance(wait_after, six.string_types) or wait_after > 0:
         return WaitAfterPolicy(wait_after)
     else:
         return None
@@ -118,7 +124,7 @@ def build_timeout_policy(policies_spec):
 
     timeout_policy = policies_spec.get_timeout()
 
-    if (isinstance(timeout_policy, six.string_types) or timeout_policy > 0):
+    if isinstance(timeout_policy, six.string_types) or timeout_policy > 0:
         return TimeoutPolicy(timeout_policy)
     else:
         return None
@@ -157,7 +163,10 @@ def _ensure_context_has_key(runtime_context, key):
 class WaitBeforePolicy(base.TaskPolicy):
     _schema = {
         "properties": {
-            "delay": {"type": "integer"}
+            "delay": {
+                "type": "integer",
+                "minimum": 0
+            }
         }
     }
 
@@ -167,6 +176,10 @@ class WaitBeforePolicy(base.TaskPolicy):
     def before_task_start(self, task_ex, task_spec):
         super(WaitBeforePolicy, self).before_task_start(task_ex, task_spec)
 
+        # No need to wait for a task if delay is 0
+        if self.delay == 0:
+            return
+
         context_key = 'wait_before_policy'
 
         runtime_context = _ensure_context_has_key(
@@ -208,7 +221,10 @@ class WaitBeforePolicy(base.TaskPolicy):
 class WaitAfterPolicy(base.TaskPolicy):
     _schema = {
         "properties": {
-            "delay": {"type": "integer"}
+            "delay": {
+                "type": "integer",
+                "minimum": 0
+            }
         }
     }
 
@@ -218,6 +234,10 @@ class WaitAfterPolicy(base.TaskPolicy):
     def after_task_complete(self, task_ex, task_spec):
         super(WaitAfterPolicy, self).after_task_complete(task_ex, task_spec)
 
+        # No need to postpone a task if delay is 0
+        if self.delay == 0:
+            return
+
         context_key = 'wait_after_policy'
 
         runtime_context = _ensure_context_has_key(
@@ -240,9 +260,9 @@ class WaitAfterPolicy(base.TaskPolicy):
         end_state = task_ex.state
         end_state_info = task_ex.state_info
 
-        # TODO(rakhmerov): Policies probably needs to have tasks.Task
-        # interface in order to change manage task state safely.
-        # Set task state to 'DELAYED'.
+        # TODO(rakhmerov): Policies probably need to have tasks.Task
+        # interface in order to manage task state safely.
+        # Set task state to 'RUNNING_DELAYED'.
         task_ex.state = states.RUNNING_DELAYED
         task_ex.state_info = (
             'Suspended by wait-after policy for %s seconds' % self.delay
@@ -262,15 +282,21 @@ class WaitAfterPolicy(base.TaskPolicy):
 class RetryPolicy(base.TaskPolicy):
     _schema = {
         "properties": {
-            "delay": {"type": "integer"},
-            "count": {"type": "integer"}
+            "delay": {
+                "type": "integer",
+                "minimum": 0
+            },
+            "count": {
+                "type": "integer",
+                "minimum": 0
+            },
         }
     }
 
     def __init__(self, count, delay, break_on, continue_on):
         self.count = count
         self.delay = delay
-        self.break_on = break_on
+        self._break_on_clause = break_on
         self._continue_on_clause = continue_on
 
     def after_task_complete(self, task_ex, task_spec):
@@ -291,6 +317,10 @@ class RetryPolicy(base.TaskPolicy):
         """
         super(RetryPolicy, self).after_task_complete(task_ex, task_spec)
 
+        # There is nothing to repeat
+        if self.count == 0:
+            return
+
         # TODO(m4dcoder): If the task_ex.action_executions and
         # task_ex.workflow_executions collection are not called,
         # then the retry_no in the runtime_context of the task_ex will not
@@ -308,6 +338,7 @@ class RetryPolicy(base.TaskPolicy):
         wf_ex = task_ex.workflow_execution
 
         ctx_view = data_flow.ContextView(
+            data_flow.get_current_task_dict(task_ex),
             data_flow.evaluate_task_outbound_context(task_ex),
             wf_ex.context,
             wf_ex.input
@@ -318,11 +349,16 @@ class RetryPolicy(base.TaskPolicy):
             ctx_view
         )
 
+        break_on_evaluation = expressions.evaluate(
+            self._break_on_clause,
+            ctx_view
+        )
+
         task_ex.runtime_context = runtime_context
 
         state = task_ex.state
 
-        if not states.is_completed(state):
+        if not states.is_completed(state) or states.is_cancelled(state):
             return
 
         policy_context = runtime_context[context_key]
@@ -333,14 +369,22 @@ class RetryPolicy(base.TaskPolicy):
             retry_no = policy_context['retry_no']
             del policy_context['retry_no']
 
-        retries_remain = retry_no + 1 < self.count
+        retries_remain = retry_no < self.count
 
-        stop_continue_flag = (task_ex.state == states.SUCCESS and
-                              not self._continue_on_clause)
-        stop_continue_flag = (stop_continue_flag or
-                              (self._continue_on_clause and
-                               not continue_on_evaluation))
-        break_triggered = task_ex.state == states.ERROR and self.break_on
+        stop_continue_flag = (
+            task_ex.state == states.SUCCESS and
+            not self._continue_on_clause
+        )
+
+        stop_continue_flag = (
+            stop_continue_flag or
+            (self._continue_on_clause and not continue_on_evaluation)
+        )
+
+        break_triggered = (
+            task_ex.state == states.ERROR and
+            break_on_evaluation
+        )
 
         if not retries_remain or break_triggered or stop_continue_flag:
             return
@@ -348,6 +392,7 @@ class RetryPolicy(base.TaskPolicy):
         _log_task_delay(task_ex, self.delay)
 
         data_flow.invalidate_task_execution_result(task_ex)
+
         task_ex.state = states.RUNNING_DELAYED
 
         policy_context['retry_no'] = retry_no + 1
@@ -360,11 +405,23 @@ class RetryPolicy(base.TaskPolicy):
             task_ex_id=task_ex.id,
         )
 
+    @staticmethod
+    def refresh_runtime_context(task_ex):
+        runtime_context = task_ex.runtime_context or {}
+        retry_task_policy = runtime_context.get('retry_task_policy')
+
+        if retry_task_policy:
+            retry_task_policy['retry_no'] = 0
+            task_ex.runtime_context['retry_task_policy'] = retry_task_policy
+
 
 class TimeoutPolicy(base.TaskPolicy):
     _schema = {
         "properties": {
-            "delay": {"type": "integer"}
+            "delay": {
+                "type": "integer",
+                "minimum": 0
+            }
         }
     }
 
@@ -374,9 +431,13 @@ class TimeoutPolicy(base.TaskPolicy):
     def before_task_start(self, task_ex, task_spec):
         super(TimeoutPolicy, self).before_task_start(task_ex, task_spec)
 
+        # No timeout if delay is 0
+        if self.delay == 0:
+            return
+
         scheduler.schedule_call(
             None,
-            'mistral.engine.policies._fail_task_if_incomplete',
+            _FAIL_IF_INCOMPLETE_TASK_PATH,
             self.delay,
             task_ex_id=task_ex.id,
             timeout=self.delay
@@ -411,14 +472,17 @@ class PauseBeforePolicy(base.TaskPolicy):
             (task_ex.name, task_ex.workflow_execution.state, states.PAUSED)
         )
 
-        task_ex.workflow_execution.state = states.PAUSED
         task_ex.state = states.IDLE
+        wf_handler.pause_workflow(task_ex.workflow_execution)
 
 
 class ConcurrencyPolicy(base.TaskPolicy):
     _schema = {
         "properties": {
-            "concurrency": {"type": "integer"},
+            "concurrency": {
+                "type": "integer",
+                "minimum": 0
+            }
         }
     }
 
@@ -428,6 +492,9 @@ class ConcurrencyPolicy(base.TaskPolicy):
     def before_task_start(self, task_ex, task_spec):
         super(ConcurrencyPolicy, self).before_task_start(task_ex, task_spec)
 
+        if self.concurrency == 0:
+            return
+
         # This policy doesn't do anything except validating "concurrency"
         # property value and setting a variable into task runtime context.
         # This variable is then used to define how many action executions
@@ -443,6 +510,8 @@ class ConcurrencyPolicy(base.TaskPolicy):
         task_ex.runtime_context = runtime_context
 
 
+@db_utils.retry_on_db_error
+@action_queue.process
 def _continue_task(task_ex_id):
     from mistral.engine import task_handler
 
@@ -450,6 +519,8 @@ def _continue_task(task_ex_id):
         task_handler.continue_task(db_api.get_task_execution(task_ex_id))
 
 
+@db_utils.retry_on_db_error
+@action_queue.process
 def _complete_task(task_ex_id, state, state_info):
     from mistral.engine import task_handler
 
@@ -461,6 +532,8 @@ def _complete_task(task_ex_id, state, state_info):
         )
 
 
+@db_utils.retry_on_db_error
+@action_queue.process
 def _fail_task_if_incomplete(task_ex_id, timeout):
     from mistral.engine import task_handler
 
