@@ -14,55 +14,60 @@
 #    limitations under the License.
 
 import copy
-import six
 
 from mistral.db.v2 import api as db_api
 from mistral import exceptions as exc
 from mistral import utils
 
 
-# TODO(rakhmerov): This method is too abstract, validation rules may vary
-#  depending on object type (action, wf), it's not clear what it can be
-# applied to.
-# TODO(rakhmerov): It must not do any manipulations with parameters
-#  (input_dict)!
-def validate_input(definition, input_dict, spec=None):
-    input_param_names = copy.deepcopy(list((input_dict or {}).keys()))
-    missing_param_names = []
+def _compare_parameters(expected_input, actual_input):
+    """Compares the expected parameters with the actual parameters.
 
-    spec_input = (spec.get_input() if spec else
-                  utils.get_dict_from_string(definition.input))
+    :param expected_input: Expected dict of parameters.
+    :param actual_input: Actual dict of parameters.
+    :return: Tuple {missing parameter names, unexpected parameter names}
+    """
 
-    for p_name, p_value in six.iteritems(spec_input):
-        if p_value is utils.NotDefined and p_name not in input_param_names:
-            missing_param_names.append(str(p_name))
+    missing_params = []
+    unexpected_params = copy.deepcopy(list((actual_input or {}).keys()))
 
-        if p_name in input_param_names:
-            input_param_names.remove(p_name)
+    for p_name, p_value in expected_input.items():
+        if p_value is utils.NotDefined and p_name not in unexpected_params:
+            missing_params.append(str(p_name))
 
-    if missing_param_names or input_param_names:
+        if p_name in unexpected_params:
+            unexpected_params.remove(p_name)
+
+    return missing_params, unexpected_params
+
+
+def validate_input(expected_input, actual_input, obj_name, obj_class):
+    actual_input = actual_input or {}
+
+    missing, unexpected = _compare_parameters(
+        expected_input,
+        actual_input
+    )
+
+    if missing or unexpected:
         msg = 'Invalid input [name=%s, class=%s'
-        msg_props = [definition.name, spec.__class__.__name__]
+        msg_props = [obj_name, obj_class]
 
-        if missing_param_names:
+        if missing:
             msg += ', missing=%s'
-            msg_props.append(missing_param_names)
+            msg_props.append(missing)
 
-        if input_param_names:
+        if unexpected:
             msg += ', unexpected=%s'
-            msg_props.append(input_param_names)
+            msg_props.append(unexpected)
 
         msg += ']'
 
-        raise exc.InputException(
-            msg % tuple(msg_props)
-        )
-    else:
-        utils.merge_dicts(input_dict, spec_input, overwrite=False)
+        raise exc.InputException(msg % tuple(msg_props))
 
 
 def resolve_workflow_definition(parent_wf_name, parent_wf_spec_name,
-                                wf_spec_name):
+                                namespace, wf_spec_name):
     wf_def = None
 
     if parent_wf_name != parent_wf_spec_name:
@@ -75,14 +80,15 @@ def resolve_workflow_definition(parent_wf_name, parent_wf_spec_name,
 
         wf_full_name = "%s.%s" % (wb_name, wf_spec_name)
 
-        wf_def = db_api.load_workflow_definition(wf_full_name)
+        wf_def = db_api.load_workflow_definition(wf_full_name, namespace)
 
     if not wf_def:
-        wf_def = db_api.load_workflow_definition(wf_spec_name)
+        wf_def = db_api.load_workflow_definition(wf_spec_name, namespace)
 
     if not wf_def:
         raise exc.WorkflowException(
-            "Failed to find workflow [name=%s]" % wf_spec_name
+            "Failed to find workflow [name=%s] [namespace=%s]" %
+            (wf_spec_name, namespace)
         )
 
     return wf_def
