@@ -1,5 +1,6 @@
 # Copyright 2016 - Nokia Networks.
 # Copyright 2016 - Brocade Communications Systems, Inc.
+# Copyright 2018 - Extreme Networks, Inc.
 #
 #    Licensed under the Apache License, Version 2.0 (the "License");
 #    you may not use this file except in compliance with the License.
@@ -14,6 +15,7 @@
 #    limitations under the License.
 
 import abc
+import json
 from oslo_config import cfg
 from oslo_log import log as logging
 from osprofiler import profiler
@@ -21,29 +23,29 @@ import six
 
 from mistral.db.v2 import api as db_api
 from mistral.db.v2.sqlalchemy import models as db_models
+from mistral.engine import action_queue
 from mistral.engine import dispatcher
-from mistral.engine.rpc_backend import rpc
-from mistral.engine import utils as eng_utils
+from mistral.engine import utils as engine_utils
 from mistral import exceptions as exc
-from mistral.services import scheduler
+from mistral import expressions as expr
+from mistral.lang import parser as spec_parser
+from mistral.notifiers import base as notif
+from mistral.notifiers import notification_events as events
+from mistral.services import triggers
 from mistral.services import workflows as wf_service
 from mistral import utils
+from mistral.utils import merge_dicts
 from mistral.utils import wf_trace
-from mistral.workbook import parser as spec_parser
 from mistral.workflow import base as wf_base
 from mistral.workflow import commands
 from mistral.workflow import data_flow
 from mistral.workflow import lookup_utils
 from mistral.workflow import states
-from mistral.workflow import utils as wf_utils
+from mistral_lib import actions as ml_actions
 
 
 LOG = logging.getLogger(__name__)
 
-_SEND_RESULT_TO_PARENT_WORKFLOW_PATH = (
-    'mistral.engine.workflows._send_result_to_parent_workflow'
-)
-
 
 @six.add_metaclass(abc.ABCMeta)
 class Workflow(object):
@@ -53,8 +55,7 @@ class Workflow(object):
     Mistral engine or its components in order to manipulate with workflows.
     """
 
-    def __init__(self, wf_def, wf_ex=None):
-        self.wf_def = wf_def
+    def __init__(self, wf_ex=None):
         self.wf_ex = wf_ex
 
         if wf_ex:
@@ -63,38 +64,72 @@ class Workflow(object):
                 wf_ex.id
             )
         else:
-            # New workflow execution.
-            self.wf_spec = spec_parser.get_workflow_spec_by_definition_id(
-                wf_def.id,
-                wf_def.updated_at
-            )
+            self.wf_spec = None
+
+    def notify(self, event):
+        publishers = self.wf_ex.params.get('notify')
+
+        if not publishers and not isinstance(publishers, list):
+            return
+
+        notifier = notif.get_notifier(cfg.CONF.notifier.type)
+
+        notifier.notify(
+            self.wf_ex.id,
+            self.wf_ex.to_dict(),
+            event,
+            self.wf_ex.updated_at,
+            publishers
+        )
 
     @profiler.trace('workflow-start')
-    def start(self, input_dict, desc='', params=None):
+    def start(self, wf_def, wf_ex_id, input_dict, desc='', params=None):
         """Start workflow.
 
+        :param wf_def: Workflow definition.
+        :param wf_ex_id: Workflow execution id.
         :param input_dict: Workflow input.
         :param desc: Workflow execution description.
         :param params: Workflow type specific parameters.
+
+        :raises
         """
 
         assert not self.wf_ex
 
-        wf_trace.info(self.wf_ex, "Starting workflow: %s" % self.wf_def)
+        # New workflow execution.
+        self.wf_spec = spec_parser.get_workflow_spec_by_definition_id(
+            wf_def.id,
+            wf_def.updated_at
+        )
 
-        # TODO(rakhmerov): This call implicitly changes input_dict! Fix it!
-        # After fix we need to move validation after adding risky fields.
-        eng_utils.validate_input(self.wf_def, input_dict, self.wf_spec)
+        wf_trace.info(
+            self.wf_ex,
+            'Starting workflow [name=%s, input=%s]' %
+            (wf_def.name, utils.cut(input_dict))
+        )
 
-        self._create_execution(input_dict, desc, params)
+        self.validate_input(input_dict)
+
+        self._create_execution(
+            wf_def,
+            wf_ex_id,
+            self.prepare_input(input_dict),
+            desc,
+            params
+        )
 
         self.set_state(states.RUNNING)
 
+        # Publish event as soon as state is set to running.
+        self.notify(events.WORKFLOW_LAUNCHED)
+
         wf_ctrl = wf_base.get_controller(self.wf_ex, self.wf_spec)
 
-        cmds = wf_ctrl.continue_workflow()
-
-        dispatcher.dispatch_workflow_commands(self.wf_ex, cmds)
+        dispatcher.dispatch_workflow_commands(
+            self.wf_ex,
+            wf_ctrl.continue_workflow()
+        )
 
     def stop(self, state, msg=None):
         """Stop workflow.
@@ -102,26 +137,38 @@ class Workflow(object):
         :param state: New workflow state.
         :param msg: Additional explaining message.
         """
-
         assert self.wf_ex
 
         if state == states.SUCCESS:
-            wf_ctrl = wf_base.get_controller(self.wf_ex)
-
-            final_context = {}
-
-            try:
-                final_context = wf_ctrl.evaluate_workflow_final_context()
-            except Exception as e:
-                LOG.warning(
-                    'Failed to get final context for %s: %s' % (self.wf_ex, e)
-                )
-
-            return self._succeed_workflow(final_context, msg)
+            self._succeed_workflow(self._get_final_context(), msg)
         elif state == states.ERROR:
-            return self._fail_workflow(msg)
+            self._fail_workflow(self._get_final_context(), msg)
         elif state == states.CANCELLED:
-            return self._cancel_workflow(msg)
+            self._cancel_workflow(msg)
+
+    def pause(self, msg=None):
+        """Pause workflow.
+
+        :param msg: Additional explaining message.
+        """
+
+        assert self.wf_ex
+
+        if states.is_paused(self.wf_ex.state):
+            return
+
+        # Set the state of this workflow to paused.
+        self.set_state(states.PAUSED, state_info=msg)
+
+        # Publish event.
+        self.notify(events.WORKFLOW_PAUSED)
+
+        # If workflow execution is a subworkflow,
+        # schedule update to the task execution.
+        if self.wf_ex.task_execution_id:
+            # Import the task_handler module here to avoid circular reference.
+            from mistral.engine import task_handler
+            task_handler.schedule_on_action_update(self.wf_ex)
 
     def resume(self, env=None):
         """Resume workflow.
@@ -133,7 +180,10 @@ class Workflow(object):
 
         wf_service.update_workflow_execution_env(self.wf_ex, env)
 
-        self.set_state(states.RUNNING, recursive=True)
+        self.set_state(states.RUNNING)
+
+        # Publish event.
+        self.notify(events.WORKFLOW_RESUMED)
 
         wf_ctrl = wf_base.get_controller(self.wf_ex)
 
@@ -142,6 +192,29 @@ class Workflow(object):
 
         self._continue_workflow(cmds)
 
+        # If workflow execution is a subworkflow,
+        # schedule update to the task execution.
+        if self.wf_ex.task_execution_id:
+            # Import the task_handler module here to avoid circular reference.
+            from mistral.engine import task_handler
+
+            task_handler.schedule_on_action_update(self.wf_ex)
+
+    def prepare_input(self, input_dict):
+        for k, v in self.wf_spec.get_input().items():
+            if k not in input_dict or input_dict[k] is utils.NotDefined:
+                input_dict[k] = v
+
+        return input_dict
+
+    def validate_input(self, input_dict):
+        engine_utils.validate_input(
+            self.wf_spec.get_input(),
+            input_dict,
+            self.wf_spec.get_name(),
+            self.wf_spec.__class__.__name__
+        )
+
     def rerun(self, task_ex, reset=True, env=None):
         """Rerun workflow from the given task.
 
@@ -156,7 +229,7 @@ class Workflow(object):
         # Since some lookup utils functions may use cache for completed tasks
         # we need to clean caches to make sure that stale objects can't be
         # retrieved.
-        lookup_utils.clean_caches()
+        lookup_utils.clear_caches()
 
         wf_service.update_workflow_execution_env(self.wf_ex, env)
 
@@ -167,14 +240,23 @@ class Workflow(object):
         # Calculate commands to process next.
         cmds = wf_ctrl.rerun_tasks([task_ex], reset=reset)
 
+        if cmds:
+            # Import the task_handler module here to avoid circular reference.
+            from mistral.engine import policies
+
+            policies.RetryPolicy.refresh_runtime_context(task_ex)
+
         self._continue_workflow(cmds)
 
+    def _get_backlog(self):
+        return self.wf_ex.runtime_context.get(dispatcher.BACKLOG_KEY)
+
     def _continue_workflow(self, cmds):
         # When resuming a workflow we need to ignore all 'pause'
         # commands because workflow controller takes tasks that
         # completed within the period when the workflow was paused.
         cmds = list(
-            filter(lambda c: not isinstance(c, commands.PauseWorkflow), cmds)
+            [c for c in cmds if not isinstance(c, commands.PauseWorkflow)]
         )
 
         # Since there's no explicit task causing the operation
@@ -185,7 +267,7 @@ class Workflow(object):
             if states.is_completed(t_ex.state) and not t_ex.processed:
                 t_ex.processed = True
 
-        if cmds:
+        if cmds or self._get_backlog():
             dispatcher.dispatch_workflow_commands(self.wf_ex, cmds)
         else:
             self.check_and_complete()
@@ -196,16 +278,37 @@ class Workflow(object):
 
         return db_api.acquire_lock(db_models.WorkflowExecution, self.wf_ex.id)
 
-    def _create_execution(self, input_dict, desc, params):
+    def _get_final_context(self):
+        final_ctx = {}
+
+        wf_ctrl = wf_base.get_controller(self.wf_ex)
+
+        try:
+            final_ctx = wf_ctrl.evaluate_workflow_final_context()
+        except Exception as e:
+            LOG.warning(
+                'Failed to get final context for workflow execution. '
+                '[wf_ex_id: %s, wf_name: %s, error: %s]',
+                self.wf_ex.id,
+                self.wf_ex.name,
+                str(e)
+            )
+
+        return final_ctx
+
+    def _create_execution(self, wf_def, wf_ex_id, input_dict, desc, params):
         self.wf_ex = db_api.create_workflow_execution({
-            'name': self.wf_def.name,
+            'id': wf_ex_id,
+            'name': wf_def.name,
             'description': desc,
-            'workflow_name': self.wf_def.name,
-            'workflow_id': self.wf_def.id,
+            'workflow_name': wf_def.name,
+            'workflow_namespace': wf_def.namespace,
+            'workflow_id': wf_def.id,
             'spec': self.wf_spec.to_dict(),
             'state': states.IDLE,
             'output': {},
             'task_execution_id': params.get('task_execution_id'),
+            'root_execution_id': params.get('root_execution_id'),
             'runtime_context': {
                 'index': params.get('index', 0)
             },
@@ -213,16 +316,12 @@ class Workflow(object):
 
         self.wf_ex.input = input_dict or {}
 
-        env = _get_environment(params)
-
-        if env:
-            params['env'] = env
+        params['env'] = _get_environment(params)
 
         self.wf_ex.params = params
 
         data_flow.add_openstack_data_to_context(self.wf_ex)
         data_flow.add_execution_to_context(self.wf_ex)
-        data_flow.add_environment_to_context(self.wf_ex)
         data_flow.add_workflow_variables_to_context(self.wf_ex, self.wf_spec)
 
         spec_parser.cache_workflow_spec_by_execution_id(
@@ -237,13 +336,27 @@ class Workflow(object):
         cur_state = self.wf_ex.state
 
         if states.is_valid_transition(cur_state, state):
-            self.wf_ex.state = state
-            self.wf_ex.state_info = state_info
+            wf_ex = db_api.update_workflow_execution_state(
+                id=self.wf_ex.id,
+                cur_state=cur_state,
+                state=state
+            )
+
+            if wf_ex is None:
+                # Do nothing because the state was updated previously.
+                return
+
+            self.wf_ex = wf_ex
+            self.wf_ex.state_info = json.dumps(state_info) \
+                if isinstance(state_info, dict) else state_info
 
             wf_trace.info(
                 self.wf_ex,
-                "Workflow '%s' [%s -> %s, msg=%s]"
-                % (self.wf_ex.workflow_name, cur_state, state, state_info)
+                "Workflow '%s' [%s -> %s, msg=%s]" %
+                (self.wf_ex.workflow_name,
+                 cur_state,
+                 state,
+                 self.wf_ex.state_info)
             )
         else:
             msg = ("Can't change workflow execution state from %s to %s. "
@@ -256,15 +369,19 @@ class Workflow(object):
         # only if it completed successfully or failed.
         self.wf_ex.accepted = states.is_completed(state)
 
+        if states.is_completed(state):
+            # No need to keep task executions of this workflow in the
+            # lookup cache anymore.
+            lookup_utils.invalidate_cached_task_executions(self.wf_ex.id)
+
+            triggers.on_workflow_complete(self.wf_ex)
+
         if recursive and self.wf_ex.task_execution_id:
             parent_task_ex = db_api.get_task_execution(
                 self.wf_ex.task_execution_id
             )
 
-            parent_wf = Workflow(
-                db_api.get_workflow_definition(parent_task_ex.workflow_id),
-                parent_task_ex.workflow_execution
-            )
+            parent_wf = Workflow(wf_ex=parent_task_ex.workflow_execution)
 
             parent_wf.lock()
 
@@ -309,41 +426,75 @@ class Workflow(object):
             self._succeed_workflow(ctx)
         else:
             msg = _build_fail_info_message(wf_ctrl, self.wf_ex)
+            final_context = wf_ctrl.evaluate_workflow_final_context()
 
-            self._fail_workflow(msg)
+            self._fail_workflow(final_context, msg)
 
         return 0
 
     def _succeed_workflow(self, final_context, msg=None):
         self.wf_ex.output = data_flow.evaluate_workflow_output(
             self.wf_ex,
-            self.wf_spec,
+            self.wf_spec.get_output(),
             final_context
         )
 
         # Set workflow execution to success until after output is evaluated.
         self.set_state(states.SUCCESS, msg)
 
-        if self.wf_ex.task_execution_id:
-            self._schedule_send_result_to_parent_workflow()
+        # Publish event.
+        self.notify(events.WORKFLOW_SUCCEEDED)
 
-    def _fail_workflow(self, msg):
+        if self.wf_ex.task_execution_id:
+            self._send_result_to_parent_workflow()
+
+    def _fail_workflow(self, final_context, msg):
         if states.is_paused_or_completed(self.wf_ex.state):
             return
 
+        output_on_error = {}
+
+        try:
+            output_on_error = data_flow.evaluate_workflow_output(
+                self.wf_ex,
+                self.wf_spec.get_output_on_error(),
+                final_context
+            )
+        except exc.MistralException as e:
+            msg = (
+                "Failed to evaluate expression in output-on-error! "
+                "(output-on-error: '%s', exception: '%s' Cause: '%s'"
+                % (self.wf_spec.get_output_on_error(), e, msg)
+            )
+            LOG.error(msg)
+
         self.set_state(states.ERROR, state_info=msg)
 
         # When we set an ERROR state we should safely set output value getting
         # w/o exceptions due to field size limitations.
-        msg = utils.cut_by_kb(
-            msg,
-            cfg.CONF.engine.execution_field_size_limit_kb
-        )
 
-        self.wf_ex.output = {'result': msg}
+        length_output_on_error = len(str(output_on_error).encode("utf-8"))
+        total_output_length = utils.get_number_of_chars_from_kilobytes(
+            cfg.CONF.engine.execution_field_size_limit_kb)
+
+        if length_output_on_error < total_output_length:
+            msg = utils.cut_by_char(
+                msg,
+                total_output_length - length_output_on_error
+            )
+        else:
+            msg = utils.cut_by_kb(
+                msg,
+                cfg.CONF.engine.execution_field_size_limit_kb
+            )
+
+        self.wf_ex.output = merge_dicts({'result': msg}, output_on_error)
+
+        # Publish event.
+        self.notify(events.WORKFLOW_FAILED)
 
         if self.wf_ex.task_execution_id:
-            self._schedule_send_result_to_parent_workflow()
+            self._send_result_to_parent_workflow()
 
     def _cancel_workflow(self, msg):
         if states.is_completed(self.wf_ex.state):
@@ -360,25 +511,51 @@ class Workflow(object):
 
         self.wf_ex.output = {'result': msg}
 
-        if self.wf_ex.task_execution_id:
-            self._schedule_send_result_to_parent_workflow()
+        # Publish event.
+        self.notify(events.WORKFLOW_CANCELLED)
 
-    def _schedule_send_result_to_parent_workflow(self):
-        scheduler.schedule_call(
-            None,
-            _SEND_RESULT_TO_PARENT_WORKFLOW_PATH,
-            0,
-            wf_ex_id=self.wf_ex.id
+        if self.wf_ex.task_execution_id:
+            self._send_result_to_parent_workflow()
+
+    def _send_result_to_parent_workflow(self):
+        if self.wf_ex.state == states.SUCCESS:
+            result = ml_actions.Result(data=self.wf_ex.output)
+        elif self.wf_ex.state == states.ERROR:
+            err_msg = (
+                self.wf_ex.state_info or
+                'Failed subworkflow [execution_id=%s]' % self.wf_ex.id
+            )
+
+            result = ml_actions.Result(error=err_msg)
+        elif self.wf_ex.state == states.CANCELLED:
+            err_msg = (
+                self.wf_ex.state_info or
+                'Cancelled subworkflow [execution_id=%s]' % self.wf_ex.id
+            )
+
+            result = ml_actions.Result(error=err_msg, cancel=True)
+        else:
+            raise RuntimeError(
+                "Method _send_result_to_parent_workflow() must never be called"
+                " if a workflow is not in SUCCESS, ERROR or CANCELLED state."
+            )
+
+        action_queue.schedule_on_action_complete(
+            self.wf_ex.id,
+            result,
+            wf_action=True
         )
 
 
 def _get_environment(params):
     env = params.get('env', {})
 
-    if isinstance(env, dict):
-        return env
+    if not env:
+        return {}
 
-    if isinstance(env, six.string_types):
+    if isinstance(env, dict):
+        env_dict = env
+    elif isinstance(env, six.string_types):
         env_db = db_api.load_environment(env)
 
         if not env_db:
@@ -386,44 +563,18 @@ def _get_environment(params):
                 'Environment is not found: %s' % env
             )
 
-        return env_db.variables
-
-    raise exc.InputException(
-        'Unexpected value type for environment [env=%s, type=%s]'
-        % (env, type(env))
-    )
-
-
-def _send_result_to_parent_workflow(wf_ex_id):
-    wf_ex = db_api.get_workflow_execution(wf_ex_id)
-
-    if wf_ex.state == states.SUCCESS:
-        result = wf_utils.Result(data=wf_ex.output)
-    elif wf_ex.state == states.ERROR:
-        err_msg = (
-            wf_ex.state_info or
-            'Failed subworkflow [execution_id=%s]' % wf_ex.id
-        )
-
-        result = wf_utils.Result(error=err_msg)
-    elif wf_ex.state == states.CANCELLED:
-        err_msg = (
-            wf_ex.state_info or
-            'Cancelled subworkflow [execution_id=%s]' % wf_ex.id
-        )
-
-        result = wf_utils.Result(error=err_msg, cancel=True)
+        env_dict = env_db.variables
     else:
-        raise RuntimeError(
-            "Method _send_result_to_parent_workflow() must never be called"
-            " if a workflow is not in SUCCESS, ERROR or CNCELLED state."
+        raise exc.InputException(
+            'Unexpected value type for environment [env=%s, type=%s]'
+            % (env, type(env))
         )
 
-    rpc.get_engine_client().on_action_complete(
-        wf_ex.id,
-        result,
-        wf_action=True
-    )
+    if ('evaluate_env' in params and
+            not params['evaluate_env']):
+        return env_dict
+    else:
+        return expr.evaluate_recursively(env_dict, {'__env': env_dict})
 
 
 def _build_fail_info_message(wf_ctrl, wf_ex):
