@@ -1,5 +1,6 @@
 # Copyright 2013 - Mirantis, Inc.
 # Copyright 2015 - StackStorm, Inc.
+# Copyright 2016 - Brocade Communications Systems, Inc.
 #
 #    Licensed under the Apache License, Version 2.0 (the "License");
 #    you may not use this file except in compliance with the License.
@@ -14,27 +15,18 @@
 #    limitations under the License.
 
 
-# TODO(rakhmerov): Can we make one parent for errors and exceptions?
+class MistralFailuresBase(Exception):
+    """Base class for mistral errors and exceptions"""
 
-class MistralError(Exception):
-    """Mistral specific error.
+    message = "An unknow failure occured"
 
-    Reserved for situations that can't be automatically handled. When it occurs
-    it signals that there is a major environmental problem like invalid startup
-    configuration or implementation problem (e.g. some code doesn't take care
-    of certain corner cases). From architectural perspective it's pointless to
-    try to handle this type of problems except doing some finalization work
-    like transaction rollback, deleting temporary files etc.
-    """
-
-    message = "An unknown error occurred"
     http_code = 500
 
     def __init__(self, message=None):
         if message is not None:
             self.message = message
 
-        super(MistralError, self).__init__(
+        super(MistralFailuresBase, self).__init__(
             '%d: %s' % (self.http_code, self.message))
 
     @property
@@ -49,7 +41,21 @@ class MistralError(Exception):
         return self.message
 
 
-class MistralException(Exception):
+class MistralError(MistralFailuresBase):
+    """Mistral specific error.
+
+    Reserved for situations that can't be automatically handled. When it occurs
+    it signals that there is a major environmental problem like invalid startup
+    configuration or implementation problem (e.g. some code doesn't take care
+    of certain corner cases). From architectural perspective it's pointless to
+    try to handle this type of problems except doing some finalization work
+    like transaction rollback, deleting temporary files etc.
+    """
+
+    message = "An unknown error occurred"
+
+
+class MistralException(MistralFailuresBase):
     """Mistral specific exception.
 
     Reserved for situations that are not critical for program continuation.
@@ -65,25 +71,6 @@ class MistralException(Exception):
     'http_code' properties.
     """
     message = "An unknown exception occurred"
-    http_code = 500
-
-    def __init__(self, message=None):
-        if message is not None:
-            self.message = message
-
-        super(MistralException, self).__init__(
-            '%d: %s' % (self.http_code, self.message))
-
-    @property
-    def code(self):
-        """This is here for webob to read.
-
-        https://github.com/Pylons/webob/blob/master/webob/exc.py
-        """
-        return self.http_code
-
-    def __str__(self):
-        return self.message
 
 
 # Database errors.
@@ -97,10 +84,6 @@ class DBDuplicateEntryError(DBError):
     message = "Database object already exists"
 
 
-class DBQueryEntryError(DBError):
-    http_code = 400
-
-
 class DBEntityNotFoundError(DBError):
     http_code = 404
     message = "Object not found"
@@ -112,8 +95,15 @@ class DSLParsingException(MistralException):
     http_code = 400
 
 
-class YaqlGrammarException(DSLParsingException):
+class ExpressionGrammarException(DSLParsingException):
     http_code = 400
+
+
+class JinjaGrammarException(ExpressionGrammarException):
+    message = "Invalid grammar of Jinja expression"
+
+
+class YaqlGrammarException(ExpressionGrammarException):
     message = "Invalid grammar of YAQL expression"
 
 
@@ -124,8 +114,15 @@ class InvalidModelException(DSLParsingException):
 
 # Various common exceptions and errors.
 
-class YaqlEvaluationException(MistralException):
+class EvaluationException(MistralException):
     http_code = 400
+
+
+class JinjaEvaluationException(EvaluationException):
+    message = "Can not evaluate Jinja expression"
+
+
+class YaqlEvaluationException(EvaluationException):
     message = "Can not evaluate YAQL expression"
 
 
@@ -174,11 +171,6 @@ class InvalidResultException(MistralException):
 class SizeLimitExceededException(MistralException):
     http_code = 400
 
-    def __init__(self, field_name, size_kb, size_limit_kb):
-        super(SizeLimitExceededException, self).__init__(
-            "Size of '%s' is %dKB which exceeds the limit of %dKB"
-            % (field_name, size_kb, size_limit_kb))
-
 
 class CoordinationException(MistralException):
     http_code = 500
@@ -200,3 +192,8 @@ class KombuException(Exception):
 
         self.exc_type = e.__class__.__name__
         self.value = str(e)
+
+
+class InvalidStateTransitionException(MistralException):
+    http_code = 400
+    message = 'Invalid state transition'
