@@ -1,5 +1,3 @@
-# -*- coding: utf-8 -*-
-#
 # Copyright 2015 - Mirantis, Inc.
 #
 #    Licensed under the Apache License, Version 2.0 (the "License");
@@ -15,43 +13,51 @@
 #    limitations under the License.
 
 from mistral.db.v2 import api as db_api_v2
+from mistral.lang import parser as spec_parser
 from mistral.services import actions
-from mistral.workbook import parser as spec_parser
 
 
-def create_workbook_v2(definition, scope='private'):
-    wb_spec = spec_parser.get_workbook_spec_from_yaml(definition)
+def create_workbook_v2(definition, namespace='', scope='private'):
+    wb_spec = spec_parser.get_workbook_spec_from_yaml(
+        definition, validate=True
+    )
 
     wb_values = _get_workbook_values(
         wb_spec,
         definition,
-        scope
+        scope,
+        namespace
     )
 
     with db_api_v2.transaction():
         wb_db = db_api_v2.create_workbook(wb_values)
 
-        _on_workbook_update(wb_db, wb_spec)
+        _on_workbook_update(wb_db, wb_spec, namespace)
 
     return wb_db
 
 
-def update_workbook_v2(definition, scope='private'):
-    wb_spec = spec_parser.get_workbook_spec_from_yaml(definition)
+def update_workbook_v2(definition, namespace='', scope='private'):
+    wb_spec = spec_parser.get_workbook_spec_from_yaml(
+        definition, validate=True
+    )
 
-    values = _get_workbook_values(wb_spec, definition, scope)
+    values = _get_workbook_values(wb_spec, definition, scope, namespace)
 
     with db_api_v2.transaction():
         wb_db = db_api_v2.update_workbook(values['name'], values)
 
-        _, db_wfs = _on_workbook_update(wb_db, wb_spec)
+        _, db_wfs = _on_workbook_update(wb_db, wb_spec, namespace)
 
     return wb_db
 
 
-def _on_workbook_update(wb_db, wb_spec):
+def _on_workbook_update(wb_db, wb_spec, namespace):
+    # TODO(hardikj) Handle actions for namespace
     db_actions = _create_or_update_actions(wb_db, wb_spec.get_actions())
-    db_wfs = _create_or_update_workflows(wb_db, wb_spec.get_workflows())
+    db_wfs = _create_or_update_workflows(wb_db,
+                                         wb_spec.get_workflows(),
+                                         namespace)
 
     return db_actions, db_wfs
 
@@ -88,7 +94,7 @@ def _create_or_update_actions(wb_db, actions_spec):
     return db_actions
 
 
-def _create_or_update_workflows(wb_db, workflows_spec):
+def _create_or_update_workflows(wb_db, workflows_spec, namespace):
     db_wfs = []
 
     if workflows_spec:
@@ -101,7 +107,9 @@ def _create_or_update_workflows(wb_db, workflows_spec):
                 'spec': wf_spec.to_dict(),
                 'scope': wb_db.scope,
                 'project_id': wb_db.project_id,
-                'tags': wf_spec.get_tags()
+                'namespace': namespace,
+                'tags': wf_spec.get_tags(),
+                'is_system': False
             }
 
             db_wfs.append(
@@ -111,13 +119,15 @@ def _create_or_update_workflows(wb_db, workflows_spec):
     return db_wfs
 
 
-def _get_workbook_values(wb_spec, definition, scope):
+def _get_workbook_values(wb_spec, definition, scope, namespace=None):
     values = {
         'name': wb_spec.get_name(),
         'tags': wb_spec.get_tags(),
         'definition': definition,
         'spec': wb_spec.to_dict(),
-        'scope': scope
+        'scope': scope,
+        'namespace': namespace,
+        'is_system': False
     }
 
     return values
