@@ -15,12 +15,15 @@
 
 import copy
 import datetime
+
 import mock
+import sqlalchemy as sa
 
 from mistral.db.v2 import api as db_api
 from mistral.db.v2.sqlalchemy import models
 from mistral import exceptions as exc
 from mistral.tests.unit.api import base
+from mistral.tests.unit import base as unit_base
 from mistral import utils
 
 WF_DEFINITION = """
@@ -58,6 +61,26 @@ WF = {
     'input': 'param1'
 }
 
+WF_DB_WITHIN_ABC_NAMESPACE = models.WorkflowDefinition(
+    id='234560fe-162a-4060-a16a-a0d9eee9b408',
+    name='flow',
+    namespace='abc',
+    definition=WF_DEFINITION,
+    created_at=datetime.datetime(1970, 1, 1),
+    updated_at=datetime.datetime(1970, 1, 1),
+    spec={'input': ['param1']}
+)
+
+WF_WITH_NAMESPACE = {
+    'id': '234560fe-162a-4060-a16a-a0d9eee9b408',
+    'name': 'flow',
+    'namespace': 'abc',
+    'definition': WF_DEFINITION,
+    'created_at': '1970-01-01 00:00:00',
+    'updated_at': '1970-01-01 00:00:00',
+    'input': 'param1'
+}
+
 WF_DEFINITION_WITH_INPUT = """
 ---
 version: '2.0'
@@ -89,6 +112,9 @@ WF_WITH_DEFAULT_INPUT = {
     'input': 'param1, param2=2'
 }
 
+WF_DB_PROJECT_ID = WF_DB.get_clone()
+WF_DB_PROJECT_ID.project_id = '<default-project>'
+
 UPDATED_WF_DEFINITION = """
 ---
 version: '2.0'
@@ -172,6 +198,19 @@ class TestWorkflowsController(base.APITest):
         self.assertEqual(200, resp.status_int)
         self.assertDictEqual(WF, resp.json)
 
+    @mock.patch.object(db_api, 'get_workflow_definition')
+    def test_get_operational_error(self, mocked_get):
+        mocked_get.side_effect = [
+            # Emulating DB OperationalError
+            sa.exc.OperationalError('Mock', 'mock', 'mock'),
+            WF_DB  # Successful run
+        ]
+
+        resp = self.app.get('/v2/workflows/123')
+
+        self.assertEqual(200, resp.status_int)
+        self.assertDictEqual(WF, resp.json)
+
     @mock.patch.object(db_api, "get_workflow_definition", MOCK_WF_WITH_INPUT)
     def test_get_with_input(self):
         resp = self.app.get('/v2/workflows/123')
@@ -216,7 +255,8 @@ class TestWorkflowsController(base.APITest):
         update_mock.assert_called_once_with(
             UPDATED_WF_DEFINITION,
             scope='private',
-            identifier='123e4567-e89b-12d3-a456-426655440000'
+            identifier='123e4567-e89b-12d3-a456-426655440000',
+            namespace=''
         )
         self.assertDictEqual(UPDATED_WF, resp.json)
 
@@ -234,7 +274,7 @@ class TestWorkflowsController(base.APITest):
 
         self.assertEqual(400, resp.status_int)
         self.assertIn(
-            "Attempt to modify a system workflow",
+            "Can not modify a system",
             resp.body.decode()
         )
 
@@ -253,6 +293,17 @@ class TestWorkflowsController(base.APITest):
 
         self.assertEqual("public", mock_update.call_args[0][1]['scope'])
 
+    def test_put_wrong_scope(self):
+        resp = self.app.put(
+            '/v2/workflows?scope=unique',
+            UPDATED_WF_DEFINITION,
+            headers={'Content-Type': 'text/plain'},
+            expect_errors=True
+        )
+
+        self.assertEqual(400, resp.status_int)
+        self.assertIn("Scope must be one of the following", resp.body.decode())
+
     @mock.patch.object(
         db_api, "update_workflow_definition", MOCK_WF_WITH_INPUT
     )
@@ -341,10 +392,7 @@ class TestWorkflowsController(base.APITest):
 
         self.assertEqual("public", mock_mtd.call_args[0][0]['scope'])
 
-    @mock.patch.object(db_api, "create_action_definition")
-    def test_post_wrong_scope(self, mock_mtd):
-        mock_mtd.return_value = WF_DB
-
+    def test_post_wrong_scope(self):
         resp = self.app.post(
             '/v2/workflows?scope=unique',
             WF_DEFINITION,
@@ -393,7 +441,7 @@ class TestWorkflowsController(base.APITest):
 
         self.assertEqual(400, resp.status_int)
         self.assertIn(
-            "Attempt to delete a system workflow",
+            "Can not modify a system",
             resp.body.decode()
         )
 
@@ -412,6 +460,21 @@ class TestWorkflowsController(base.APITest):
         self.assertEqual(1, len(resp.json['workflows']))
         self.assertDictEqual(WF, resp.json['workflows'][0])
 
+    @mock.patch.object(db_api, 'get_workflow_definitions')
+    def test_get_all_operational_error(self, mocked_get_all):
+        mocked_get_all.side_effect = [
+            # Emulating DB OperationalError
+            sa.exc.OperationalError('Mock', 'mock', 'mock'),
+            [WF_DB]  # Successful run
+        ]
+
+        resp = self.app.get('/v2/workflows')
+
+        self.assertEqual(200, resp.status_int)
+
+        self.assertEqual(1, len(resp.json['workflows']))
+        self.assertDictEqual(WF, resp.json['workflows'][0])
+
     @mock.patch.object(db_api, "get_workflow_definitions", MOCK_EMPTY)
     def test_get_all_empty(self):
         resp = self.app.get('/v2/workflows')
@@ -420,6 +483,26 @@ class TestWorkflowsController(base.APITest):
 
         self.assertEqual(0, len(resp.json['workflows']))
 
+    @mock.patch('mistral.db.v2.api.get_workflow_definitions')
+    @mock.patch('mistral.context.MistralContext.from_environ')
+    def test_get_all_projects_admin(self, mock_context, mock_get_wf_defs):
+        admin_ctx = unit_base.get_context(admin=True)
+        mock_context.return_value = admin_ctx
+
+        resp = self.app.get('/v2/workflows?all_projects=true')
+
+        self.assertEqual(200, resp.status_int)
+
+        self.assertTrue(mock_get_wf_defs.call_args[1].get('insecure', False))
+
+    def test_get_all_projects_normal_user(self):
+        resp = self.app.get(
+            '/v2/workflows?all_projects=true',
+            expect_errors=True
+        )
+
+        self.assertEqual(403, resp.status_int)
+
     @mock.patch.object(db_api, "get_workflow_definitions", MOCK_WFS)
     def test_get_all_pagination(self):
         resp = self.app.get(
@@ -507,6 +590,32 @@ class TestWorkflowsController(base.APITest):
 
         self.assertDictEqual(expected_dict, resp.json['workflows'][0])
 
+    @mock.patch('mistral.db.v2.api.get_workflow_definitions')
+    def test_get_all_with_fields_input_filter(self, mock_get_db_wfs):
+        def mock_get_defintions(fields=None, session=None, **kwargs):
+            if fields and 'input' in fields:
+                fields.remove('input')
+                fields.append('spec')
+
+            return [
+                ('65df1f59-938f-4c17-bc2a-562524ef5e40',
+                 {'input': ['param1', {'param2': 2}]})
+            ]
+
+        mock_get_db_wfs.side_effect = mock_get_defintions
+
+        resp = self.app.get('/v2/workflows?fields=input')
+
+        self.assertEqual(200, resp.status_int)
+        self.assertEqual(1, len(resp.json['workflows']))
+
+        expected_dict = {
+            'id': '65df1f59-938f-4c17-bc2a-562524ef5e40',
+            'input': 'param1, param2=2'
+        }
+
+        self.assertDictEqual(expected_dict, resp.json['workflows'][0])
+
     def test_get_all_with_invalid_field(self):
         resp = self.app.get(
             '/v2/workflows?fields=name,nonexist',
@@ -578,3 +687,53 @@ class TestWorkflowsController(base.APITest):
         self.assertEqual(200, resp.status_int)
         self.assertFalse(resp.json['valid'])
         self.assertIn("Invalid DSL", resp.json['error'])
+
+    @mock.patch("mistral.services.workflows.update_workflows")
+    @mock.patch.object(db_api, "create_workflow_definition")
+    def test_workflow_within_namespace(self, mock_mtd, update_mock):
+        mock_mtd.return_value = WF_DB_WITHIN_ABC_NAMESPACE
+
+        namespace = 'abc'
+        resp = self.app.post(
+            '/v2/workflows?namespace=%s' % namespace,
+            WF_DEFINITION,
+            headers={'Content-Type': 'text/plain'}
+        )
+
+        self.assertEqual(201, resp.status_int)
+        self.assertDictEqual({'workflows': [WF_WITH_NAMESPACE]}, resp.json)
+
+        self.assertEqual(1, mock_mtd.call_count)
+
+        spec = mock_mtd.call_args[0][0]['spec']
+
+        self.assertIsNotNone(spec)
+        self.assertEqual(WF_DB.name, spec['name'])
+        self.assertEqual(WF_DB_WITHIN_ABC_NAMESPACE.namespace, namespace)
+
+        update_mock.return_value = [WF_DB_WITHIN_ABC_NAMESPACE]
+
+        id_ = '234560fe-162a-4060-a16a-a0d9eee9b408'
+
+        resp = self.app.put(
+            '/v2/workflows/%s?namespace=%s' % (id_, namespace),
+            WF_DEFINITION,
+            headers={'Content-Type': 'text/plain'}
+        )
+
+        self.assertEqual(200, resp.status_int)
+        update_mock.assert_called_once_with(
+            WF_DEFINITION,
+            scope='private',
+            identifier=id_,
+            namespace='abc'
+        )
+        self.assertDictEqual(WF_WITH_NAMESPACE, resp.json)
+
+    @mock.patch.object(db_api, "get_workflow_definition")
+    def test_workflow_within_project_id(self, mock_get):
+        mock_get.return_value = WF_DB_PROJECT_ID
+        resp = self.app.get(
+            '/v2/workflows/123e4567-e89b-12d3-a456-426655440000')
+        self.assertEqual(200, resp.status_int)
+        self.assertTrue('project_id' in resp.json)
