@@ -23,19 +23,19 @@ import mock
 from oslo_config import cfg
 from oslo_log import log as logging
 from oslotest import base
-import six
 import testtools.matchers as ttm
 
+from mistral import config
 from mistral import context as auth_context
 from mistral.db.sqlalchemy import base as db_sa_base
 from mistral.db.sqlalchemy import sqlite_lock
 from mistral.db.v2 import api as db_api
+from mistral.lang import parser as spec_parser
 from mistral.services import action_manager
 from mistral.services import security
 from mistral.tests.unit import config as test_config
 from mistral.utils import inspect_utils as i_utils
 from mistral import version
-from mistral.workbook import parser as spec_parser
 from mistral.workflow import lookup_utils
 
 RESOURCES_PATH = 'tests/resources/'
@@ -53,21 +53,21 @@ def get_resource(resource_name):
 
 def get_context(default=True, admin=False):
     if default:
-        return auth_context.MistralContext(
-            user_id='1-2-3-4',
-            project_id=security.DEFAULT_PROJECT_ID,
-            user_name='test-user',
-            project_name='test-project',
-            is_admin=admin
-        )
+        return auth_context.MistralContext.from_dict({
+            'user_name': 'test-user',
+            'user': '1-2-3-4',
+            'tenant': security.DEFAULT_PROJECT_ID,
+            'project_name': 'test-project',
+            'is_admin': admin
+        })
     else:
-        return auth_context.MistralContext(
-            user_id='9-0-44-5',
-            project_id='99-88-33',
-            user_name='test-user',
-            project_name='test-another',
-            is_admin=admin
-        )
+        return auth_context.MistralContext.from_dict({
+            'user_name': 'test-user',
+            'user': '9-0-44-5',
+            'tenant': '99-88-33',
+            'project_name': 'test-another',
+            'is_admin': admin
+        })
 
 
 def register_action_class(name, cls, attributes=None, desc=None):
@@ -81,7 +81,7 @@ def register_action_class(name, cls, attributes=None, desc=None):
 
 class FakeHTTPResponse(object):
     def __init__(self, text, status_code, reason=None, headers=None,
-                 history=None, encoding='utf8', url='', cookies=None,
+                 history=None, encoding='utf-8', url='', cookies=None,
                  elapsed=None):
         self.text = text
         self.content = text
@@ -94,8 +94,8 @@ class FakeHTTPResponse(object):
         self.cookies = cookies or {}
         self.elapsed = elapsed or datetime.timedelta(milliseconds=123)
 
-    def json(self):
-        return json.loads(self.text)
+    def json(self, **kwargs):
+        return json.loads(self.text, **kwargs)
 
 
 class BaseTest(base.BaseTestCase):
@@ -104,6 +104,25 @@ class BaseTest(base.BaseTestCase):
 
         self.addCleanup(spec_parser.clear_caches)
 
+    def register_action_class(self, name, cls, attributes=None, desc=None):
+        # Added for convenience (to avoid unnecessary imports).
+        register_action_class(name, cls, attributes, desc)
+
+    def assertRaisesWithMessage(self, exception, msg, func, *args, **kwargs):
+        try:
+            func(*args, **kwargs)
+            self.assertFail()
+        except exception as e:
+            self.assertEqual(msg, e.message)
+
+    def assertRaisesWithMessageContaining(self, exception, msg, func, *args,
+                                          **kwargs):
+        try:
+            func(*args, **kwargs)
+            self.assertFail()
+        except exception as e:
+            self.assertIn(msg, e.message)
+
     def assertListEqual(self, l1, l2):
         if tuple(sys.version_info)[0:2] < (2, 7):
             # for python 2.6 compatibility
@@ -123,7 +142,7 @@ class BaseTest(base.BaseTestCase):
 
     def _assert_multiple_items(self, items, count, **props):
         def _matches(item, **props):
-            for prop_name, prop_val in six.iteritems(props):
+            for prop_name, prop_val in props.items():
                 v = item[prop_name] if isinstance(
                     item, dict) else getattr(item, prop_name)
 
@@ -133,13 +152,13 @@ class BaseTest(base.BaseTestCase):
             return True
 
         filtered_items = list(
-            filter(lambda item: _matches(item, **props), items)
+            [item for item in items if _matches(item, **props)]
         )
 
         found = len(filtered_items)
 
         if found != count:
-            LOG.info("[failed test ctx] items=%s, expected_props=%s" % (str(
+            LOG.info("[failed test ctx] items=%s, expected_props=%s", (str(
                 items), props))
             self.fail("Wrong number of items found [props=%s, "
                       "expected=%s, found=%s]" % (props, count, found))
@@ -156,7 +175,7 @@ class BaseTest(base.BaseTestCase):
         missing = []
         mismatched = []
 
-        for key, value in six.iteritems(expected):
+        for key, value in expected.items():
             if key not in actual:
                 missing.append(key)
             elif value != actual[key]:
@@ -178,7 +197,8 @@ class BaseTest(base.BaseTestCase):
 
         self.fail(self._formatMessage(msg, standardMsg))
 
-    def _await(self, predicate, delay=1, timeout=60):
+    def _await(self, predicate, delay=1, timeout=60, fail_message="no detail",
+               fail_message_formatter=lambda x: x):
         """Awaits for predicate function to evaluate to True.
 
         If within a configured timeout predicate function hasn't evaluated
@@ -187,6 +207,8 @@ class BaseTest(base.BaseTestCase):
         :param delay: Delay in seconds between predicate function calls.
         :param timeout: Maximum amount of time to wait for predication
             function to evaluate to True.
+        :param fail_message: explains what was expected
+        :param fail_message_formatter: lambda that formats the fail_message
         :return:
         """
         end_time = time.time() + timeout
@@ -196,7 +218,10 @@ class BaseTest(base.BaseTestCase):
                 break
 
             if time.time() + delay > end_time:
-                raise AssertionError("Failed to wait for expected result.")
+                raise AssertionError(
+                    "Failed to wait for expected result: " +
+                    fail_message_formatter(fail_message)
+                )
 
             time.sleep(delay)
 
@@ -206,6 +231,7 @@ class BaseTest(base.BaseTestCase):
     def override_config(self, name, override, group=None):
         """Cleanly override CONF variables."""
         cfg.CONF.set_override(name, override, group)
+
         self.addCleanup(cfg.CONF.clear_override, name, group)
 
 
@@ -237,6 +263,14 @@ class DbTestCase(BaseTest):
         if cfg.CONF.database.connection.startswith('sqlite'):
             cfg.CONF.set_default('connection', 'sqlite://', group='database')
 
+        # This option is normally registered in sync_db.py so we have to
+        # register it here specifically for tests.
+        cfg.CONF.register_opt(config.os_actions_mapping_path)
+
+        cfg.CONF.set_default(
+            'openstack_actions_mapping_path',
+            'tests/resources/openstack/test_mapping.json'
+        )
         cfg.CONF.set_default('max_overflow', -1, group='database')
         cfg.CONF.set_default('max_pool_size', 1000, group='database')
 
@@ -245,7 +279,7 @@ class DbTestCase(BaseTest):
         action_manager.sync_db()
 
     def _clean_db(self):
-        lookup_utils.clean_caches()
+        lookup_utils.clear_caches()
 
         contexts = [
             get_context(default=False),
