@@ -42,8 +42,8 @@ EXPECTED_ENV_AUTH = ('librarian', 'password123')
 WORKFLOW1 = """
 ---
 version: "2.0"
+
 wf1:
-  type: direct
   tasks:
     task1:
       action: std.http url="https://api.library.org/books"
@@ -54,8 +54,8 @@ wf1:
 WORKFLOW2 = """
 ---
 version: "2.0"
+
 wf2:
-  type: direct
   tasks:
     task1:
       action: std.http url="https://api.library.org/books" timeout=60
@@ -66,10 +66,11 @@ wf2:
 WORKFLOW1_WITH_ITEMS = """
 ---
 version: "2.0"
+
 wf1_with_items:
-  type: direct
   input:
     - links
+
   tasks:
     task1:
       with-items: link in <% $.links %>
@@ -81,10 +82,11 @@ wf1_with_items:
 WORKFLOW2_WITH_ITEMS = """
 ---
 version: "2.0"
+
 wf2_with_items:
-  type: direct
   input:
     - links
+
   tasks:
     task1:
       with-items: link in <% $.links %>
@@ -95,7 +97,6 @@ wf2_with_items:
 
 
 class ActionDefaultTest(base.EngineTestCase):
-
     @mock.patch.object(
         requests, 'request',
         mock.MagicMock(return_value=test_base.FakeHTTPResponse('', 200, 'OK')))
@@ -105,21 +106,29 @@ class ActionDefaultTest(base.EngineTestCase):
     def test_action_defaults_from_env(self):
         wf_service.create_workflows(WORKFLOW1)
 
-        wf_ex = self.engine.start_workflow('wf1', None, env=ENV)
+        wf_ex = self.engine.start_workflow('wf1', env=ENV)
 
         self.await_workflow_success(wf_ex.id)
 
-        wf_ex = db_api.get_workflow_execution(wf_ex.id)
+        with db_api.transaction():
+            wf_ex = db_api.get_workflow_execution(wf_ex.id)
 
-        self.assertEqual(states.SUCCESS, wf_ex.state)
-        self._assert_single_item(wf_ex.task_executions, name='task1')
+            self.assertEqual(states.SUCCESS, wf_ex.state)
+            self._assert_single_item(wf_ex.task_executions, name='task1')
 
         requests.request.assert_called_with(
-            'GET', 'https://api.library.org/books',
-            params=None, data=None, headers=None, cookies=None,
-            allow_redirects=None, proxies=None, verify=None,
+            'GET',
+            'https://api.library.org/books',
+            params=None,
+            data=None,
+            headers=None,
+            cookies=None,
+            allow_redirects=None,
+            proxies=None,
+            verify=None,
             auth=EXPECTED_ENV_AUTH,
-            timeout=ENV['__actions']['std.http']['timeout'])
+            timeout=ENV['__actions']['std.http']['timeout']
+        )
 
     @mock.patch.object(
         requests, 'request',
@@ -130,14 +139,15 @@ class ActionDefaultTest(base.EngineTestCase):
     def test_action_defaults_from_env_not_applied(self):
         wf_service.create_workflows(WORKFLOW2)
 
-        wf_ex = self.engine.start_workflow('wf2', None, env=ENV)
+        wf_ex = self.engine.start_workflow('wf2', env=ENV)
 
         self.await_workflow_success(wf_ex.id)
 
-        wf_ex = db_api.get_workflow_execution(wf_ex.id)
+        with db_api.transaction():
+            wf_ex = db_api.get_workflow_execution(wf_ex.id)
 
-        self.assertEqual(states.SUCCESS, wf_ex.state)
-        self._assert_single_item(wf_ex.task_executions, name='task1')
+            self.assertEqual(states.SUCCESS, wf_ex.state)
+            self._assert_single_item(wf_ex.task_executions, name='task1')
 
         requests.request.assert_called_with(
             'GET', 'https://api.library.org/books',
@@ -165,23 +175,24 @@ class ActionDefaultTest(base.EngineTestCase):
 
         wf_ex = self.engine.start_workflow(
             'wf1_with_items',
-            wf_input,
+            wf_input=wf_input,
             env=ENV
         )
 
         self.await_workflow_success(wf_ex.id)
 
-        wf_ex = db_api.get_workflow_execution(wf_ex.id)
+        with db_api.transaction():
+            wf_ex = db_api.get_workflow_execution(wf_ex.id)
 
-        self.assertEqual(states.SUCCESS, wf_ex.state)
-        self._assert_single_item(wf_ex.task_executions, name='task1')
+            self.assertEqual(states.SUCCESS, wf_ex.state)
+            self._assert_single_item(wf_ex.task_executions, name='task1')
 
-        calls = [mock.call('GET', url, params=None, data=None,
-                           headers=None, cookies=None,
-                           allow_redirects=None, proxies=None,
-                           auth=EXPECTED_ENV_AUTH, verify=None,
-                           timeout=ENV['__actions']['std.http']['timeout'])
-                 for url in wf_input['links']]
+            calls = [mock.call('GET', url, params=None, data=None,
+                               headers=None, cookies=None,
+                               allow_redirects=None, proxies=None,
+                               auth=EXPECTED_ENV_AUTH, verify=None,
+                               timeout=ENV['__actions']['std.http']['timeout'])
+                     for url in wf_input['links']]
 
         requests.request.assert_has_calls(calls, any_order=True)
 
@@ -203,16 +214,17 @@ class ActionDefaultTest(base.EngineTestCase):
 
         wf_ex = self.engine.start_workflow(
             'wf2_with_items',
-            wf_input,
+            wf_input=wf_input,
             env=ENV
         )
 
         self.await_workflow_success(wf_ex.id)
 
-        wf_ex = db_api.get_workflow_execution(wf_ex.id)
+        with db_api.transaction():
+            wf_ex = db_api.get_workflow_execution(wf_ex.id)
 
-        self.assertEqual(states.SUCCESS, wf_ex.state)
-        self._assert_single_item(wf_ex.task_executions, name='task1')
+            self.assertEqual(states.SUCCESS, wf_ex.state)
+            self._assert_single_item(wf_ex.task_executions, name='task1')
 
         calls = [mock.call('GET', url, params=None, data=None,
                            headers=None, cookies=None,
