@@ -58,47 +58,78 @@ class SimpleEngineCommandsTest(base.EngineTestCase):
         wb_service.create_workbook_v2(WORKBOOK1)
 
     def test_fail(self):
-        wf_ex = self.engine.start_workflow('my_wb.wf', {'my_var': 1})
+        wf_ex = self.engine.start_workflow('my_wb.wf', wf_input={'my_var': 1})
 
         self.await_workflow_error(wf_ex.id)
 
-        wf_ex = db_api.get_workflow_execution(wf_ex.id)
+        with db_api.transaction():
+            wf_ex = db_api.get_workflow_execution(wf_ex.id)
 
-        self.assertEqual(1, len(wf_ex.task_executions))
+            task_execs = wf_ex.task_executions
+
+        self.assertEqual(1, len(task_execs))
         self._assert_single_item(
-            wf_ex.task_executions,
+            task_execs,
             name='task1',
             state=states.SUCCESS
         )
 
     def test_succeed(self):
-        wf_ex = self.engine.start_workflow('my_wb.wf', {'my_var': 2})
+        wf_ex = self.engine.start_workflow('my_wb.wf', wf_input={'my_var': 2})
 
         self.await_workflow_success(wf_ex.id)
 
-        wf_ex = db_api.get_workflow_execution(wf_ex.id)
+        with db_api.transaction():
+            wf_ex = db_api.get_workflow_execution(wf_ex.id)
 
-        self.assertEqual(1, len(wf_ex.task_executions))
+            task_execs = wf_ex.task_executions
+
+        self.assertEqual(1, len(task_execs))
         self._assert_single_item(
-            wf_ex.task_executions,
+            task_execs,
             name='task1',
             state=states.SUCCESS
         )
 
     def test_pause(self):
-        wf_ex = self.engine.start_workflow('my_wb.wf', {'my_var': 3})
+        wf_ex = self.engine.start_workflow('my_wb.wf', wf_input={'my_var': 3})
 
         self.await_workflow_paused(wf_ex.id)
 
-        wf_ex = db_api.get_workflow_execution(wf_ex.id)
+        with db_api.transaction():
+            wf_ex = db_api.get_workflow_execution(wf_ex.id)
 
-        self.assertEqual(1, len(wf_ex.task_executions))
+            task_execs = wf_ex.task_executions
+
+        self.assertEqual(1, len(task_execs))
         self._assert_single_item(
-            wf_ex.task_executions,
+            task_execs,
             name='task1',
             state=states.SUCCESS
         )
 
+        # Let's resume the workflow and wait till it succeeds.
+        self.engine.resume_workflow(wf_ex.id)
+
+        self.await_workflow_success(wf_ex.id)
+
+        with db_api.transaction():
+            wf_ex = db_api.get_workflow_execution(wf_ex.id)
+
+            task_execs = wf_ex.task_executions
+
+        self.assertEqual(2, len(task_execs))
+        self._assert_single_item(
+            task_execs,
+            name='task1',
+            state=states.SUCCESS
+        )
+        self._assert_single_item(
+            task_execs,
+            name='task2',
+            state=states.SUCCESS
+        )
+
 
 WORKBOOK2 = """
 ---
@@ -135,43 +166,52 @@ class SimpleEngineWorkflowLevelCommandsTest(base.EngineTestCase):
         wb_service.create_workbook_v2(WORKBOOK2)
 
     def test_fail(self):
-        wf_ex = self.engine.start_workflow('my_wb.wf', {'my_var': 1})
+        wf_ex = self.engine.start_workflow('my_wb.wf', wf_input={'my_var': 1})
 
         self.await_workflow_error(wf_ex.id)
 
-        wf_ex = db_api.get_workflow_execution(wf_ex.id)
+        with db_api.transaction():
+            wf_ex = db_api.get_workflow_execution(wf_ex.id)
 
-        self.assertEqual(1, len(wf_ex.task_executions))
+            task_execs = wf_ex.task_executions
+
+        self.assertEqual(1, len(task_execs))
         self._assert_single_item(
-            wf_ex.task_executions,
+            task_execs,
             name='task1',
             state=states.SUCCESS
         )
 
     def test_succeed(self):
-        wf_ex = self.engine.start_workflow('my_wb.wf', {'my_var': 2})
+        wf_ex = self.engine.start_workflow('my_wb.wf', wf_input={'my_var': 2})
 
         self.await_workflow_success(wf_ex.id)
 
-        wf_ex = db_api.get_workflow_execution(wf_ex.id)
+        with db_api.transaction():
+            wf_ex = db_api.get_workflow_execution(wf_ex.id)
 
-        self.assertEqual(1, len(wf_ex.task_executions))
+            task_execs = wf_ex.task_executions
+
+        self.assertEqual(1, len(task_execs))
         self._assert_single_item(
-            wf_ex.task_executions,
+            task_execs,
             name='task1',
             state=states.SUCCESS
         )
 
     def test_pause(self):
-        wf_ex = self.engine.start_workflow('my_wb.wf', {'my_var': 3})
+        wf_ex = self.engine.start_workflow('my_wb.wf', wf_input={'my_var': 3})
 
         self.await_workflow_paused(wf_ex.id)
 
-        wf_ex = db_api.get_workflow_execution(wf_ex.id)
+        with db_api.transaction():
+            wf_ex = db_api.get_workflow_execution(wf_ex.id)
 
-        self.assertEqual(1, len(wf_ex.task_executions))
+            task_execs = wf_ex.task_executions
+
+        self.assertEqual(1, len(task_execs))
         self._assert_single_item(
-            wf_ex.task_executions,
+            task_execs,
             name='task1',
             state=states.SUCCESS
         )
@@ -245,71 +285,77 @@ class OrderEngineCommandsTest(base.EngineTestCase):
         wb_service.create_workbook_v2(WORKBOOK3)
 
     def test_fail_first(self):
-        wf_ex = self.engine.start_workflow('my_wb.fail_first_wf', None)
+        wf_ex = self.engine.start_workflow('my_wb.fail_first_wf')
 
         self.await_workflow_error(wf_ex.id)
 
-        wf_ex = db_api.get_workflow_execution(wf_ex.id)
+        with db_api.transaction():
+            wf_ex = db_api.get_workflow_execution(wf_ex.id)
 
-        self.assertEqual(1, len(wf_ex.task_executions))
+            task_execs = wf_ex.task_executions
+
+        self.assertEqual(1, len(task_execs))
         self._assert_single_item(
-            wf_ex.task_executions,
+            task_execs,
             name='task1',
             state=states.SUCCESS
         )
 
     def test_fail_second(self):
-        wf_ex = self.engine.start_workflow('my_wb.fail_second_wf', None)
+        wf_ex = self.engine.start_workflow('my_wb.fail_second_wf')
 
         self.await_workflow_error(wf_ex.id)
 
-        wf_ex = db_api.get_workflow_execution(wf_ex.id)
+        with db_api.transaction():
+            wf_ex = db_api.get_workflow_execution(wf_ex.id)
 
-        self.assertEqual(2, len(wf_ex.task_executions))
+            task_execs = wf_ex.task_executions
+
+        self.assertEqual(2, len(task_execs))
         self._assert_single_item(
-            wf_ex.task_executions,
+            task_execs,
             name='task1',
             state=states.SUCCESS
         )
-        task2_db = self._assert_single_item(
-            wf_ex.task_executions,
-            name='task2'
-        )
+        task2_db = self._assert_single_item(task_execs, name='task2')
 
         self.await_task_success(task2_db.id)
         self.await_workflow_error(wf_ex.id)
 
     def test_succeed_first(self):
-        wf_ex = self.engine.start_workflow('my_wb.succeed_first_wf', None)
+        wf_ex = self.engine.start_workflow('my_wb.succeed_first_wf')
 
         self.await_workflow_success(wf_ex.id)
 
-        wf_ex = db_api.get_workflow_execution(wf_ex.id)
+        with db_api.transaction():
+            wf_ex = db_api.get_workflow_execution(wf_ex.id)
 
-        self.assertEqual(1, len(wf_ex.task_executions))
+            task_execs = wf_ex.task_executions
+
+        self.assertEqual(1, len(task_execs))
         self._assert_single_item(
-            wf_ex.task_executions,
+            task_execs,
             name='task1',
             state=states.SUCCESS
         )
 
     def test_succeed_second(self):
-        wf_ex = self.engine.start_workflow('my_wb.succeed_second_wf', None)
+        wf_ex = self.engine.start_workflow('my_wb.succeed_second_wf')
 
         self.await_workflow_success(wf_ex.id)
 
-        wf_ex = db_api.get_workflow_execution(wf_ex.id)
+        with db_api.transaction():
+            wf_ex = db_api.get_workflow_execution(wf_ex.id)
 
-        self.assertEqual(2, len(wf_ex.task_executions))
+            task_execs = wf_ex.task_executions
+
+        self.assertEqual(2, len(task_execs))
         self._assert_single_item(
-            wf_ex.task_executions,
+            task_execs,
             name='task1',
             state=states.SUCCESS
         )
-        task2_db = self._assert_single_item(
-            wf_ex.task_executions,
-            name='task2'
-        )
+        task2_db = self._assert_single_item(task_execs, name='task2')
 
         self.await_task_error(task2_db.id)
         self.await_workflow_success(wf_ex.id)
@@ -347,15 +393,18 @@ class SimpleEngineCmdsWithMsgTest(base.EngineTestCase):
         wb_service.create_workbook_v2(WORKBOOK4)
 
     def test_fail(self):
-        wf_ex = self.engine.start_workflow('my_wb.wf', {'my_var': 1})
+        wf_ex = self.engine.start_workflow('my_wb.wf', wf_input={'my_var': 1})
 
         self.await_workflow_error(wf_ex.id)
 
-        wf_ex = db_api.get_workflow_execution(wf_ex.id)
+        with db_api.transaction():
+            wf_ex = db_api.get_workflow_execution(wf_ex.id)
 
-        self.assertEqual(1, len(wf_ex.task_executions))
+            task_execs = wf_ex.task_executions
+
+        self.assertEqual(1, len(task_execs))
         self._assert_single_item(
-            wf_ex.task_executions,
+            task_execs,
             name='task1',
             state=states.SUCCESS
         )
@@ -363,15 +412,18 @@ class SimpleEngineCmdsWithMsgTest(base.EngineTestCase):
         self.assertEqual('my_var value is 1', wf_ex.state_info)
 
     def test_succeed(self):
-        wf_ex = self.engine.start_workflow('my_wb.wf', {'my_var': 2})
+        wf_ex = self.engine.start_workflow('my_wb.wf', wf_input={'my_var': 2})
 
         self.await_workflow_success(wf_ex.id)
 
-        wf_ex = db_api.get_workflow_execution(wf_ex.id)
+        with db_api.transaction():
+            wf_ex = db_api.get_workflow_execution(wf_ex.id)
 
-        self.assertEqual(1, len(wf_ex.task_executions))
+            task_execs = wf_ex.task_executions
+
+        self.assertEqual(1, len(task_execs))
         self._assert_single_item(
-            wf_ex.task_executions,
+            task_execs,
             name='task1',
             state=states.SUCCESS
         )
@@ -379,15 +431,18 @@ class SimpleEngineCmdsWithMsgTest(base.EngineTestCase):
         self.assertEqual("my_var value is 2", wf_ex.state_info)
 
     def test_pause(self):
-        wf_ex = self.engine.start_workflow('my_wb.wf', {'my_var': 3})
+        wf_ex = self.engine.start_workflow('my_wb.wf', wf_input={'my_var': 3})
 
         self.await_workflow_paused(wf_ex.id)
 
-        wf_ex = db_api.get_workflow_execution(wf_ex.id)
+        with db_api.transaction():
+            wf_ex = db_api.get_workflow_execution(wf_ex.id)
 
-        self.assertEqual(1, len(wf_ex.task_executions))
+            task_execs = wf_ex.task_executions
+
+        self.assertEqual(1, len(task_execs))
         self._assert_single_item(
-            wf_ex.task_executions,
+            task_execs,
             name='task1',
             state=states.SUCCESS
         )
@@ -429,15 +484,18 @@ class SimpleEngineWorkflowLevelCmdsWithMsgTest(base.EngineTestCase):
         wb_service.create_workbook_v2(WORKBOOK5)
 
     def test_fail(self):
-        wf_ex = self.engine.start_workflow('my_wb.wf', {'my_var': 1})
+        wf_ex = self.engine.start_workflow('my_wb.wf', wf_input={'my_var': 1})
 
         self.await_workflow_error(wf_ex.id)
 
-        wf_ex = db_api.get_workflow_execution(wf_ex.id)
+        with db_api.transaction():
+            wf_ex = db_api.get_workflow_execution(wf_ex.id)
 
-        self.assertEqual(1, len(wf_ex.task_executions))
+            task_execs = wf_ex.task_executions
+
+        self.assertEqual(1, len(task_execs))
         self._assert_single_item(
-            wf_ex.task_executions,
+            task_execs,
             name='task1',
             state=states.SUCCESS
         )
@@ -445,15 +503,18 @@ class SimpleEngineWorkflowLevelCmdsWithMsgTest(base.EngineTestCase):
         self.assertEqual("my_var value is 1", wf_ex.state_info)
 
     def test_succeed(self):
-        wf_ex = self.engine.start_workflow('my_wb.wf', {'my_var': 2})
+        wf_ex = self.engine.start_workflow('my_wb.wf', wf_input={'my_var': 2})
 
         self.await_workflow_success(wf_ex.id)
 
-        wf_ex = db_api.get_workflow_execution(wf_ex.id)
+        with db_api.transaction():
+            wf_ex = db_api.get_workflow_execution(wf_ex.id)
 
-        self.assertEqual(1, len(wf_ex.task_executions))
+            task_execs = wf_ex.task_executions
+
+        self.assertEqual(1, len(task_execs))
         self._assert_single_item(
-            wf_ex.task_executions,
+            task_execs,
             name='task1',
             state=states.SUCCESS
         )
@@ -461,15 +522,18 @@ class SimpleEngineWorkflowLevelCmdsWithMsgTest(base.EngineTestCase):
         self.assertEqual("my_var value is 2", wf_ex.state_info)
 
     def test_pause(self):
-        wf_ex = self.engine.start_workflow('my_wb.wf', {'my_var': 3})
+        wf_ex = self.engine.start_workflow('my_wb.wf', wf_input={'my_var': 3})
 
         self.await_workflow_paused(wf_ex.id)
 
-        wf_ex = db_api.get_workflow_execution(wf_ex.id)
+        with db_api.transaction():
+            wf_ex = db_api.get_workflow_execution(wf_ex.id)
 
-        self.assertEqual(1, len(wf_ex.task_executions))
+            task_execs = wf_ex.task_executions
+
+        self.assertEqual(1, len(task_execs))
         self._assert_single_item(
-            wf_ex.task_executions,
+            task_execs,
             name='task1',
             state=states.SUCCESS
         )
