@@ -16,7 +16,9 @@ import datetime
 import mock
 from oslo_config import cfg
 
+from mistral import context as auth_ctx
 from mistral.db.v2 import api as db_api
+from mistral import exceptions as exc
 from mistral.services import periodic
 from mistral.services import security
 from mistral.services import triggers
@@ -42,7 +44,8 @@ class ProcessCronTriggerTest(base.EngineTestCase):
     @mock.patch.object(security,
                        'create_trust',
                        type('trust', (object,), {'id': 'my_trust_id'}))
-    def test_start_workflow(self):
+    @mock.patch('mistral.rpc.clients.get_engine_client')
+    def test_start_workflow(self, get_engine_client_mock):
         cfg.CONF.set_default('auth_enable', True, group='pecan')
 
         wf = workflows.create_workflows(WORKFLOW_LIST)[0]
@@ -65,10 +68,27 @@ class ProcessCronTriggerTest(base.EngineTestCase):
         next_trigger = triggers.get_next_cron_triggers()[0]
         next_execution_time_before = next_trigger.next_execution_time
 
-        periodic.MistralPeriodicTasks(cfg.CONF).process_cron_triggers_v2(None)
+        periodic.process_cron_triggers_v2(None, None)
+        start_wf_mock = get_engine_client_mock.return_value.start_workflow
+
+        start_wf_mock.assert_called_once()
+
+        # Check actual parameters of the call.
+        self.assertEqual(
+            ('my_wf', '', None, {}),
+            start_wf_mock.mock_calls[0][1]
+        )
+        self.assertIn(
+            t.id,
+            start_wf_mock.mock_calls[0][2]['description']
+        )
+
+        self._await(
+            lambda: triggers.get_next_cron_triggers(),
+            fail_message="No triggers were found"
+        )
 
         next_triggers = triggers.get_next_cron_triggers()
-
         self.assertEqual(1, len(next_triggers))
 
         next_trigger = next_triggers[0]
@@ -103,8 +123,13 @@ class ProcessCronTriggerTest(base.EngineTestCase):
 
         next_trigger = next_triggers[0]
         next_execution_time_before = next_trigger.next_execution_time
+        ts_before = datetime.datetime.utcnow()
 
-        periodic.MistralPeriodicTasks(cfg.CONF).process_cron_triggers_v2(None)
+        periodic.process_cron_triggers_v2(None, None)
+        self._await(
+            lambda: triggers.get_next_cron_triggers(),
+            fail_message="No triggers were found"
+        )
 
         next_triggers = triggers.get_next_cron_triggers()
 
@@ -113,6 +138,11 @@ class ProcessCronTriggerTest(base.EngineTestCase):
         next_trigger = next_triggers[0]
         next_execution_time_after = next_trigger.next_execution_time
 
+        self.assertGreater(
+            next_execution_time_after,
+            ts_before
+        )
+
         self.assertNotEqual(
             next_execution_time_before,
             next_execution_time_after
@@ -127,7 +157,7 @@ class ProcessCronTriggerTest(base.EngineTestCase):
 
         # Make the first_time 1 sec later than current time, in order to make
         # it executed by next cron-trigger task.
-        first_time = datetime.datetime.now() + datetime.timedelta(0, 1)
+        first_time = datetime.datetime.utcnow() + datetime.timedelta(0, 1)
 
         # Creates a cron-trigger with pattern and first time, ensure the
         # cron-trigger can be executed more than once, and cron-trigger will
@@ -145,12 +175,14 @@ class ProcessCronTriggerTest(base.EngineTestCase):
             None
         )
 
-        self.assertEqual(
-            first_time,
-            cron_trigger.next_execution_time
-        )
+        interval = (cron_trigger.next_execution_time - first_time)
 
-        periodic.MistralPeriodicTasks(cfg.CONF).process_cron_triggers_v2(None)
+        self.assertLessEqual(interval.total_seconds(), 3.0)
+
+        periodic.process_cron_triggers_v2(None, None)
+
+        # After process_triggers context is set to None, need to reset it.
+        auth_ctx.set_ctx(self.ctx)
 
         next_time = triggers.get_next_execution_time(
             cron_trigger.pattern,
@@ -160,7 +192,20 @@ class ProcessCronTriggerTest(base.EngineTestCase):
         cron_trigger_db = db_api.get_cron_trigger(trigger_name)
 
         self.assertIsNotNone(cron_trigger_db)
-        self.assertEqual(
-            next_time,
-            cron_trigger_db.next_execution_time
+
+        interval = (cron_trigger_db.next_execution_time - next_time)
+
+        self.assertLessEqual(interval.total_seconds(), 3.0)
+
+    def test_validate_cron_trigger_input_first_time(self):
+        cfg.CONF.set_default('auth_enable', False, group='pecan')
+
+        first_time = datetime.datetime.utcnow() + datetime.timedelta(0, 1)
+
+        self.assertRaises(
+            exc.InvalidModelException,
+            triggers.validate_cron_trigger_input,
+            None,
+            first_time,
+            None
         )
