@@ -21,6 +21,7 @@ from mistral.actions import std_actions
 from mistral.db.v2 import api as db_api
 from mistral import exceptions as exc
 from mistral.services import workbooks as wb_service
+from mistral.services import workflows as wf_service
 from mistral.tests.unit.engine import base
 from mistral.workflow import states
 
@@ -224,18 +225,21 @@ class DirectWorkflowRerunTest(base.EngineTestCase):
         wb_service.create_workbook_v2(SIMPLE_WORKBOOK)
 
         # Run workflow and fail task.
-        wf_ex = self.engine.start_workflow('wb1.wf1', {})
+        wf_ex = self.engine.start_workflow('wb1.wf1')
 
         self.await_workflow_error(wf_ex.id)
 
-        wf_ex = db_api.get_workflow_execution(wf_ex.id)
+        with db_api.transaction():
+            wf_ex = db_api.get_workflow_execution(wf_ex.id)
+
+            task_execs = wf_ex.task_executions
 
         self.assertEqual(states.ERROR, wf_ex.state)
         self.assertIsNotNone(wf_ex.state_info)
-        self.assertEqual(2, len(wf_ex.task_executions))
+        self.assertEqual(2, len(task_execs))
 
-        task_1_ex = self._assert_single_item(wf_ex.task_executions, name='t1')
-        task_2_ex = self._assert_single_item(wf_ex.task_executions, name='t2')
+        task_1_ex = self._assert_single_item(task_execs, name='t1')
+        task_2_ex = self._assert_single_item(task_execs, name='t2')
 
         self.assertEqual(states.SUCCESS, task_1_ex.state)
         self.assertEqual(states.ERROR, task_2_ex.state)
@@ -252,15 +256,18 @@ class DirectWorkflowRerunTest(base.EngineTestCase):
         # Wait for the workflow to succeed.
         self.await_workflow_success(wf_ex.id)
 
-        wf_ex = db_api.get_workflow_execution(wf_ex.id)
+        with db_api.transaction():
+            wf_ex = db_api.get_workflow_execution(wf_ex.id)
+
+            task_execs = wf_ex.task_executions
 
         self.assertEqual(states.SUCCESS, wf_ex.state)
         self.assertIsNone(wf_ex.state_info)
-        self.assertEqual(3, len(wf_ex.task_executions))
+        self.assertEqual(3, len(task_execs))
 
-        task_1_ex = self._assert_single_item(wf_ex.task_executions, name='t1')
-        task_2_ex = self._assert_single_item(wf_ex.task_executions, name='t2')
-        task_3_ex = self._assert_single_item(wf_ex.task_executions, name='t3')
+        task_1_ex = self._assert_single_item(task_execs, name='t1')
+        task_2_ex = self._assert_single_item(task_execs, name='t2')
+        task_3_ex = self._assert_single_item(task_execs, name='t3')
 
         # Check action executions of task 1.
         self.assertEqual(states.SUCCESS, task_1_ex.state)
@@ -281,8 +288,10 @@ class DirectWorkflowRerunTest(base.EngineTestCase):
         )
 
         self.assertEqual(2, len(task_2_action_exs))
-        self.assertEqual(states.ERROR, task_2_action_exs[0].state)
-        self.assertEqual(states.SUCCESS, task_2_action_exs[1].state)
+        # Check there is exactly 1 action in Success and 1 in error state.
+        # Order doesn't matter.
+        self._assert_single_item(task_2_action_exs, state=states.SUCCESS)
+        self._assert_single_item(task_2_action_exs, state=states.ERROR)
 
         # Check action executions of task 3.
         self.assertEqual(states.SUCCESS, task_3_ex.state)
@@ -318,7 +327,7 @@ class DirectWorkflowRerunTest(base.EngineTestCase):
         }
 
         # Run workflow and fail task.
-        wf_ex = self.engine.start_workflow('wb1.wf1', {}, env=env)
+        wf_ex = self.engine.start_workflow('wb1.wf1', env=env)
 
         self.await_workflow_error(wf_ex.id)
 
@@ -330,7 +339,6 @@ class DirectWorkflowRerunTest(base.EngineTestCase):
         self.assertIsNotNone(wf_ex.state_info)
         self.assertEqual(3, len(task_execs))
         self.assertDictEqual(env, wf_ex.params['env'])
-        self.assertDictEqual(env, wf_ex.context['__env'])
 
         task_10_ex = self._assert_single_item(task_execs, name='t10')
         task_21_ex = self._assert_single_item(task_execs, name='t21')
@@ -354,7 +362,6 @@ class DirectWorkflowRerunTest(base.EngineTestCase):
         self.assertEqual(states.RUNNING, wf_ex.state)
         self.assertIsNone(wf_ex.state_info)
         self.assertDictEqual(updated_env, wf_ex.params['env'])
-        self.assertDictEqual(updated_env, wf_ex.context['__env'])
 
         # Await t30 success.
         self.await_task_success(task_30_ex.id)
@@ -399,17 +406,25 @@ class DirectWorkflowRerunTest(base.EngineTestCase):
         )
 
         self.assertEqual(2, len(task_21_action_exs))
-        self.assertEqual(states.ERROR, task_21_action_exs[0].state)
-        self.assertEqual(states.SUCCESS, task_21_action_exs[1].state)
+        # Check there is exactly 1 action in Success and 1 in error state.
+        # Order doesn't matter.
+        task_21_action_exs_1 = self._assert_single_item(
+            task_21_action_exs,
+            state=states.ERROR
+        )
+        task_21_action_exs_2 = self._assert_single_item(
+            task_21_action_exs,
+            state=states.SUCCESS
+        )
 
         self.assertDictEqual(
             {'output': env['var1']},
-            task_21_action_exs[0].input
+            task_21_action_exs_1.input
         )
 
         self.assertDictEqual(
             {'output': updated_env['var1']},
-            task_21_action_exs[1].input
+            task_21_action_exs_2.input
         )
 
         # Check action executions of task 22.
@@ -456,17 +471,18 @@ class DirectWorkflowRerunTest(base.EngineTestCase):
         wb_service.create_workbook_v2(SIMPLE_WORKBOOK)
 
         # Run workflow and fail task.
-        wf_ex = self.engine.start_workflow('wb1.wf1', {})
+        wf_ex = self.engine.start_workflow('wb1.wf1')
 
         self.await_workflow_error(wf_ex.id)
 
-        wf_ex = db_api.get_workflow_execution(wf_ex.id)
+        with db_api.transaction():
+            wf_ex = db_api.get_workflow_execution(wf_ex.id)
+
+            task_execs = wf_ex.task_executions
 
         self.assertEqual(states.ERROR, wf_ex.state)
         self.assertIsNotNone(wf_ex.state_info)
 
-        task_execs = wf_ex.task_executions
-
         self.assertEqual(2, len(task_execs))
 
         task_1_ex = self._assert_single_item(
@@ -509,17 +525,18 @@ class DirectWorkflowRerunTest(base.EngineTestCase):
         wb_service.create_workbook_v2(WITH_ITEMS_WORKBOOK)
 
         # Run workflow and fail task.
-        wf_ex = self.engine.start_workflow('wb3.wf1', {})
+        wf_ex = self.engine.start_workflow('wb3.wf1')
 
         self.await_workflow_error(wf_ex.id)
 
-        wf_ex = db_api.get_workflow_execution(wf_ex.id)
+        with db_api.transaction():
+            wf_ex = db_api.get_workflow_execution(wf_ex.id)
+
+            task_execs = wf_ex.task_executions
 
         self.assertEqual(states.ERROR, wf_ex.state)
         self.assertIsNotNone(wf_ex.state_info)
 
-        task_execs = wf_ex.task_executions
-
         self.assertEqual(1, len(task_execs))
 
         task_1_ex = self._assert_single_item(task_execs, name='t1')
@@ -541,15 +558,16 @@ class DirectWorkflowRerunTest(base.EngineTestCase):
         self.assertEqual(states.RUNNING, wf_ex.state)
         self.assertIsNone(wf_ex.state_info)
 
-        self.await_workflow_success(wf_ex.id, delay=10)
+        self.await_workflow_success(wf_ex.id)
 
-        wf_ex = db_api.get_workflow_execution(wf_ex.id)
+        with db_api.transaction():
+            wf_ex = db_api.get_workflow_execution(wf_ex.id)
+
+            task_execs = wf_ex.task_executions
 
         self.assertEqual(states.SUCCESS, wf_ex.state)
         self.assertIsNone(wf_ex.state_info)
 
-        task_execs = wf_ex.task_executions
-
         self.assertEqual(2, len(task_execs))
 
         task_1_ex = self._assert_single_item(task_execs, name='t1')
@@ -599,7 +617,7 @@ class DirectWorkflowRerunTest(base.EngineTestCase):
         wb_service.create_workbook_v2(WITH_ITEMS_WORKBOOK_CONCURRENCY)
 
         # Run workflow and fail task.
-        wf_ex = self.engine.start_workflow('wb3.wf1', {})
+        wf_ex = self.engine.start_workflow('wb3.wf1')
 
         self.await_workflow_error(wf_ex.id)
 
@@ -627,7 +645,7 @@ class DirectWorkflowRerunTest(base.EngineTestCase):
         self.assertEqual(states.RUNNING, wf_ex.state)
         self.assertIsNone(wf_ex.state_info)
 
-        self.await_workflow_success(wf_ex.id, delay=10)
+        self.await_workflow_success(wf_ex.id)
 
         wf_ex = db_api.get_workflow_execution(wf_ex.id)
 
@@ -681,17 +699,20 @@ class DirectWorkflowRerunTest(base.EngineTestCase):
         env = {'var1': 'fee fi fo fum'}
 
         # Run workflow and fail task.
-        wf_ex = self.engine.start_workflow('wb3.wf1', {}, env=env)
+        wf_ex = self.engine.start_workflow('wb3.wf1', env=env)
 
         self.await_workflow_error(wf_ex.id)
 
-        wf_ex = db_api.get_workflow_execution(wf_ex.id)
+        with db_api.transaction():
+            wf_ex = db_api.get_workflow_execution(wf_ex.id)
+
+            task_execs = wf_ex.task_executions
 
         self.assertEqual(states.ERROR, wf_ex.state)
         self.assertIsNotNone(wf_ex.state_info)
-        self.assertEqual(1, len(wf_ex.task_executions))
+        self.assertEqual(1, len(task_execs))
 
-        task_1_ex = self._assert_single_item(wf_ex.task_executions, name='t1')
+        task_1_ex = self._assert_single_item(task_execs, name='t1')
 
         self.assertEqual(states.ERROR, task_1_ex.state)
         self.assertIsNotNone(task_1_ex.state_info)
@@ -703,9 +724,7 @@ class DirectWorkflowRerunTest(base.EngineTestCase):
         self.assertEqual(3, len(task_1_action_exs))
 
         # Update env in workflow execution with the following.
-        updated_env = {
-            'var1': 'foobar'
-        }
+        updated_env = {'var1': 'foobar'}
 
         # Resume workflow and re-run failed task.
         self.engine.rerun_workflow(
@@ -719,16 +738,19 @@ class DirectWorkflowRerunTest(base.EngineTestCase):
         self.assertEqual(states.RUNNING, wf_ex.state)
         self.assertIsNone(wf_ex.state_info)
 
-        self.await_workflow_success(wf_ex.id, delay=10)
+        self.await_workflow_success(wf_ex.id)
 
-        wf_ex = db_api.get_workflow_execution(wf_ex.id)
+        with db_api.transaction():
+            wf_ex = db_api.get_workflow_execution(wf_ex.id)
+
+            task_execs = wf_ex.task_executions
 
         self.assertEqual(states.SUCCESS, wf_ex.state)
         self.assertIsNone(wf_ex.state_info)
-        self.assertEqual(2, len(wf_ex.task_executions))
+        self.assertEqual(2, len(task_execs))
 
-        task_1_ex = self._assert_single_item(wf_ex.task_executions, name='t1')
-        task_2_ex = self._assert_single_item(wf_ex.task_executions, name='t2')
+        task_1_ex = self._assert_single_item(task_execs, name='t1')
+        task_2_ex = self._assert_single_item(task_execs, name='t2')
 
         # Check action executions of task 1.
         self.assertEqual(states.SUCCESS, task_1_ex.state)
@@ -746,13 +768,17 @@ class DirectWorkflowRerunTest(base.EngineTestCase):
             'Task 1.2 [%s]' % updated_env['var1']   # Task 1 item 2 (rerun).
         ]
 
-        result = zip(task_1_action_exs, expected_inputs)
+        # Assert that every expected input is in actual task input.
+        for action_ex in task_1_action_exs:
+            self.assertIn(action_ex.input['output'], expected_inputs)
 
-        for (action_ex, expected_input) in result:
-            self.assertDictEqual(
-                {'output': expected_input},
-                action_ex.input
-            )
+        # Assert that there was same number of unique inputs as action execs.
+        self.assertEqual(
+            len(task_1_action_exs),
+            len(set(
+                [action_ex.input['output'] for action_ex in task_1_action_exs]
+            ))
+        )
 
         # Check action executions of task 2.
         self.assertEqual(states.SUCCESS, task_2_ex.state)
@@ -779,21 +805,24 @@ class DirectWorkflowRerunTest(base.EngineTestCase):
         wb_service.create_workbook_v2(JOIN_WORKBOOK)
 
         # Run workflow and fail task.
-        wf_ex = self.engine.start_workflow('wb1.wf1', {})
+        wf_ex = self.engine.start_workflow('wb1.wf1')
 
         wf_ex = db_api.get_workflow_execution(wf_ex.id)
 
         self.await_workflow_error(wf_ex.id)
 
-        wf_ex = db_api.get_workflow_execution(wf_ex.id)
+        with db_api.transaction():
+            wf_ex = db_api.get_workflow_execution(wf_ex.id)
+
+            task_execs = wf_ex.task_executions
 
         self.assertEqual(states.ERROR, wf_ex.state)
         self.assertIsNotNone(wf_ex.state_info)
-        self.assertEqual(3, len(wf_ex.task_executions))
+        self.assertEqual(3, len(task_execs))
 
-        task_1_ex = self._assert_single_item(wf_ex.task_executions, name='t1')
-        task_2_ex = self._assert_single_item(wf_ex.task_executions, name='t2')
-        task_3_ex = self._assert_single_item(wf_ex.task_executions, name='t3')
+        task_1_ex = self._assert_single_item(task_execs, name='t1')
+        task_2_ex = self._assert_single_item(task_execs, name='t2')
+        task_3_ex = self._assert_single_item(task_execs, name='t3')
 
         self.assertEqual(states.SUCCESS, task_1_ex.state)
         self.assertEqual(states.SUCCESS, task_2_ex.state)
@@ -811,15 +840,18 @@ class DirectWorkflowRerunTest(base.EngineTestCase):
         # Wait for the workflow to succeed.
         self.await_workflow_success(wf_ex.id)
 
-        wf_ex = db_api.get_workflow_execution(wf_ex.id)
+        with db_api.transaction():
+            wf_ex = db_api.get_workflow_execution(wf_ex.id)
+
+            task_execs = wf_ex.task_executions
 
         self.assertEqual(states.SUCCESS, wf_ex.state)
         self.assertIsNone(wf_ex.state_info)
-        self.assertEqual(3, len(wf_ex.task_executions))
+        self.assertEqual(3, len(task_execs))
 
-        task_1_ex = self._assert_single_item(wf_ex.task_executions, name='t1')
-        task_2_ex = self._assert_single_item(wf_ex.task_executions, name='t2')
-        task_3_ex = self._assert_single_item(wf_ex.task_executions, name='t3')
+        task_1_ex = self._assert_single_item(task_execs, name='t1')
+        task_2_ex = self._assert_single_item(task_execs, name='t2')
+        task_3_ex = self._assert_single_item(task_execs, name='t3')
 
         # Check action executions of task 1.
         self.assertEqual(states.SUCCESS, task_1_ex.state)
@@ -846,12 +878,14 @@ class DirectWorkflowRerunTest(base.EngineTestCase):
         self.assertIsNone(task_3_ex.state_info)
 
         task_3_action_exs = db_api.get_action_executions(
-            task_execution_id=wf_ex.task_executions[2].id
+            task_execution_id=task_3_ex.id
         )
 
         self.assertEqual(2, len(task_3_action_exs))
-        self.assertEqual(states.ERROR, task_3_action_exs[0].state)
-        self.assertEqual(states.SUCCESS, task_3_action_exs[1].state)
+        # Check there is exactly 1 action in Success and 1 in error state.
+        # Order doesn't matter.
+        self._assert_single_item(task_3_action_exs, state=states.SUCCESS)
+        self._assert_single_item(task_3_action_exs, state=states.ERROR)
 
     @mock.patch.object(
         std_actions.EchoAction,
@@ -870,7 +904,7 @@ class DirectWorkflowRerunTest(base.EngineTestCase):
         wb_service.create_workbook_v2(JOIN_WORKBOOK)
 
         # Run workflow and fail task.
-        wf_ex = self.engine.start_workflow('wb1.wf1', {})
+        wf_ex = self.engine.start_workflow('wb1.wf1')
 
         with db_api.transaction():
             wf_ex = db_api.get_workflow_execution(wf_ex.id)
@@ -968,8 +1002,10 @@ class DirectWorkflowRerunTest(base.EngineTestCase):
         )
 
         self.assertEqual(2, len(task_1_action_exs))
-        self.assertEqual(states.ERROR, task_1_action_exs[0].state)
-        self.assertEqual(states.SUCCESS, task_1_action_exs[1].state)
+        # Check there is exactly 1 action in Success and 1 in error state.
+        # Order doesn't matter.
+        self._assert_single_item(task_1_action_exs, state=states.SUCCESS)
+        self._assert_single_item(task_1_action_exs, state=states.ERROR)
 
         # Check action executions of task 2.
         self.assertEqual(states.SUCCESS, task_2_ex.state)
@@ -980,14 +1016,16 @@ class DirectWorkflowRerunTest(base.EngineTestCase):
         )
 
         self.assertEqual(2, len(task_2_action_exs))
-        self.assertEqual(states.ERROR, task_2_action_exs[0].state)
-        self.assertEqual(states.SUCCESS, task_2_action_exs[1].state)
+        # Check there is exactly 1 action in Success and 1 in error state.
+        # Order doesn't matter.
+        self._assert_single_item(task_2_action_exs, state=states.SUCCESS)
+        self._assert_single_item(task_2_action_exs, state=states.ERROR)
 
         # Check action executions of task 3.
         self.assertEqual(states.SUCCESS, task_3_ex.state)
 
         task_3_action_exs = db_api.get_action_executions(
-            task_execution_id=task_execs[2].id
+            task_execution_id=task_3_ex.id
         )
 
         self.assertEqual(1, len(task_3_action_exs))
@@ -1018,17 +1056,20 @@ class DirectWorkflowRerunTest(base.EngineTestCase):
         wb_service.create_workbook_v2(WITH_ITEMS_WORKBOOK)
 
         # Run workflow and fail task.
-        wf_ex = self.engine.start_workflow('wb3.wf1', {})
+        wf_ex = self.engine.start_workflow('wb3.wf1')
 
         self.await_workflow_error(wf_ex.id)
 
-        wf_ex = db_api.get_workflow_execution(wf_ex.id)
+        with db_api.transaction():
+            wf_ex = db_api.get_workflow_execution(wf_ex.id)
+
+            task_execs = wf_ex.task_executions
 
         self.assertEqual(states.ERROR, wf_ex.state)
         self.assertIsNotNone(wf_ex.state_info)
-        self.assertEqual(1, len(wf_ex.task_executions))
+        self.assertEqual(1, len(task_execs))
 
-        task_1_ex = self._assert_single_item(wf_ex.task_executions, name='t1')
+        task_1_ex = self._assert_single_item(task_execs, name='t1')
 
         self.await_task_error(task_1_ex.id)
 
@@ -1046,7 +1087,7 @@ class DirectWorkflowRerunTest(base.EngineTestCase):
         self.assertEqual(states.RUNNING, wf_ex.state)
         self.assertIsNone(wf_ex.state_info)
 
-        self.await_workflow_error(wf_ex.id, delay=10)
+        self.await_workflow_error(wf_ex.id)
 
         wf_ex = db_api.get_workflow_execution(wf_ex.id)
 
@@ -1067,7 +1108,7 @@ class DirectWorkflowRerunTest(base.EngineTestCase):
         self.assertEqual(states.RUNNING, wf_ex.state)
         self.assertIsNone(wf_ex.state_info)
 
-        self.await_workflow_error(wf_ex.id, delay=10)
+        self.await_workflow_error(wf_ex.id)
 
         wf_ex = db_api.get_workflow_execution(wf_ex.id)
 
@@ -1088,7 +1129,7 @@ class DirectWorkflowRerunTest(base.EngineTestCase):
         self.assertEqual(states.RUNNING, wf_ex.state)
         self.assertIsNone(wf_ex.state_info)
 
-        self.await_workflow_error(wf_ex.id, delay=10)
+        self.await_workflow_error(wf_ex.id)
 
         wf_ex = db_api.get_workflow_execution(wf_ex.id)
 
@@ -1109,16 +1150,19 @@ class DirectWorkflowRerunTest(base.EngineTestCase):
         self.assertEqual(states.RUNNING, wf_ex.state)
         self.assertIsNone(wf_ex.state_info)
 
-        self.await_workflow_success(wf_ex.id, delay=10)
+        self.await_workflow_success(wf_ex.id)
 
-        wf_ex = db_api.get_workflow_execution(wf_ex.id)
+        with db_api.transaction():
+            wf_ex = db_api.get_workflow_execution(wf_ex.id)
+
+            task_execs = wf_ex.task_executions
 
         self.assertEqual(states.SUCCESS, wf_ex.state)
         self.assertIsNone(wf_ex.state_info)
-        self.assertEqual(2, len(wf_ex.task_executions))
+        self.assertEqual(2, len(task_execs))
 
-        task_1_ex = self._assert_single_item(wf_ex.task_executions, name='t1')
-        task_2_ex = self._assert_single_item(wf_ex.task_executions, name='t2')
+        task_1_ex = self._assert_single_item(task_execs, name='t1')
+        task_2_ex = self._assert_single_item(task_execs, name='t2')
 
         # Check action executions of task 1.
         self.assertEqual(states.SUCCESS, task_1_ex.state)
@@ -1131,8 +1175,10 @@ class DirectWorkflowRerunTest(base.EngineTestCase):
         # The single action execution that succeeded should not re-run.
         self.assertEqual(12, len(task_1_action_exs))
 
-        self.assertListEqual(['Task 1.0', 'Task 1.1', 'Task 1.2'],
-                             task_1_ex.published.get('v1'))
+        self.assertListEqual(
+            ['Task 1.0', 'Task 1.1', 'Task 1.2'],
+            task_1_ex.published.get('v1')
+        )
 
         # Check action executions of task 2.
         self.assertEqual(states.SUCCESS, task_2_ex.state)
@@ -1159,18 +1205,21 @@ class DirectWorkflowRerunTest(base.EngineTestCase):
         wb_service.create_workbook_v2(SUBFLOW_WORKBOOK)
 
         # Run workflow and fail task.
-        wf_ex = self.engine.start_workflow('wb1.wf1', {})
+        wf_ex = self.engine.start_workflow('wb1.wf1')
 
         self.await_workflow_error(wf_ex.id)
 
-        wf_ex = db_api.get_workflow_execution(wf_ex.id)
+        with db_api.transaction():
+            wf_ex = db_api.get_workflow_execution(wf_ex.id)
+
+            task_execs = wf_ex.task_executions
 
         self.assertEqual(states.ERROR, wf_ex.state)
         self.assertIsNotNone(wf_ex.state_info)
-        self.assertEqual(2, len(wf_ex.task_executions))
+        self.assertEqual(2, len(task_execs))
 
-        task_1_ex = self._assert_single_item(wf_ex.task_executions, name='t1')
-        task_2_ex = self._assert_single_item(wf_ex.task_executions, name='t2')
+        task_1_ex = self._assert_single_item(task_execs, name='t1')
+        task_2_ex = self._assert_single_item(task_execs, name='t2')
 
         self.assertEqual(states.SUCCESS, task_1_ex.state)
         self.assertEqual(states.ERROR, task_2_ex.state)
@@ -1178,6 +1227,7 @@ class DirectWorkflowRerunTest(base.EngineTestCase):
 
         # Resume workflow and re-run failed task.
         self.engine.rerun_workflow(task_2_ex.id)
+
         wf_ex = db_api.get_workflow_execution(wf_ex.id)
 
         self.assertEqual(states.RUNNING, wf_ex.state)
@@ -1186,15 +1236,18 @@ class DirectWorkflowRerunTest(base.EngineTestCase):
         # Wait for the workflow to succeed.
         self.await_workflow_success(wf_ex.id)
 
-        wf_ex = db_api.get_workflow_execution(wf_ex.id)
+        with db_api.transaction():
+            wf_ex = db_api.get_workflow_execution(wf_ex.id)
+
+            task_execs = wf_ex.task_executions
 
         self.assertEqual(states.SUCCESS, wf_ex.state)
         self.assertIsNone(wf_ex.state_info)
-        self.assertEqual(3, len(wf_ex.task_executions))
+        self.assertEqual(3, len(task_execs))
 
-        task_1_ex = self._assert_single_item(wf_ex.task_executions, name='t1')
-        task_2_ex = self._assert_single_item(wf_ex.task_executions, name='t2')
-        task_3_ex = self._assert_single_item(wf_ex.task_executions, name='t3')
+        task_1_ex = self._assert_single_item(task_execs, name='t1')
+        task_2_ex = self._assert_single_item(task_execs, name='t2')
+        task_3_ex = self._assert_single_item(task_execs, name='t3')
 
         # Check action executions of task 1.
         self.assertEqual(states.SUCCESS, task_1_ex.state)
@@ -1214,14 +1267,17 @@ class DirectWorkflowRerunTest(base.EngineTestCase):
         )
 
         self.assertEqual(2, len(task_2_action_exs))
-        self.assertEqual(states.ERROR, task_2_action_exs[0].state)
-        self.assertEqual(states.SUCCESS, task_2_action_exs[1].state)
+        # Check there is exactly 1 action in Success and 1 in error state.
+        # Order doesn't matter.
+        self._assert_single_item(task_2_action_exs, state=states.SUCCESS)
+        self._assert_single_item(task_2_action_exs, state=states.ERROR)
 
         # Check action executions of task 3.
         self.assertEqual(states.SUCCESS, task_3_ex.state)
 
         task_3_action_exs = db_api.get_action_executions(
-            task_execution_id=task_3_ex.id)
+            task_execution_id=task_3_ex.id
+        )
 
         self.assertEqual(1, len(task_3_action_exs))
         self.assertEqual(states.SUCCESS, task_3_action_exs[0].state)
@@ -1242,36 +1298,41 @@ class DirectWorkflowRerunTest(base.EngineTestCase):
         wb_service.create_workbook_v2(SUBFLOW_WORKBOOK)
 
         # Run workflow and fail task.
-        wf_ex = self.engine.start_workflow('wb1.wf1', {})
+        wf_ex = self.engine.start_workflow('wb1.wf1')
 
         self.await_workflow_error(wf_ex.id)
 
-        wf_ex = db_api.get_workflow_execution(wf_ex.id)
+        with db_api.transaction():
+            wf_ex = db_api.get_workflow_execution(wf_ex.id)
+
+            task_execs = wf_ex.task_executions
 
         self.assertEqual(states.ERROR, wf_ex.state)
         self.assertIsNotNone(wf_ex.state_info)
-        self.assertEqual(2, len(wf_ex.task_executions))
+        self.assertEqual(2, len(task_execs))
 
-        task_1_ex = self._assert_single_item(wf_ex.task_executions, name='t1')
-        task_2_ex = self._assert_single_item(wf_ex.task_executions, name='t2')
+        task_1_ex = self._assert_single_item(task_execs, name='t1')
+        task_2_ex = self._assert_single_item(task_execs, name='t2')
 
         self.assertEqual(states.SUCCESS, task_1_ex.state)
         self.assertEqual(states.ERROR, task_2_ex.state)
         self.assertIsNotNone(task_2_ex.state_info)
 
-        # Get subworkflow and related task
-        sub_wf_exs = db_api.get_workflow_executions(
-            task_execution_id=task_2_ex.id
-        )
+        with db_api.transaction():
+            # Get subworkflow and related task
+            sub_wf_exs = db_api.get_workflow_executions(
+                task_execution_id=task_2_ex.id
+            )
 
-        sub_wf_ex = sub_wf_exs[0]
+            sub_wf_ex = sub_wf_exs[0]
+            sub_wf_task_execs = sub_wf_ex.task_executions
 
         self.assertEqual(states.ERROR, sub_wf_ex.state)
         self.assertIsNotNone(sub_wf_ex.state_info)
-        self.assertEqual(1, len(sub_wf_ex.task_executions))
+        self.assertEqual(1, len(sub_wf_task_execs))
 
         sub_wf_task_ex = self._assert_single_item(
-            sub_wf_ex.task_executions,
+            sub_wf_task_execs,
             name='wf2_t1'
         )
 
@@ -1294,14 +1355,17 @@ class DirectWorkflowRerunTest(base.EngineTestCase):
         # Wait for the subworkflow to succeed.
         self.await_workflow_success(sub_wf_ex.id)
 
-        sub_wf_ex = db_api.get_workflow_execution(sub_wf_ex.id)
+        with db_api.transaction():
+            sub_wf_ex = db_api.get_workflow_execution(sub_wf_ex.id)
+
+            sub_wf_task_execs = sub_wf_ex.task_executions
 
         self.assertEqual(states.SUCCESS, sub_wf_ex.state)
         self.assertIsNone(sub_wf_ex.state_info)
-        self.assertEqual(1, len(sub_wf_ex.task_executions))
+        self.assertEqual(1, len(sub_wf_task_execs))
 
         sub_wf_task_ex = self._assert_single_item(
-            sub_wf_ex.task_executions,
+            sub_wf_task_execs,
             name='wf2_t1'
         )
 
@@ -1310,30 +1374,40 @@ class DirectWorkflowRerunTest(base.EngineTestCase):
         self.assertIsNone(sub_wf_task_ex.state_info)
 
         sub_wf_task_ex_action_exs = db_api.get_action_executions(
-            task_execution_id=sub_wf_task_ex.id)
+            task_execution_id=sub_wf_task_ex.id
+        )
 
         self.assertEqual(2, len(sub_wf_task_ex_action_exs))
-        self.assertEqual(states.ERROR, sub_wf_task_ex_action_exs[0].state)
-        self.assertEqual(states.SUCCESS, sub_wf_task_ex_action_exs[1].state)
+        # Check there is exactly 1 action in Success and 1 in error state.
+        # Order doesn't matter.
+        self._assert_single_item(
+            sub_wf_task_ex_action_exs,
+            state=states.SUCCESS
+        )
+        self._assert_single_item(sub_wf_task_ex_action_exs, state=states.ERROR)
 
         # Wait for the main workflow to succeed.
         self.await_workflow_success(wf_ex.id)
 
-        wf_ex = db_api.get_workflow_execution(wf_ex.id)
+        with db_api.transaction():
+            wf_ex = db_api.get_workflow_execution(wf_ex.id)
+
+            task_execs = wf_ex.task_executions
 
         self.assertEqual(states.SUCCESS, wf_ex.state)
         self.assertIsNone(wf_ex.state_info)
-        self.assertEqual(3, len(wf_ex.task_executions))
+        self.assertEqual(3, len(task_execs))
 
-        task_1_ex = self._assert_single_item(wf_ex.task_executions, name='t1')
-        task_2_ex = self._assert_single_item(wf_ex.task_executions, name='t2')
-        task_3_ex = self._assert_single_item(wf_ex.task_executions, name='t3')
+        task_1_ex = self._assert_single_item(task_execs, name='t1')
+        task_2_ex = self._assert_single_item(task_execs, name='t2')
+        task_3_ex = self._assert_single_item(task_execs, name='t3')
 
         # Check action executions of task 1.
         self.assertEqual(states.SUCCESS, task_1_ex.state)
 
         task_1_action_exs = db_api.get_action_executions(
-            task_execution_id=task_1_ex.id)
+            task_execution_id=task_1_ex.id
+        )
 
         self.assertEqual(1, len(task_1_action_exs))
         self.assertEqual(states.SUCCESS, task_1_action_exs[0].state)
@@ -1357,3 +1431,52 @@ class DirectWorkflowRerunTest(base.EngineTestCase):
 
         self.assertEqual(1, len(task_3_action_exs))
         self.assertEqual(states.SUCCESS, task_3_action_exs[0].state)
+
+    def test_rerun_task_with_retry_policy(self):
+        wf_service.create_workflows("""---
+        version: '2.0'
+        wf_fail:
+          tasks:
+            task1:
+              action: std.fail
+              retry:
+                delay: 0
+                count: 2""")
+
+        wf_ex = self.engine.start_workflow("wf_fail")
+
+        self.await_workflow_error(wf_ex.id)
+
+        with db_api.transaction():
+            wf_ex = db_api.get_workflow_execution(wf_ex.id)
+
+            task_ex = self._assert_single_item(wf_ex.task_executions,
+                                               name="task1")
+            action_executions = task_ex.executions
+
+        self.assertEqual(states.ERROR, wf_ex.state)
+        self.assertIsNotNone(wf_ex.state_info)
+        self.assertEqual(3, len(action_executions))
+        self.assertTrue(all(a.state == states.ERROR
+                            for a in action_executions))
+
+        self.engine.rerun_workflow(task_ex.id)
+
+        wf_ex = db_api.get_workflow_execution(wf_ex.id)
+
+        self.assertEqual(states.RUNNING, wf_ex.state)
+
+        self.await_workflow_error(wf_ex.id)
+
+        with db_api.transaction():
+            wf_ex = db_api.get_workflow_execution(wf_ex.id)
+
+            task_ex = self._assert_single_item(wf_ex.task_executions,
+                                               name="task1")
+            action_executions = task_ex.executions
+
+        self.assertEqual(states.ERROR, wf_ex.state)
+        self.assertIsNotNone(wf_ex.state_info)
+        self.assertEqual(6, len(action_executions))
+        self.assertTrue(all(a.state == states.ERROR
+                            for a in action_executions))
