@@ -16,8 +16,10 @@ from oslo_config import cfg
 import testtools
 
 from mistral.db.v2 import api as db_api
+from mistral.lang.v2 import tasks as tasks_lang
 from mistral.services import workflows as wf_service
 from mistral.tests.unit.engine import base
+from mistral import utils
 from mistral.workflow import states
 
 
@@ -47,10 +49,19 @@ class JoinEngineTest(base.EngineTestCase):
 
         wf_service.create_workflows(wf_text)
 
-        wf_ex = self.engine.start_workflow('wf', {})
+        wf_ex = self.engine.start_workflow('wf')
 
         self.await_workflow_success(wf_ex.id)
 
+        with db_api.transaction():
+            wf_ex = db_api.get_workflow_execution(wf_ex.id)
+
+            t_execs = wf_ex.task_executions
+
+        self._assert_single_item(t_execs, name='task1')
+        self._assert_single_item(t_execs, name='task2')
+        self._assert_single_item(t_execs, name='join_task')
+
     def test_full_join_without_errors(self):
         wf_text = """---
         version: '2.0'
@@ -86,7 +97,7 @@ class JoinEngineTest(base.EngineTestCase):
         wf_service.create_workflows(wf_text)
 
         # Start workflow.
-        wf_ex = self.engine.start_workflow('wf', {})
+        wf_ex = self.engine.start_workflow('wf')
 
         self.await_workflow_success(wf_ex.id)
 
@@ -139,7 +150,7 @@ class JoinEngineTest(base.EngineTestCase):
         wf_service.create_workflows(wf_text)
 
         # Start workflow.
-        wf_ex = self.engine.start_workflow('wf', {})
+        wf_ex = self.engine.start_workflow('wf')
 
         self.await_workflow_success(wf_ex.id)
 
@@ -200,16 +211,20 @@ class JoinEngineTest(base.EngineTestCase):
         wf_service.create_workflows(wf_text)
 
         # Start workflow.
-        wf_ex = self.engine.start_workflow('wf', {})
+        wf_ex = self.engine.start_workflow('wf')
 
-        self._await(
-            lambda:
-            len(db_api.get_workflow_execution(wf_ex.id).task_executions) == 4
-        )
+        def _num_of_tasks():
+            return len(
+                db_api.get_task_executions(workflow_execution_id=wf_ex.id)
+            )
 
-        # Note: We need to reread execution to access related tasks.
-        wf_ex = db_api.get_workflow_execution(wf_ex.id)
-        tasks = wf_ex.task_executions
+        self._await(lambda: _num_of_tasks() == 4)
+
+        with db_api.transaction():
+            # Note: We need to reread execution to access related tasks.
+            wf_ex = db_api.get_workflow_execution(wf_ex.id)
+
+            tasks = wf_ex.task_executions
 
         task1 = self._assert_single_item(tasks, name='task1')
         task2 = self._assert_single_item(tasks, name='task2')
@@ -217,13 +232,18 @@ class JoinEngineTest(base.EngineTestCase):
         task4 = self._assert_single_item(tasks, name='task4')
 
         # NOTE(xylan): We ensure task4 is successful here because of the
-        # uncertainty of its running parallelly with task3.
+        # uncertainty of its running in parallel with task3.
         self.await_task_success(task4.id)
 
         self.assertEqual(states.RUNNING, wf_ex.state)
         self.assertEqual(states.SUCCESS, task1.state)
         self.assertEqual(states.SUCCESS, task2.state)
-        self.assertEqual(states.WAITING, task3.state)
+
+        # NOTE(rakhmerov): Task 3 must fail because task2->task3 transition
+        # will never trigger due to its condition.
+        self.await_task_error(task3.id)
+
+        self.await_workflow_error(wf_ex.id)
 
     def test_partial_join(self):
         wf_text = """---
@@ -272,7 +292,7 @@ class JoinEngineTest(base.EngineTestCase):
         wf_service.create_workflows(wf_text)
 
         # Start workflow.
-        wf_ex = self.engine.start_workflow('wf', {})
+        wf_ex = self.engine.start_workflow('wf')
 
         self.await_workflow_success(wf_ex.id)
 
@@ -354,7 +374,7 @@ class JoinEngineTest(base.EngineTestCase):
         wf_service.create_workflows(wf_text)
 
         # Start workflow.
-        wf_ex = self.engine.start_workflow('wf', {})
+        wf_ex = self.engine.start_workflow('wf')
 
         self.await_workflow_success(wf_ex.id)
 
@@ -431,7 +451,7 @@ class JoinEngineTest(base.EngineTestCase):
         wf_service.create_workflows(wf_text)
 
         # Start workflow.
-        wf_ex = self.engine.start_workflow('wf', {})
+        wf_ex = self.engine.start_workflow('wf')
 
         self.await_workflow_success(wf_ex.id)
 
@@ -513,7 +533,7 @@ class JoinEngineTest(base.EngineTestCase):
         wf_service.create_workflows(wfs_tasks_join_complex)
 
         # Start workflow.
-        wf_ex = self.engine.start_workflow('main', {})
+        wf_ex = self.engine.start_workflow('main')
 
         self.await_workflow_success(wf_ex.id)
 
@@ -578,7 +598,7 @@ class JoinEngineTest(base.EngineTestCase):
         wf_service.create_workflows(wf_text)
 
         # Start workflow.
-        wf_ex = self.engine.start_workflow('main', {})
+        wf_ex = self.engine.start_workflow('main')
 
         self.await_workflow_success(wf_ex.id)
 
@@ -637,7 +657,7 @@ class JoinEngineTest(base.EngineTestCase):
 
         wf_service.create_workflows(wf_text)
 
-        wf_ex = self.engine.start_workflow('main', {})
+        wf_ex = self.engine.start_workflow('main')
 
         self.await_workflow_error(wf_ex.id)
 
@@ -691,7 +711,7 @@ class JoinEngineTest(base.EngineTestCase):
 
         wf_service.create_workflows(wf_text)
 
-        wf_ex = self.engine.start_workflow('test-join', {})
+        wf_ex = self.engine.start_workflow('test-join')
 
         self.await_workflow_success(wf_ex.id)
 
@@ -723,7 +743,7 @@ class JoinEngineTest(base.EngineTestCase):
 
         wf_service.create_workflows(wf_text)
 
-        wf_ex = self.engine.start_workflow('wf', {})
+        wf_ex = self.engine.start_workflow('wf')
 
         self.await_workflow_success(wf_ex.id)
 
@@ -767,7 +787,7 @@ class JoinEngineTest(base.EngineTestCase):
 
         wf_service.create_workflows(wf_text)
 
-        wf_ex = self.engine.start_workflow('wf', {})
+        wf_ex = self.engine.start_workflow('wf')
 
         self.await_workflow_success(wf_ex.id)
 
@@ -800,7 +820,7 @@ class JoinEngineTest(base.EngineTestCase):
 
         wf_service.create_workflows(wf_text)
 
-        wf_ex = self.engine.start_workflow('wf', {})
+        wf_ex = self.engine.start_workflow('wf')
 
         self.await_workflow_success(wf_ex.id)
 
@@ -812,7 +832,7 @@ class JoinEngineTest(base.EngineTestCase):
 
         self._assert_multiple_items(task_execs, 4, state=states.SUCCESS)
 
-    def delete_join_completion_check_on_stop(self):
+    def test_delete_join_completion_check_on_stop(self):
         wf_text = """---
         version: '2.0'
 
@@ -833,11 +853,11 @@ class JoinEngineTest(base.EngineTestCase):
 
         wf_service.create_workflows(wf_text)
 
-        wf_ex = self.engine.start_workflow('wf', {})
+        wf_ex = self.engine.start_workflow('wf')
 
         tasks = db_api.get_task_executions(workflow_execution_id=wf_ex.id)
 
-        self.assertTrue(len(tasks) >= 2)
+        self.assertGreaterEqual(len(tasks), 2)
 
         task1 = self._assert_single_item(tasks, name='task1')
 
@@ -857,7 +877,10 @@ class JoinEngineTest(base.EngineTestCase):
 
         mtd_name = 'mistral.engine.task_handler._refresh_task_state'
 
-        self._assert_single_item(calls, target_method_name=mtd_name)
+        cnt = sum([1 for c in calls if c.target_method_name == mtd_name])
+
+        # There can be 2 calls with different value of 'processing' flag.
+        self.assertTrue(cnt == 1 or cnt == 2)
 
         # Stop the workflow.
         self.engine.stop_workflow(wf_ex.id, state=states.CANCELLED)
@@ -867,7 +890,7 @@ class JoinEngineTest(base.EngineTestCase):
             len(db_api.get_delayed_calls(target_method_name=mtd_name)) == 0
         )
 
-    def delete_join_completion_check_on_execution_delete(self):
+    def test_delete_join_completion_check_on_execution_delete(self):
         wf_text = """---
         version: '2.0'
 
@@ -888,11 +911,11 @@ class JoinEngineTest(base.EngineTestCase):
 
         wf_service.create_workflows(wf_text)
 
-        wf_ex = self.engine.start_workflow('wf', {})
+        wf_ex = self.engine.start_workflow('wf')
 
         tasks = db_api.get_task_executions(workflow_execution_id=wf_ex.id)
 
-        self.assertTrue(len(tasks) >= 2)
+        self.assertGreaterEqual(len(tasks), 2)
 
         task1 = self._assert_single_item(tasks, name='task1')
 
@@ -912,7 +935,10 @@ class JoinEngineTest(base.EngineTestCase):
 
         mtd_name = 'mistral.engine.task_handler._refresh_task_state'
 
-        self._assert_single_item(calls, target_method_name=mtd_name)
+        cnt = sum([1 for c in calls if c.target_method_name == mtd_name])
+
+        # There can be 2 calls with different value of 'processing' flag.
+        self.assertTrue(cnt == 1 or cnt == 2)
 
         # Stop the workflow.
         db_api.delete_workflow_execution(wf_ex.id)
@@ -921,3 +947,364 @@ class JoinEngineTest(base.EngineTestCase):
             lambda:
             len(db_api.get_delayed_calls(target_method_name=mtd_name)) == 0
         )
+
+    def test_join_with_deep_dependencies_tree(self):
+        wf_text = """---
+        version: '2.0'
+
+        wf:
+          tasks:
+            task_a_1:
+              on-success:
+                - task_with_join
+
+            task_b_1:
+              action: std.fail
+              on-success:
+                - task_b_2
+
+            task_b_2:
+              on-success:
+                - task_b_3
+
+            task_b_3:
+              on-success:
+                - task_with_join
+
+            task_with_join:
+              join: all
+        """
+
+        wf_service.create_workflows(wf_text)
+
+        wf_ex = self.engine.start_workflow('wf')
+
+        self.await_workflow_error(wf_ex.id)
+
+        with db_api.transaction():
+            wf_ex = db_api.get_workflow_execution(wf_ex.id)
+
+            task_execs = wf_ex.task_executions
+
+        self.assertEqual(3, len(task_execs))
+        self._assert_single_item(
+            task_execs,
+            name='task_a_1',
+            state=states.SUCCESS
+        )
+        self._assert_single_item(
+            task_execs,
+            name='task_b_1',
+            state=states.ERROR
+        )
+        self._assert_single_item(
+            task_execs,
+            name='task_with_join',
+            state=states.ERROR
+        )
+
+    def test_no_workflow_error_after_inbound_error(self):
+        wf_text = """---
+        version: "2.0"
+
+        wf:
+          output:
+            continue_flag: <% $.get(continue_flag) %>
+
+          task-defaults:
+            on-error:
+              - change_continue_flag
+
+          tasks:
+            task_a:
+              action: std.fail
+              on-success:
+                - task_c: <% $.get(continue_flag) = null %>
+                - task_a_process
+
+            task_a_process:
+              action: std.noop
+
+            task_b:
+              on-success:
+                - task_c: <% $.get(continue_flag) = null %>
+
+            task_c:
+              join: all
+
+            change_continue_flag:
+              publish:
+                continue_flag: false
+        """
+
+        wf_service.create_workflows(wf_text)
+
+        wf_ex = self.engine.start_workflow('wf')
+
+        self.await_workflow_success(wf_ex.id)
+
+    def test_triggered_by_success(self):
+        wf_text = """---
+        version: '2.0'
+
+        wf:
+          type: direct
+
+          tasks:
+            join_task:
+              join: all
+
+            task1:
+              on-success: join_task
+
+            task2:
+              on-success: join_task
+        """
+
+        wf_service.create_workflows(wf_text)
+
+        wf_ex = self.engine.start_workflow('wf')
+
+        self.await_workflow_success(wf_ex.id)
+
+        with db_api.transaction():
+            wf_ex = db_api.get_workflow_execution(wf_ex.id)
+
+            t_execs = wf_ex.task_executions
+
+        task1 = self._assert_single_item(t_execs, name='task1')
+        task2 = self._assert_single_item(t_execs, name='task2')
+        join_task = self._assert_single_item(t_execs, name='join_task')
+
+        key = 'triggered_by'
+
+        self.assertIsNone(task1.runtime_context.get(key))
+        self.assertIsNone(task2.runtime_context.get(key))
+
+        self.assertIn(
+            {
+                "task_id": task1.id,
+                "event": "on-success"
+            },
+            join_task.runtime_context.get(key)
+        )
+        self.assertIn(
+            {
+                "task_id": task2.id,
+                "event": "on-success"
+            },
+            join_task.runtime_context.get(key)
+        )
+
+    def test_triggered_by_error(self):
+        wf_text = """---
+        version: '2.0'
+
+        wf:
+          type: direct
+
+          tasks:
+            task1:
+              on-success: join_task
+
+            task2:
+              action: std.fail
+              on-success: join_task
+
+            task3:
+              action: std.noop
+              on-error: join_task
+
+            join_task:
+              join: all
+        """
+
+        wf_service.create_workflows(wf_text)
+
+        wf_ex = self.engine.start_workflow('wf')
+
+        self.await_workflow_error(wf_ex.id)
+
+        with db_api.transaction():
+            wf_ex = db_api.get_workflow_execution(wf_ex.id)
+
+            t_execs = wf_ex.task_executions
+
+        task1 = self._assert_single_item(
+            t_execs,
+            name='task1',
+            state=states.SUCCESS
+        )
+        task2 = self._assert_single_item(
+            t_execs,
+            name='task2',
+            state=states.ERROR
+        )
+        task3 = self._assert_single_item(
+            t_execs,
+            name='task3',
+            state=states.SUCCESS
+        )
+        join_task = self._assert_single_item(
+            t_execs,
+            name='join_task',
+            state=states.ERROR
+        )
+
+        key = 'triggered_by'
+
+        self.assertIsNone(task1.runtime_context.get(key))
+        self.assertIsNone(task2.runtime_context.get(key))
+        self.assertIsNone(task3.runtime_context.get(key))
+
+        self.assertIn(
+            {
+                "task_id": task2.id,
+                "event": "not triggered"
+            },
+            join_task.runtime_context.get(key)
+        )
+        self.assertIn(
+            {
+                "task_id": task3.id,
+                "event": "not triggered"
+            },
+            join_task.runtime_context.get(key)
+        )
+
+    def test_triggered_by_impossible_route(self):
+        wf_text = """---
+        version: '2.0'
+
+        wf:
+          type: direct
+
+          tasks:
+            task1:
+              on-success: join_task
+
+            task2:
+              action: std.fail
+              on-success: task3
+
+            task3:
+              action: std.noop
+              on-success: join_task
+
+            join_task:
+              join: all
+        """
+
+        wf_service.create_workflows(wf_text)
+
+        wf_ex = self.engine.start_workflow('wf')
+
+        self.await_workflow_error(wf_ex.id)
+
+        with db_api.transaction():
+            wf_ex = db_api.get_workflow_execution(wf_ex.id)
+
+            t_execs = wf_ex.task_executions
+
+        task1 = self._assert_single_item(
+            t_execs,
+            name='task1',
+            state=states.SUCCESS
+        )
+        task2 = self._assert_single_item(
+            t_execs,
+            name='task2',
+            state=states.ERROR
+        )
+        join_task = self._assert_single_item(
+            t_execs,
+            name='join_task',
+            state=states.ERROR
+        )
+
+        self.assertEqual(3, len(t_execs))
+
+        key = 'triggered_by'
+
+        self.assertIsNone(task1.runtime_context.get(key))
+        self.assertIsNone(task2.runtime_context.get(key))
+
+        # Note: in case if execution does not exist for a previous
+        # task we can't track it in "triggered_by" because we need
+        # to know its ID so we leave it blank.
+        self.assertFalse(join_task.runtime_context.get(key))
+
+    def test_join_saving_task_context_with_all(self):
+        workflow = """---
+        version: '2.0'
+
+        test_workflow:
+          type: direct
+
+          tasks:
+            task1:
+              action: std.echo output='task1'
+              on-success:
+                - task2
+              publish:
+                result: <% task().result %>
+
+            task2:
+              action: std.echo output='task2'
+              join: all
+              publish:
+                result: <% task().result %>
+        """
+
+        wf_service.create_workflows(workflow)
+
+        wf_ex = self.engine.start_workflow('test_workflow')
+
+        self.await_workflow_success(wf_ex.id)
+
+        with db_api.transaction():
+            wf_ex = db_api.get_workflow_execution(wf_ex.id)
+            tasks = wf_ex.task_executions
+
+            for task in tasks:
+                task_result = task.published["result"]
+
+                self.assertEqual(task.name, task_result,
+                                 "The result of task must equal own name")
+
+    def test_join_with_long_name(self):
+        long_task_name = utils.generate_string(
+            tasks_lang.MAX_LENGTH_JOIN_TASK_NAME
+        )
+        wf_text = """---
+        version: '2.0'
+
+        wf:
+          tasks:
+            task1:
+              on-success:
+                - {0}
+
+            task2:
+              on-success:
+                - {0}
+
+            {0}:
+              join: all
+        """.format(long_task_name)
+
+        wf_service.create_workflows(wf_text)
+        wf_ex = self.engine.start_workflow('wf')
+        self.await_workflow_success(wf_ex.id)
+
+        with db_api.transaction():
+            wf_ex = db_api.get_workflow_execution(wf_ex.id)
+            tasks = wf_ex.task_executions
+
+        task1 = self._assert_single_item(tasks, name='task1')
+        task2 = self._assert_single_item(tasks, name='task2')
+        task3 = self._assert_single_item(tasks, name=long_task_name)
+
+        self.assertEqual(states.SUCCESS, task1.state)
+        self.assertEqual(states.SUCCESS, task2.state)
+        self.assertEqual(states.SUCCESS, task3.state)
