@@ -19,10 +19,12 @@ from oslo_config import cfg
 
 from mistral.db.v2.sqlalchemy import api as db_api
 from mistral import exceptions as exc
+from mistral.lang import parser as spec_parser
+from mistral.lang.v2 import tasks
+from mistral.lang.v2 import workflows
 from mistral.services import workflows as wf_service
 from mistral.tests.unit import base
 from mistral import utils
-from mistral.workbook import parser as spec_parser
 from mistral.workflow import states
 
 
@@ -80,18 +82,19 @@ wf1:
         result: "{$}"
 """
 
-WORKFLOW = """
+WORKFLOW_WITH_VAR_TASK_NAME = """
 ---
 version: '2.0'
 
-list_servers:
+engine_command_{task_name}:
 
   tasks:
-    list_servers:
+    {task_name}:
       action: nova.servers_list
-
 """
 
+WORKFLOW = WORKFLOW_WITH_VAR_TASK_NAME.format(task_name='task1')
+
 INVALID_WORKFLOW = """
 ---
 verstion: '2.0'
@@ -103,6 +106,34 @@ wf:
       action: std.echo output="Task 1"
 """
 
+WORKFLOW_WITH_LONG_TASK_NAME = """
+---
+version: '2.0'
+
+test_workflow:
+
+  tasks:
+    {long_task_name}:
+      action: std.noop
+
+"""
+
+WORKFLOW_WITH_LONG_JOIN_TASK_NAME = """
+---
+version: '2.0'
+
+test_workflow:
+
+  tasks:
+    task1:
+      on-success:
+        - {long_task_name}
+
+    {long_task_name}:
+      join: all
+
+"""
+
 
 class WorkflowServiceTest(base.DbTestCase):
     def test_create_workflows(self):
@@ -125,6 +156,11 @@ class WorkflowServiceTest(base.DbTestCase):
         self.assertEqual('wf2', wf2_spec.get_name())
         self.assertEqual('direct', wf2_spec.get_type())
 
+    def test_engine_commands_are_valid_task_names(self):
+        for name in workflows.ENGINE_COMMANDS:
+            wf = WORKFLOW_WITH_VAR_TASK_NAME.format(task_name=name)
+            wf_service.create_workflows(wf)
+
     def test_update_workflows(self):
         db_wfs = wf_service.create_workflows(WORKFLOW_LIST)
 
@@ -170,7 +206,7 @@ class WorkflowServiceTest(base.DbTestCase):
             WORKFLOW
         )
 
-        self.assertIn("Workflow not found", exception.message)
+        self.assertIn("Workflow not found", str(exception))
 
     def test_invalid_workflow_list(self):
         exception = self.assertRaises(
@@ -179,7 +215,7 @@ class WorkflowServiceTest(base.DbTestCase):
             INVALID_WORKFLOW
         )
 
-        self.assertIn("Invalid DSL", exception.message)
+        self.assertIn("Invalid DSL", str(exception))
 
     def test_update_workflow_execution_env(self):
         wf_exec_template = {
@@ -219,13 +255,12 @@ class WorkflowServiceTest(base.DbTestCase):
                     update_env
                 )
 
-            self.assertDictEqual(update_env, updated.params['env'])
-            self.assertDictEqual(update_env, updated.context['__env'])
+                self.assertDictEqual(update_env, updated.params['env'])
 
-            fetched = db_api.get_workflow_execution(created.id)
+                fetched = db_api.get_workflow_execution(created.id)
 
-            self.assertEqual(updated, fetched)
-            self.assertIsNotNone(fetched.updated_at)
+                self.assertEqual(updated, fetched)
+                self.assertIsNotNone(fetched.updated_at)
 
     def test_update_workflow_execution_env_wrong_state(self):
         wf_exec_template = {
@@ -279,3 +314,41 @@ class WorkflowServiceTest(base.DbTestCase):
                 wf_exec['context']['__env'],
                 fetched.context['__env']
             )
+
+    def test_with_long_task_name(self):
+        long_task_name = utils.generate_string(tasks.MAX_LENGTH_TASK_NAME + 1)
+        workflow = WORKFLOW_WITH_LONG_TASK_NAME.format(
+            long_task_name=long_task_name)
+
+        self.assertRaises(
+            exc.InvalidModelException,
+            wf_service.create_workflows,
+            workflow
+        )
+
+    def test_upper_bound_length_task_name(self):
+        long_task_name = utils.generate_string(tasks.MAX_LENGTH_TASK_NAME)
+        workflow = WORKFLOW_WITH_LONG_TASK_NAME.format(
+            long_task_name=long_task_name)
+
+        wf_service.create_workflows(workflow)
+
+    def test_with_long_join_task_name(self):
+        long_task_name = utils.generate_string(
+            tasks.MAX_LENGTH_JOIN_TASK_NAME + 1
+        )
+        workflow = WORKFLOW_WITH_LONG_JOIN_TASK_NAME.format(
+            long_task_name=long_task_name)
+
+        self.assertRaises(
+            exc.InvalidModelException,
+            wf_service.create_workflows,
+            workflow
+        )
+
+    def test_upper_bound_length_join_task_name(self):
+        long_task_name = utils.generate_string(tasks.MAX_LENGTH_JOIN_TASK_NAME)
+        workflow = WORKFLOW_WITH_LONG_JOIN_TASK_NAME.format(
+            long_task_name=long_task_name)
+
+        wf_service.create_workflows(workflow)
