@@ -1,5 +1,3 @@
-# -*- encoding: utf-8 -*-
-#
 # Copyright 2014 Rackspace Hosting.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
@@ -18,6 +16,7 @@ import six
 
 from mistral import exceptions
 from mistral.tests.unit import base
+from mistral.utils import inspect_utils
 
 
 class ExceptionTestCase(base.BaseTest):
@@ -25,27 +24,46 @@ class ExceptionTestCase(base.BaseTest):
 
     def test_nf_with_message(self):
         exc = exceptions.DBEntityNotFoundError('check_for_this')
-        self.assertIn('check_for_this',
-                      six.text_type(exc))
+
+        self.assertIn('check_for_this', six.text_type(exc))
         self.assertEqual(404, exc.http_code)
 
     def test_nf_with_no_message(self):
         exc = exceptions.DBEntityNotFoundError()
-        self.assertIn("Object not found",
-                      six.text_type(exc))
+
+        self.assertIn("Object not found", six.text_type(exc))
         self.assertEqual(404, exc.http_code,)
 
     def test_duplicate_obj_code(self):
         exc = exceptions.DBDuplicateEntryError()
-        self.assertIn("Database object already exists",
-                      six.text_type(exc))
+
+        self.assertIn("Database object already exists", six.text_type(exc))
         self.assertEqual(409, exc.http_code,)
 
     def test_default_code(self):
         exc = exceptions.EngineException()
+
         self.assertEqual(500, exc.http_code)
 
     def test_default_message(self):
         exc = exceptions.EngineException()
-        self.assertIn("An unknown exception occurred",
-                      six.text_type(exc))
+
+        self.assertIn("An unknown exception occurred", six.text_type(exc))
+
+    def test_one_param_initializer(self):
+        # NOTE: this test is needed because at some places in the code we
+        # have to assume that every class derived from MistralException
+        # has an initializer with one "message" parameter.
+
+        # Let's traverse the MistralException class hierarchy recursively
+        # and check if it has the required initializer.
+        base_classes = [exceptions.MistralException]
+
+        for base_class in base_classes:
+            for subclass in base_class.__subclasses__():
+                arg_list = inspect_utils.get_arg_list(subclass.__init__)
+
+                self.assertEqual(1, len(arg_list))
+                self.assertEqual('message', arg_list[0])
+
+                base_classes.extend(subclass.__subclasses__())
