@@ -1,7 +1,6 @@
-# -*- coding: utf-8 -*-
-#
 # Copyright 2013 - Mirantis, Inc.
 # Copyright 2015 - Huawei Technologies Co. Ltd
+# Copyright 2016 - Brocade Communications Systems, Inc.
 #
 #    Licensed under the Apache License, Version 2.0 (the "License");
 #    you may not use this file except in compliance with the License.
@@ -16,33 +15,42 @@
 #    limitations under the License.
 
 import contextlib
+import datetime
+import functools
 import json
-import logging
 import os
 from os import path
 import shutil
-import six
 import socket
+import string
+import sys
 import tempfile
 import threading
-import uuid
 
 import eventlet
 from eventlet import corolocal
 from oslo_concurrency import processutils
+from oslo_log import log as logging
+from oslo_utils import timeutils
+from oslo_utils import uuidutils
 import pkg_resources as pkg
 import random
 
 from mistral import exceptions as exc
-from mistral import version
 
 
 # Thread local storage.
 _th_loc_storage = threading.local()
+ACTION_TASK_TYPE = 'ACTION'
+WORKFLOW_TASK_TYPE = 'WORKFLOW'
 
 
 def generate_unicode_uuid():
-    return six.text_type(str(uuid.uuid4()))
+    return uuidutils.generate_uuid()
+
+
+def is_valid_uuid(uuid_string):
+    return uuidutils.is_uuid_like(uuid_string)
 
 
 def _get_greenlet_local_storage():
@@ -73,7 +81,7 @@ def get_thread_local(var_name):
 
 
 def set_thread_local(var_name, val):
-    if not val and has_thread_local(var_name):
+    if val is None and has_thread_local(var_name):
         gl_storage = _get_greenlet_local_storage()
 
         # Delete variable from greenlet local storage.
@@ -84,7 +92,7 @@ def set_thread_local(var_name, val):
         if gl_storage and len(gl_storage) == 0:
             del _th_loc_storage.greenlet_locals[corolocal.get_ident()]
 
-    if val:
+    if val is not None:
         gl_storage = _get_greenlet_local_storage()
         if not gl_storage:
             gl_storage = _th_loc_storage.greenlet_locals[
@@ -100,6 +108,7 @@ def log_exec(logger, level=logging.DEBUG):
     """
 
     def _decorator(func):
+        @functools.wraps(func)
         def _logged(*args, **kw):
             params_repr = ("[args=%s, kw=%s]" % (str(args), str(kw))
                            if args or kw else "")
@@ -133,7 +142,7 @@ def merge_dicts(left, right, overwrite=True):
     if right is None:
         return left
 
-    for k, v in six.iteritems(right):
+    for k, v in right.items():
         if k not in left:
             left[k] = v
         else:
@@ -147,26 +156,161 @@ def merge_dicts(left, right, overwrite=True):
     return left
 
 
+def update_dict(left, right):
+    """Updates left dict with content from right dict
+
+    :param left: Left dict.
+    :param right: Right dict.
+    :return: the updated left dictionary.
+    """
+
+    if left is None:
+        return right
+
+    if right is None:
+        return left
+
+    left.update(right)
+
+    return left
+
+
 def get_file_list(directory):
-    base_path = pkg.resource_filename(
-        version.version_info.package,
-        directory
-    )
+    base_path = pkg.resource_filename("mistral", directory)
 
     return [path.join(base_path, f) for f in os.listdir(base_path)
             if path.isfile(path.join(base_path, f))]
 
 
+def cut_dict(d, length=100):
+    """Removes dictionary entries according to the given length.
+
+    This method removes a number of entries, if needed, so that a
+    string representation would fit into the given length.
+    The intention of this method is to optimize truncation of string
+    representation for dictionaries where the exact precision is not
+    critically important. Otherwise, we'd always have to convert a dict
+    into a string first and then shrink it to a needed size which will
+    increase memory footprint and reduce performance in case of large
+    dictionaries (i.e. tens of thousands entries).
+    Note that the method, due to complexity of the algorithm, has some
+    non-zero precision which depends on exact keys and values placed into
+    the dict. So for some dicts their reduced string representations will
+    be only approximately equal to the given value (up to around several
+    chars difference).
+
+    :param d: A dictionary.
+    :param length: A length limiting the dictionary string representation.
+    :return: A dictionary which is a subset of the given dictionary.
+    """
+
+    if not isinstance(d, dict):
+        raise ValueError("A dictionary is expected, got: %s" % type(d))
+
+    res = "{"
+
+    idx = 0
+
+    for key, value in d.items():
+        k = str(key)
+        v = str(value)
+
+        # Processing key.
+        new_len = len(k)
+
+        is_str = isinstance(key, str)
+
+        if is_str:
+            new_len += 2    # Account for the quotation marks
+
+        if 0 <= length <= new_len + len(res):
+            res += "'%s" % k if is_str else k
+            break
+        else:
+            res += "'%s'" % k if is_str else k
+
+        res += ": "
+
+        # Processing value.
+        new_len = len(v)
+
+        is_str = isinstance(value, str)
+
+        if is_str:
+            new_len += 2
+
+        if 0 <= length <= new_len + len(res):
+            res += "'%s" % v if is_str else v
+            break
+        else:
+            res += "'%s'" % v if is_str else v
+
+        res += ', ' if idx < len(d) - 1 else '}'
+
+        idx += 1
+
+    if 0 <= length <= len(res) and res[length - 1] is not '}':
+        res = res[:length - 3] + '...'
+
+    return res
+
+
+def cut_list(l, length=100):
+    if not isinstance(l, list):
+        raise ValueError("A list is expected, got: %s" % type(l))
+
+    res = '['
+
+    for idx, item in enumerate(l):
+        s = str(item)
+
+        new_len = len(res) + len(s)
+
+        is_str = isinstance(item, str)
+
+        if is_str:
+            new_len += 2
+
+        if 0 <= length <= new_len:
+            res += "'%s" % s if is_str else s
+            break
+        else:
+            res += "'%s'" % s if is_str else s
+        res += ', ' if idx < len(l) - 1 else ']'
+
+    if 0 <= length <= len(res) and res[length - 1] is not ']':
+        res = res[:length - 3] + '...'
+
+    return res
+
+
+def cut_string(s, length=100):
+    if 0 <= length < len(s):
+        return "%s..." % s[:length]
+
+    return s
+
+
 def cut(data, length=100):
     if not data:
         return data
 
-    string = str(data)
+    if isinstance(data, list):
+        return cut_list(data, length=length)
 
-    if len(string) > length:
-        return "%s..." % string[:length]
-    else:
-        return string
+    if isinstance(data, dict):
+        return cut_dict(data, length=length)
+
+    return cut_string(str(data), length=length)
+
+
+def cut_by_kb(data, kilobytes):
+    length = get_number_of_chars_from_kilobytes(kilobytes)
+    return cut(data, length)
+
+
+def cut_by_char(data, length):
+    return cut(data, length)
 
 
 def iter_subclasses(cls, _seen=None):
@@ -179,7 +323,7 @@ def iter_subclasses(cls, _seen=None):
 
     try:
         subs = cls.__subclasses__()
-    except TypeError:   # fails only when cls is type
+    except TypeError:  # fails only when cls is type
         subs = cls.__subclasses__(cls)
 
     for sub in subs:
@@ -206,54 +350,67 @@ def random_sleep(limit=1):
 
 
 class NotDefined(object):
-    """This class is just a marker of input params without value."""
+    """Marker of an empty value.
+
+    In a number of cases None can't be used to express the semantics of
+    a not defined value because None is just a normal value rather than
+    a value set to denote that it's not defined. This class can be used
+    in such cases instead of None.
+    """
 
     pass
 
 
-def get_dict_from_string(input_string, delimiter=','):
-    if not input_string:
+def get_number_of_chars_from_kilobytes(kilobytes):
+    bytes_per_char = sys.getsizeof('s') - sys.getsizeof('')
+    total_number_of_chars = int(kilobytes * 1024 / bytes_per_char)
+    return total_number_of_chars
+
+
+def get_dict_from_string(string, delimiter=','):
+    if not string:
         return {}
 
-    raw_inputs = input_string.split(delimiter)
+    kv_dicts = []
 
-    inputs = []
-
-    for raw in raw_inputs:
-        input = raw.strip()
-        name_value = input.split('=')
-
-        if len(name_value) > 1:
+    for kv_pair_str in string.split(delimiter):
+        kv_str = kv_pair_str.strip()
+        kv_list = kv_str.split('=')
 
+        if len(kv_list) > 1:
             try:
-                value = json.loads(name_value[1])
+                value = json.loads(kv_list[1])
             except ValueError:
-                value = name_value[1]
+                value = kv_list[1]
 
-            inputs += [{name_value[0]: value}]
+            kv_dicts += [{kv_list[0]: value}]
         else:
-            inputs += [name_value[0]]
+            kv_dicts += [kv_list[0]]
 
-    return get_input_dict(inputs)
+    return get_dict_from_entries(kv_dicts)
 
 
-def get_input_dict(inputs):
-    """Transform input list to dictionary.
+def get_dict_from_entries(entries):
+    """Transforms a list of entries into dictionary.
 
-    Ensure every input param has a default value(it will be a NotDefined
-    object if it's not provided).
+    :param entries: A list of entries.
+        If an entry is a dictionary the method simply updates the result
+        dictionary with its content.
+        If an entry is not a dict adds {entry, NotDefined} into the result.
     """
-    input_dict = {}
-    for x in inputs:
-        if isinstance(x, dict):
-            input_dict.update(x)
-        else:
-            # NOTE(xylan): we put a NotDefined class here as the value of
-            # param without value specified, to distinguish from the valid
-            # values such as None, ''(empty string), etc.
-            input_dict[x] = NotDefined
 
-    return input_dict
+    result = {}
+
+    for e in entries:
+        if isinstance(e, dict):
+            result.update(e)
+        else:
+            # NOTE(kong): we put NotDefined here as the value of
+            # param without value specified, to distinguish from
+            # the valid values such as None, ''(empty string), etc.
+            result[e] = NotDefined
+
+    return result
 
 
 def get_process_identifier():
@@ -329,3 +486,55 @@ def generate_key_pair(key_length=2048):
         public_key = open(public_key_path).read()
 
         return private_key, public_key
+
+
+def utc_now_sec():
+    """Returns current time and drops microseconds."""
+
+    return drop_microseconds(timeutils.utcnow())
+
+
+def drop_microseconds(date):
+    """Drops microseconds and returns date."""
+    return date.replace(microsecond=0)
+
+
+def datetime_to_str(val, sep=' '):
+    """Converts datetime value to string.
+
+    If the given value is not an instance of datetime then the method
+    returns the same value.
+
+    :param val: datetime value.
+    :param sep: Separator between date and time.
+    :return: Datetime as a string.
+    """
+    if isinstance(val, datetime.datetime):
+        return val.isoformat(sep)
+
+    return val
+
+
+def datetime_to_str_in_dict(d, key, sep=' '):
+    """Converts datetime value in te given dict to string.
+
+    :param d: A dictionary.
+    :param key: The key for which we need to convert the value.
+    :param sep: Separator between date and time.
+    """
+    val = d.get(key)
+
+    if val is not None:
+        d[key] = datetime_to_str(d[key], sep=sep)
+
+
+def generate_string(length):
+    """Returns random string.
+
+    :param length: the length of returned string
+    """
+
+    return ''.join(random.choice(
+        string.ascii_uppercase + string.digits)
+        for _ in range(length)
+    )
