@@ -1,6 +1,5 @@
-# -*- coding: utf-8 -*-
-#
 # Copyright 2014 - Mirantis, Inc.
+# Copyright 2016 - Brocade Communications Systems, Inc.
 #
 #    Licensed under the Apache License, Version 2.0 (the "License");
 #    you may not use this file except in compliance with the License.
@@ -17,15 +16,25 @@
 import functools
 import json
 
+from oslo_db import exception as db_exc
+from oslo_log import log as logging
 import pecan
 import six
-
+import sqlalchemy as sa
+import tenacity
 import webob
 from wsme import exc as wsme_exc
 
+
+from mistral import context as auth_ctx
+from mistral.db import utils as db_utils
+from mistral.db.v2.sqlalchemy import api as db_api
 from mistral import exceptions as exc
 
 
+LOG = logging.getLogger(__name__)
+
+
 def wrap_wsme_controller_exception(func):
     """Decorator for controllers method.
 
@@ -39,6 +48,8 @@ def wrap_wsme_controller_exception(func):
         except (exc.MistralException, exc.MistralError) as e:
             pecan.response.translatable_error = e
 
+            LOG.error('Error during API call: %s', str(e))
+
             raise wsme_exc.ClientSideError(
                 msg=six.text_type(e),
                 status_code=e.http_code
@@ -58,10 +69,13 @@ def wrap_pecan_controller_exception(func):
         try:
             return func(*args, **kwargs)
         except (exc.MistralException, exc.MistralError) as e:
+            LOG.error('Error during API call: %s', str(e))
+
             return webob.Response(
                 status=e.http_code,
                 content_type='application/json',
-                body=json.dumps(dict(faultstring=six.text_type(e)))
+                body=json.dumps(dict(faultstring=six.text_type(e))),
+                charset='UTF-8'
             )
 
     return wrapped
@@ -118,11 +132,13 @@ def filters_to_dict(**kwargs):
 
 def get_all(list_cls, cls, get_all_function, get_function,
             resource_function=None, marker=None, limit=None,
-            sort_keys='created_at', sort_dirs='asc', fields='', **filters):
+            sort_keys=None, sort_dirs=None, fields=None,
+            all_projects=False, **filters):
     """Return a list of cls.
 
-    :param list_cls: Collection class (e.g.: Actions, Workflows, ...).
-    :param cls: Class (e.g.: Action, Workflow, ...).
+    :param list_cls: REST Resource collection class (e.g.: Actions,
+        Workflows, ...)
+    :param cls: REST Resource class (e.g.: Action, Workflow, ...)
     :param get_all_function: Request function to get all elements with
                              filtering (limit, marker, sort_keys, sort_dirs,
                              fields)
@@ -132,73 +148,90 @@ def get_all(list_cls, cls, get_all_function, get_function,
     :param limit: Optional. Maximum number of resources to return in a
                   single result. Default value is None for backward
                   compatibility.
-    :param sort_keys: Optional. Columns to sort results by.
-                      Default: created_at.
-    :param sort_dirs: Optional. Directions to sort corresponding to
+    :param sort_keys: Optional. List of columns to sort results by.
+                      Default: ['created_at'].
+    :param sort_dirs: Optional. List of directions to sort corresponding to
                       sort_keys, "asc" or "desc" can be chosen.
-                      Default: asc.
+                      Default: ['asc'].
     :param fields: Optional. A specified list of fields of the resource to
                    be returned. 'id' will be included automatically in
                    fields if it's provided, since it will be used when
                    constructing 'next' link.
     :param filters: Optional. A specified dictionary of filters to match.
+    :param all_projects: Optional. Get resources of all projects.
     """
+    sort_keys = ['created_at'] if sort_keys is None else sort_keys
+    sort_dirs = ['asc'] if sort_dirs is None else sort_dirs
+    fields = [] if fields is None else fields
+
     if fields and 'id' not in fields:
         fields.insert(0, 'id')
 
     validate_query_params(limit, sort_keys, sort_dirs)
     validate_fields(fields, cls.get_fields())
 
+    # Admin user can get all tenants resources, no matter they are private or
+    # public.
+    insecure = False
+
+    if (all_projects or
+            (auth_ctx.ctx().is_admin and filters.get('project_id', ''))):
+        insecure = True
+
     marker_obj = None
 
     if marker:
         marker_obj = get_function(marker)
 
-    list_to_return = []
+    def _get_all_function():
+        with db_api.transaction():
+            db_models = get_all_function(
+                limit=limit,
+                marker=marker_obj,
+                sort_keys=sort_keys,
+                sort_dirs=sort_dirs,
+                insecure=insecure,
+                **filters
+            )
 
-    if resource_function:
-        # do not filter fields yet, resource_function needs the ORM object
-        db_list = get_all_function(
-            limit=limit,
-            marker=marker_obj,
-            sort_keys=sort_keys,
-            sort_dirs=sort_dirs,
-            **filters
-        )
+            for db_model in db_models:
+                if resource_function:
+                    rest_resource = resource_function(db_model)
+                else:
+                    rest_resource = cls.from_db_model(db_model)
 
-        for data in db_list:
-            obj = resource_function(data)
+                rest_resources.append(rest_resource)
 
-            # filter fields using a loop instead of the ORM
-            if fields:
-                data = []
-                for f in fields:
-                    if hasattr(obj, f):
-                        data.append(getattr(obj, f))
+    rest_resources = []
 
-                dict_data = dict(zip(fields, data))
-            else:
-                dict_data = obj.to_dict()
+    r = create_db_retry_object()
 
-            list_to_return.append(cls.from_dict(dict_data))
-    else:
-        db_list = get_all_function(
+    # If only certain fields are requested then we ignore "resource_function"
+    # parameter because it doesn't make sense anymore.
+    if fields:
+        # Use retries to prevent possible failures.
+        db_list = r.call(
+            get_all_function,
             limit=limit,
             marker=marker_obj,
             sort_keys=sort_keys,
             sort_dirs=sort_dirs,
             fields=fields,
+            insecure=insecure,
             **filters
         )
 
-        for data in db_list:
-            dict_data = (dict(zip(fields, data)) if fields else
-                         data.to_dict())
-
-            list_to_return.append(cls.from_dict(dict_data))
+        for obj_values in db_list:
+            # Note: in case if only certain fields have been requested
+            # "db_list" contains tuples with values of db objects.
+            rest_resources.append(
+                cls.from_tuples(zip(fields, obj_values))
+            )
+    else:
+        r.call(_get_all_function)
 
     return list_cls.convert_with_links(
-        list_to_return,
+        rest_resources,
         limit,
         pecan.request.host_url,
         sort_keys=','.join(sort_keys),
@@ -206,3 +239,29 @@ def get_all(list_cls, cls, get_all_function, get_function,
         fields=','.join(fields) if fields else '',
         **filters
     )
+
+
+class MistralRetrying(tenacity.Retrying):
+    def call(self, fn, *args, **kwargs):
+        try:
+            return super(MistralRetrying, self).call(fn, *args, **kwargs)
+        except tenacity.RetryError:
+            raise exc.MistralError("The service is temporarily unavailable")
+
+
+def create_db_retry_object():
+    return MistralRetrying(
+        retry=tenacity.retry_if_exception_type(
+            (
+                sa.exc.OperationalError,
+                db_exc.DBDeadlock,
+                db_exc.DBConnectionError
+            )
+        ),
+        stop=tenacity.stop_after_attempt(10),
+        wait=tenacity.wait_incrementing(increment=0.5)  # 0.5 seconds
+    )
+
+
+def rest_retry_on_db_error(func):
+    return db_utils.retry_on_db_error(func, create_db_retry_object())
