@@ -13,20 +13,18 @@
 #    See the License for the specific language governing permissions and
 #    limitations under the License.
 
-import copy
-
 from oslo_config import cfg
 from oslo_log import log as logging
 
 from mistral import context as auth_ctx
+from mistral.db.v2 import api as db_api
 from mistral.db.v2.sqlalchemy import models
 from mistral import exceptions as exc
 from mistral import expressions as expr
+from mistral.lang import parser as spec_parser
 from mistral import utils
 from mistral.utils import inspect_utils
-from mistral.workbook import parser as spec_parser
 from mistral.workflow import states
-from mistral.workflow import with_items
 
 LOG = logging.getLogger(__name__)
 CONF = cfg.CONF
@@ -174,7 +172,10 @@ def get_task_execution_result(task_ex):
     task_spec = spec_parser.get_task_spec(task_ex.spec)
 
     if task_spec.get_with_items():
-        if with_items.get_count(task_ex) > 0:
+        # TODO(rakhmerov): Smell: violation of 'with-items' encapsulation.
+        with_items_ctx = task_ex.runtime_context.get('with_items')
+
+        if with_items_ctx and with_items_ctx.get('count') > 0:
             return results
         else:
             return []
@@ -183,28 +184,50 @@ def get_task_execution_result(task_ex):
 
 
 def publish_variables(task_ex, task_spec):
-    if task_ex.state != states.SUCCESS:
+    if task_ex.state not in [states.SUCCESS, states.ERROR]:
         return
 
     wf_ex = task_ex.workflow_execution
 
     expr_ctx = ContextView(
+        get_current_task_dict(task_ex),
         task_ex.in_context,
+        get_workflow_environment_dict(wf_ex),
         wf_ex.context,
         wf_ex.input
     )
 
     if task_ex.name in expr_ctx:
         LOG.warning(
-            'Shadowing context variable with task name while publishing: %s' %
+            'Shadowing context variable with task name while '
+            'publishing: %s',
             task_ex.name
         )
 
-    task_ex.published = expr.evaluate_recursively(
-        task_spec.get_publish(),
-        expr_ctx
+    publish_spec = task_spec.get_publish(task_ex.state)
+
+    if not publish_spec:
+        return
+
+    # Publish branch variables.
+    branch_vars = publish_spec.get_branch()
+
+    task_ex.published = expr.evaluate_recursively(branch_vars, expr_ctx)
+
+    # Publish global variables.
+    global_vars = publish_spec.get_global()
+
+    utils.merge_dicts(
+        task_ex.workflow_execution.context,
+        expr.evaluate_recursively(global_vars, expr_ctx)
     )
 
+    # TODO(rakhmerov):
+    # 1. Publish atomic variables.
+    # 2. Add the field "publish" in TaskExecution model similar to "published"
+    #    but containing info as
+    #    {'branch': {vars}, 'global': {vars}, 'atomic': {vars}}
+
 
 def evaluate_task_outbound_context(task_ex):
     """Evaluates task outbound Data Flow context.
@@ -214,42 +237,55 @@ def evaluate_task_outbound_context(task_ex):
     :param task_ex: DB task.
     :return: Outbound task Data Flow context.
     """
+
+    # NOTE(rakhmerov): 'task_ex.in_context' has the SQLAlchemy specific
+    # type MutableDict. So we need to create a shallow copy using dict(...)
+    # initializer and use it. It's enough to be safe in order to manipulate
+    # with entries of the result dictionary, like adding more entries.
+    # However, we must not change values themselves because they are
+    # shared between the original dictionary and the newly created.
+    # It's better to avoid using the method copy.deepcopy() because on
+    # dictionaries with many entries it significantly increases memory
+    # footprint and reduces performance.
     in_context = (
-        copy.deepcopy(dict(task_ex.in_context))
+        dict(task_ex.in_context)
         if task_ex.in_context is not None else {}
     )
 
     return utils.update_dict(in_context, task_ex.published)
 
 
-def evaluate_workflow_output(wf_ex, wf_spec, ctx):
+def evaluate_workflow_output(wf_ex, wf_output, ctx):
     """Evaluates workflow output.
 
     :param wf_ex: Workflow execution.
-    :param wf_spec: Workflow specification.
+    :param wf_output: Workflow output.
     :param ctx: Final Data Flow context (cause task's outbound context).
     """
 
-    output_dict = wf_spec.get_output()
-
     # Evaluate workflow 'output' clause using the final workflow context.
-    ctx_view = ContextView(ctx, wf_ex.context, wf_ex.input)
+    ctx_view = ContextView(
+        ctx,
+        get_workflow_environment_dict(wf_ex),
+        wf_ex.context,
+        wf_ex.input
+    )
 
-    output = expr.evaluate_recursively(output_dict, ctx_view)
+    output = expr.evaluate_recursively(wf_output, ctx_view)
 
     # TODO(rakhmerov): Many don't like that we return the whole context
     # if 'output' is not explicitly defined.
     return output or ctx
 
 
-def add_current_task_to_context(ctx, task_id, task_name):
-    ctx['__task_execution'] = {
-        'id': task_id,
-        'name': task_name
+def get_current_task_dict(task_ex):
+    return {
+        '__task_execution': {
+            'id': task_ex.id,
+            'name': task_ex.name
+        }
     }
 
-    return ctx
-
 
 def add_openstack_data_to_context(wf_ex):
     wf_ex.context = wf_ex.context or {}
@@ -257,8 +293,6 @@ def add_openstack_data_to_context(wf_ex):
     if CONF.pecan.auth_enable:
         exec_ctx = auth_ctx.ctx()
 
-        LOG.debug('Data flow security context: %s' % exec_ctx)
-
         if exec_ctx:
             wf_ex.context.update({'openstack': exec_ctx.to_dict()})
 
@@ -266,21 +300,7 @@ def add_openstack_data_to_context(wf_ex):
 def add_execution_to_context(wf_ex):
     wf_ex.context = wf_ex.context or {}
 
-    wf_ex.context['__execution'] = {
-        'id': wf_ex.id
-    }
-
-
-def add_environment_to_context(wf_ex):
-    # TODO(rakhmerov): This is redundant, we can always get env from WF params
-    wf_ex.context = wf_ex.context or {}
-
-    # If env variables are provided, add an evaluated copy into the context.
-    if 'env' in wf_ex.params:
-        env = copy.deepcopy(wf_ex.params['env'])
-
-        # An env variable can be an expression of other env variables.
-        wf_ex.context['__env'] = expr.evaluate_recursively(env, {'__env': env})
+    wf_ex.context['__execution'] = {'id': wf_ex.id}
 
 
 def add_workflow_variables_to_context(wf_ex, wf_spec):
@@ -288,7 +308,11 @@ def add_workflow_variables_to_context(wf_ex, wf_spec):
 
     # The context for calculating workflow variables is workflow input
     # and other data already stored in workflow initial context.
-    ctx_view = ContextView(wf_ex.context, wf_ex.input)
+    ctx_view = ContextView(
+        get_workflow_environment_dict(wf_ex),
+        wf_ex.context,
+        wf_ex.input
+    )
 
     wf_vars = expr.evaluate_recursively(wf_spec.get_vars(), ctx_view)
 
@@ -302,3 +326,17 @@ def evaluate_object_fields(obj, context):
 
     for k, v in evaluated_fields.items():
         setattr(obj, k, v)
+
+
+def get_workflow_environment_dict(wf_ex):
+    if not wf_ex:
+        return {}
+
+    if wf_ex.root_execution_id:
+        return get_workflow_environment_dict(
+            db_api.get_workflow_execution(wf_ex.root_execution_id)
+        )
+
+    env_dict = wf_ex.params['env'] if 'env' in wf_ex.params else {}
+
+    return {'__env': env_dict}
