@@ -18,14 +18,18 @@ import inspect
 import json
 import os
 
+from aodhclient.v2 import base as aodh_base
+from aodhclient.v2 import client as aodhclient
 from barbicanclient import base as barbican_base
 from barbicanclient import client as barbicanclient
-from ceilometerclient.v2 import client as ceilometerclient
-from cinderclient.openstack.common.apiclient import base as cinder_base
+from cinderclient.apiclient import base as cinder_base
 from cinderclient.v2 import client as cinderclient
 from designateclient import client as designateclient
 from glanceclient.v2 import client as glanceclient
-from heatclient.openstack.common.apiclient import base as heat_base
+from glareclient.v1 import client as glareclient
+from gnocchiclient.v1 import base as gnocchi_base
+from gnocchiclient.v1 import client as gnocchiclient
+from heatclient.common import base as heat_base
 from heatclient.v1 import client as heatclient
 from ironicclient.common import base as ironic_base
 from ironicclient.v1 import client as ironicclient
@@ -33,6 +37,8 @@ from keystoneclient import base as keystone_base
 from keystoneclient.v3 import client as keystoneclient
 from magnumclient.common import base as magnum_base
 from magnumclient.v1 import client as magnumclient
+from manilaclient import base as manila_base
+from manilaclient.v2 import client as manilaclient
 from mistralclient.api import base as mistral_base
 from mistralclient.api.v2 import client as mistralclient
 from muranoclient.common import base as murano_base
@@ -84,8 +90,11 @@ BASE_MISTRAL_MANAGER = mistral_base.ResourceManager
 BASE_TROVE_MANAGER = trove_base.Manager
 BASE_IRONIC_MANAGER = ironic_base.Manager
 BASE_BARBICAN_MANAGER = barbican_base.BaseEntityManager
+BASE_MANILA_MANAGER = manila_base.Manager
 BASE_MAGNUM_MANAGER = magnum_base.Manager
 BASE_MURANO_MANAGER = murano_base.Manager
+BASE_AODH_MANAGER = aodh_base.Manager
+BASE_GNOCCHI_MANAGER = gnocchi_base.Manager
 
 
 def get_parser():
@@ -132,11 +141,6 @@ GLANCE_NAMESPACE_LIST = [
     'metadefs_tag', 'metadefs_namespace', 'versions'
 ]
 
-CEILOMETER_NAMESPACE_LIST = [
-    'alarms', 'capabilities', 'event_types', 'events', 'meters',
-    'new_samples', 'query_alarm_history', 'query_alarms', 'query_samples',
-    'resources', 'samples', 'statistics', 'trait_descriptions', 'traits'
-]
 
 DESIGNATE_NAMESPACE_LIST = [
     'diagnostics', 'domains', 'quotas', 'records', 'reports', 'servers',
@@ -144,6 +148,9 @@ DESIGNATE_NAMESPACE_LIST = [
 ]
 
 
+GLARE_NAMESPACE_LIST = ['artifacts', 'versions']
+
+
 def get_nova_client(**kwargs):
     return novaclient.Client(2)
 
@@ -160,10 +167,6 @@ def get_heat_client(**kwargs):
     return heatclient.Client('')
 
 
-def get_ceilometer_client(**kwargs):
-    return ceilometerclient.Client('')
-
-
 def get_cinder_client(**kwargs):
     return cinderclient.Client()
 
@@ -199,10 +202,28 @@ def get_murano_client(**kwargs):
     return muranoclient.Client('')
 
 
+def get_aodh_client(**kwargs):
+    return aodhclient.Client('')
+
+
+def get_gnocchi_client(**kwargs):
+    return gnocchiclient.Client()
+
+
+def get_glare_client(**kwargs):
+    return glareclient.Client('')
+
+
+def get_manila_client(**kwargs):
+    return manilaclient.Client(
+        input_auth_token='token',
+        service_catalog_url='http://127.0.0.1:8786'
+    )
+
+
 CLIENTS = {
     'nova': get_nova_client,
     'heat': get_heat_client,
-    'ceilometer': get_ceilometer_client,
     'cinder': get_cinder_client,
     'keystone': get_keystone_client,
     'glance': get_glance_client,
@@ -213,6 +234,10 @@ CLIENTS = {
     'designate': get_designate_client,
     'magnum': get_magnum_client,
     'murano': get_murano_client,
+    'aodh': get_aodh_client,
+    'gnocchi': get_gnocchi_client,
+    'glare': get_glare_client,
+    'manila': get_manila_client,
     # 'neutron': get_nova_client
     # 'baremetal_introspection': ...
     # 'swift': ...
@@ -221,7 +246,6 @@ CLIENTS = {
 BASE_MANAGERS = {
     'nova': BASE_NOVA_MANAGER,
     'heat': BASE_HEAT_MANAGER,
-    'ceilometer': None,
     'cinder': BASE_CINDER_MANAGER,
     'keystone': BASE_KEYSTONE_MANAGER,
     'glance': None,
@@ -232,6 +256,10 @@ BASE_MANAGERS = {
     'designate': None,
     'magnum': BASE_MAGNUM_MANAGER,
     'murano': BASE_MURANO_MANAGER,
+    'aodh': BASE_AODH_MANAGER,
+    'gnocchi': BASE_GNOCCHI_MANAGER,
+    'glare': None,
+    'manila': BASE_MANILA_MANAGER,
     # 'neutron': BASE_NOVA_MANAGER
     # 'baremetal_introspection': ...
     # 'swift': ...
@@ -239,8 +267,8 @@ BASE_MANAGERS = {
 }
 NAMESPACES = {
     'glance': GLANCE_NAMESPACE_LIST,
-    'ceilometer': CEILOMETER_NAMESPACE_LIST,
-    'designate': DESIGNATE_NAMESPACE_LIST
+    'designate': DESIGNATE_NAMESPACE_LIST,
+    'glare': GLARE_NAMESPACE_LIST
 }
 ALLOWED_ATTRS = ['service_catalog', 'catalog']
 FORBIDDEN_METHODS = [
